/* Copyright (c) 2021, Haakan T. Johansson */
/* All rights reserved. */

/* Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the authors nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
  gcc -O3 trigmatch.c -o trigmatch
 */

/* Program to check that trigger idle pattern cannot match
 * for other offsets in idle pattern, or at data bursts.
 */

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define SYNC_LENGTH   16
#define SHORT_LENGTH  24
#define LONG_LENGTH   40
#define AUX_LENGTH     2
#define HALF_LENGTH  (LONG_LENGTH + AUX_LENGTH)
#define FULL_LENGTH  (SYNC_LENGTH + SHORT_LENGTH + AUX_LENGTH+\
		      LONG_LENGTH + AUX_LENGTH)
#define TEST_LENGTH  (2*FULL_LENGTH)

char ptnsync[SYNC_LENGTH+1];
int v1i, v2i;

int min_dist;
int best_dist;

char invert01(char c)
{
  if (c == '0')
    return '1';
  if (c == '1')
    return '0';
  return c;
}

void match_in_stream_off(const char *target,
			 const char *stream,
			 size_t off, size_t len,
			 int expect0mod64)
{
  int mismatch = 0;
  size_t i, j, k;
  char v_value = '.';

  char fix[TEST_LENGTH+1];
  char want[TEST_LENGTH+1];
  char diff[TEST_LENGTH+1];
  
  strcpy (fix, stream);
  memset (want, ' ', TEST_LENGTH);
  want[TEST_LENGTH] = 0;
  memcpy(want + off, target, len);
  memset (diff, ' ', TEST_LENGTH);
  diff[TEST_LENGTH] = 0;

  /* The idle pattern is fixed 01, so just copy it wherever it appears. */
  for (k = 0; k < TEST_LENGTH; k++)
    {
      if (fix[k] == '-')
	fix[k] = '0' + !(k & 1);
      if (fix[k] == 's')
	fix[k] = '0' +  (k & 1);
    }

  /* The sync pattern is fixed, so copy where it appears. */
  for (j = 0; j < 2; j++)
    for (i = 0; i < SYNC_LENGTH; i++)
      {
	k = j * FULL_LENGTH + i;

	if (fix[k] == '=')
	  fix[k] = ptnsync[i];
      }

  /* The data and aux pattern might be whatever, but in pairs. */
  for (j = 0; j < FULL_LENGTH; j++)
    {
      size_t k1 = j*2 + 0;
      size_t k2 = j*2 + 1;

      if (fix[k1] == '_' || fix[k1] == 'a' || fix[k1] == 'A')
	{
	  if (want[k1] == '0' || want[k1] == '1')
	    {
	      fix[k1] = want[k1];
	      if (fix[k2] == '_' || fix[k2] == 'a' || fix[k2] == 'A')
		fix[k2] = invert01(fix[k1]);
	      goto next_data;
	    }
	}
      if (fix[k2] == '_' || fix[k2] == 'a' || fix[k2] == 'A')
	{
	  if (want[k2] == '0' || want[k2] == '1')
	    {
	      fix[k2] = want[k2];
	      if (fix[k1] == '_' || fix[k1] == 'a' || fix[k1] == 'A')
		fix[k1] = invert01(fix[k2]);
	    }
	}
    next_data:
      ;
    }

  /* The v bits might be just right... */
  if (v1i != -1 && v2i != -1)
    for (j = 0; j < 2; j++)
      {
	size_t k1 = j * FULL_LENGTH + v1i;
	size_t k2 = j * FULL_LENGTH + v2i;

	if (want[k1] == '0' || want[k1] == '1')
	  {
	    fix[k1] = want[k1];
	    fix[k2] = invert01(fix[k1]);
	  }
	if (want[k2] == '0' || want[k2] == '1')
	  {
	    fix[k2] = want[k2];
	    fix[k1] = invert01(fix[k2]);
	  }
      }

  /* Calculate matches. */
  for (i = 0; i < SYNC_LENGTH; i++)
    {
      k = off + i;
      
      if (want[k] == '0' || want[k] == '1')
	{
	  if (fix[k] != want[k])
	    {
	      diff[k] = 'x';
	      mismatch++;
	    }
	}
      else if (want[k] == 'v')
	{
	  v_value = fix[k];
	}
      else if (want[k] == 'V')
	{
	  if ((fix[k] == '0' || fix[k] == '1') &&
	      (v_value == '0' || v_value == '1') &&
	      fix[k] != invert01(v_value))
	    {
	      diff[k] = 'X';
	      mismatch++;
	    }
	}
    }

  if (mismatch < min_dist)
    {
      if (off % FULL_LENGTH == 0 &&
	  expect0mod64)
	return;

      min_dist = mismatch;

      if (0)
	{
	  printf ("  Distance: %d (%s)\n",
		  mismatch,
		  expect0mod64 ? "plain" : "inverted");

	  printf ("%s\n", stream);
	  printf ("%s\n", fix);
	  printf ("%s\n", want);
	  printf ("%s\n", diff);
	}
    }
}

void match_in_stream(const char *target,
		     const char *stream,
		     int expect0mod64)
{
  size_t len_target;
  size_t len_stream;
  size_t off;

  len_target = strlen(target);
  len_stream = strlen(stream);

  for (off = 0; off < len_stream - len_target; off++)
    {
      match_in_stream_off(target, stream, off, len_target, expect0mod64);

      if (min_dist < best_dist)
	return;
    }
}

void test_ptnsync()
{
  char target[25];
  char target_inv[25];

  char stream[TEST_LENGTH+1];
  char stream2[TEST_LENGTH+1];
  int s1, s2;
  int i, j;

  strcpy(target, ptnsync);
  strcpy(target_inv, ptnsync);

  for (i = 0; i < SYNC_LENGTH; i++)
    target_inv[i] = invert01(target_inv[i]);

  memset(stream, '-', TEST_LENGTH);
  stream[TEST_LENGTH] = 0;

  for (i = 0; i < SYNC_LENGTH; i++)
    {
      char str;
      
      if (ptnsync[i] == '0' || ptnsync[i] == '1')
	str = '=';
      else
	str = ptnsync[i];

      stream[i            ] = str;
      stream[i+FULL_LENGTH] = str;
    }

  for (j = 0; j < TEST_LENGTH / (FULL_LENGTH/2); j++)
    for (i = 0; i < AUX_LENGTH; i++)
      {
	char str = 'a';
      
	stream[i+j*(FULL_LENGTH/2)+(FULL_LENGTH/2)-AUX_LENGTH] = str;
	stream[i+j*(FULL_LENGTH/2)+(FULL_LENGTH/2)-AUX_LENGTH] = str;
      }

  match_in_stream(target,     stream, 1);
  match_in_stream(target_inv, stream, 0);

  for (s1 = 0; s1 <= TEST_LENGTH;
       s1 += (s1 % HALF_LENGTH == HALF_LENGTH-AUX_LENGTH-8) ? 10 : 8)
  for (s2 = s1; s2 <= TEST_LENGTH;
       s2 += (s2 % HALF_LENGTH == HALF_LENGTH-AUX_LENGTH-8) ? 10 : 8)
      {
	int s1_data_start = s1 + 8;
	int s1_data_end   = s1_data_start + 32;
	int s1_aux_end    = s1_data_end   + 2;
	  
	if (s2 >= s1_data_end || s2 == TEST_LENGTH)
	  {
	    int data_bits;
	    int check_end;

	    int s2_data_start = s2 + 8;
	    int s2_data_end   = s2_data_start + 32;
	    int s2_aux_end    = s2_data_end + 2;

	    int i;

	    strcpy(stream2, stream);

	    for (i = s1; i < s1_data_start && i < TEST_LENGTH; i++)
	      if (stream2[i] != 'v' && stream2[i] != 'V')
		stream2[i] = 's';
	    for (i = s1_data_start; i < s1_data_end && i < TEST_LENGTH; i++)
	      stream2[i] = '_';
	    for (i = s1_data_end; i < s1_aux_end && i < TEST_LENGTH; i++)
	      stream2[i] = 'A';
	    for (i = s2; i < s2_data_start && i < TEST_LENGTH; i++)
	      if (stream2[i] != 'v' && stream2[i] != 'V')
		stream2[i] = 's';
	    for (i = s2_data_start; i < s2_data_end && i < TEST_LENGTH; i++)
	      stream2[i] = '_';
	    for (i = s2_data_end; i < s2_aux_end && i < TEST_LENGTH; i++)
	      stream2[i] = 'A';

	    /* printf ("%3d %3d \n", s1, s2); */

	    match_in_stream(target,     stream2, 1);
	    match_in_stream(target_inv, stream2, 0);

	    if (min_dist < best_dist)
	      return;
	  }
      }
}

/* The new search strategy with the following:
 *
 * Idle pattern is 64 bits:
 * - First 24 bits has the match-pattern, i.e. will not match
 *   anything else, however shifted or inverted.
 * - The match-pattern contains two aux bits.
 *   TODO! Change: this can be in the 40 bits, first two bits.
 * - Last 40 (38) bits is a repeated 10-pattern.
 *
 * A data transmission consist of 32 bits of data, where each
 * bit is repeated twice: first normal followed by inverted bit.
 *
 * A data transmission can start on any 8-slot boundary in
 * the idle pattern.  It is identfied by the transmission
 * being inverted for all 8 bits.  Aux bits have fixed value
 * 10 when this happens.
 *
 *
 */

/* An idle pattern consist of (8+32+2)*2 = 42*2 = 84 bits.
 *
 * - n (16 / 24) bits of match pattern.
 * - (24 / 16 bits of 10
 * - 2 aux bits
 * - 40 bits of 10
 * - 2 aux bits
 *
 * A data burst can happen at any 8-slot boundary, i.e.  at an offset
 * of 0, 8, 16, 24, 32.  Or the same added 42.  It cannot happen at
 * the aux bit emission point.
 *
 * The transmission is signified by the next 8 idle bits being
 * inverted.  This is then followed by 32 payload bits.  And then
 * followed by the two aux bits.  In case the start offset was 0, then
 * it is the natural aux position.  If not, then we have jumped passed
 * two aux bits, and instead inject them at the end of the data.  Thus
 * the overall period of the idle pattern is not affected.
 *
 * Thus aux bit delivery is never an issue.  However, the matching
 * pattern, or long stretch (40) of 10 pattern (which is used for
 * initial signal speed lock) may be interrupted.  The transmitter
 * will monitor this, and if a data transmission happened during all
 * of the last three idle blocks of the same kind, transmission of a
 * new data burst will be delayed.
 */







int main()
{
  uint32_t scan;
  uint32_t ptn;
  int v1 = -1, v2 = -1;
  int i, j;

  best_dist = 1;

  /* The two vV cannot be in the same 8-group. */
  /*
  for (v1 = 0; v1 < 16; v1++)
    for (v2 = v1 - (v1 % 8) + 8; v2 < 24; v2++)
  */
  /*
  for (v1 = 0; v1 < 23; v1++)
    for (v2 = v1 + 1; v2 < ((v1 + 7) & ~7); v2++)
      for (scan = 0; scan < 0x400000; scan++)
  */
  for (scan = 0; scan < (1 << SYNC_LENGTH); scan++)
    {
	  uint32_t scan1, scan2, scan3;
	  uint32_t ptnv0, ptnv1;
	  int maxdisparity = 0;

	  /* Parts of scan between the v locations. */
	  /*
	  scan1 = (scan          ) & ((1 <<  v1      ) - 1);
	  scan2 = (scan >>  v1   ) & ((1 << (v2-v1-1)) - 1);
	  scan3 = (scan >> (v2-1)) & ((1 << (24-v2-1)) - 1);

	  ptn = scan1 | (scan2 << (v1+1)) | (scan3 << (v2+1));
	  */
	  ptn = scan;
	  
	  /*
	  printf ("%2d %2d: %06x  %06x %06x %06x  %06x %06x\n",
		  v1, v2, scan, scan1, scan2, scan3, ptn, (~ptn) & 0xffffff);
	  */
	  ptnv0 = ptnv1 = ptn;
	  /*  
	  ptnv0 = ptn | (1 << v2);
	  ptnv1 = ptn | (1 << v1);
	  */
	  /*
	  for (i = 0; i < 24; i++)
	    if (i != v1 && i != v2)
	      ones += (ptn >> i) & 1;

	  if (ones != 11)
	    continue;
	  */
	  /* Make sure sync pattern has is balanced for each 8 bits. */
	  for (j = 0; j < SYNC_LENGTH; j += 8)
	    {
	      int ones = 0;

	      for (i = 0; i < 8; i++)
		ones += (ptnv0 >> (j+i)) & 1;
	      if (ones != 4)
		goto next_iter;
	    }

#define SAMEBIT(ptn,x,y) ((((ptn)>>(x))&1)==(((ptn)>>(y))&1))

	  /* Make sure edge bits, or next-edge bits are different, for
	   * each 8-bit group.  'ab......' or '.bc.....'  shall be
	   * different, and same for end '......gh' or '.....fg.'.
	   */
	  for (j = 0; j < SYNC_LENGTH; j += 8)
	    {
	      if ((SAMEBIT(ptnv0, j+0, j+1) && SAMEBIT(ptnv0, j+1, j+2)) ||
		  (SAMEBIT(ptnv1, j+0, j+1) && SAMEBIT(ptnv1, j+1, j+2)) ||
		  (SAMEBIT(ptnv0, j+6, j+7) && SAMEBIT(ptnv0, j+5, j+6)) ||
		  (SAMEBIT(ptnv1, j+6, j+7) && SAMEBIT(ptnv1, j+5, j+6)))
		goto next_iter;
	      /* Since we can be stricter, require the edges to be
	       * different.
	       */
	      if ((SAMEBIT(ptnv0, j+0, j+1)) ||
		  (SAMEBIT(ptnv1, j+0, j+1)) ||
		  (SAMEBIT(ptnv0, j+6, j+7)) ||
		  (SAMEBIT(ptnv1, j+6, j+7)))
		goto next_iter;
	    }

	  /* We do not want more than three of each kind in a row. */
	  for (i = 0; i < SYNC_LENGTH-3; i++)
	    if (((ptnv0 >> i) & 0xf) == 0x0 ||
		((ptnv0 >> i) & 0xf) == 0xf ||
		((ptnv1 >> i) & 0xf) == 0x0 ||
		((ptnv1 >> i) & 0xf) == 0xf)
	      goto next_iter;

	  for (j = 0; j < SYNC_LENGTH; j += 8)
	    {
	      int onesv0 = 0;
	      int onesv1 = 0;

	      for (i = 0; i < SYNC_LENGTH - j; i++)
		{
		  int zerosv0, zerosv1;

		  onesv0 += (ptnv0 >> (j+i)) & 1;
		  onesv1 += (ptnv1 >> (j+i)) & 1;

		  zerosv0 = i + 1 - onesv0;
		  zerosv1 = i + 1 - onesv1;

		  if (onesv0 - zerosv0 > maxdisparity)
		    maxdisparity = onesv0 - zerosv0;
		  if (onesv1 - zerosv1 > maxdisparity)
		    maxdisparity = onesv1 - zerosv1;
		  if (zerosv0 - onesv0 > maxdisparity)
		    maxdisparity = zerosv0 - onesv0;
		  if (zerosv1 - onesv1 > maxdisparity)
		    maxdisparity = zerosv1 - onesv1;
		}
	    }

	  if (maxdisparity > 3)
	    continue;

	  for (i = 0; i < SYNC_LENGTH; i++)
	    ptnsync[i] = ((ptn >> i) & 1) + '0';
	  /*
	  ptnsync[v1] = 'v';
	  ptnsync[v2] = 'V';
	  */
	  ptnsync[SYNC_LENGTH] = 0;

	  /* With v outermost, we still most have an edge inside that,
	   * to not have more than three bits the same in a row.
	   */
	  /*
	  if ((v1 ==  0 && ptnsync[ 1] == ptnsync[ 2]) ||
	      (v1 ==  1 && ptnsync[ 0] == ptnsync[ 2]) ||
	      (v1  >  1 && ptnsync[ 0] == ptnsync[ 1]) ||
	      (v2 == 23 && ptnsync[21] == ptnsync[22]) ||
	      (v2 == 22 && ptnsync[21] == ptnsync[23]) ||
	      (v2  < 22 && ptnsync[22] == ptnsync[23]))
	    continue;
	  */
	  min_dist = 100;

	  v1i = v1; v2i = v2;
	  test_ptnsync();

	  if (min_dist >= best_dist)
	    {
	      printf ("%x: %d %d   %s\n",
		      ptn, min_dist, maxdisparity,
		      ptnsync);

	      best_dist = min_dist;
	    }

	next_iter:
	  ;
	}
}
