-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.ALL;
use ieee.std_logic_unsigned.ALL;
use ieee.numeric_std.all;

entity tb_timing is
  port (clk     :  in  std_logic;
        clk90   :  in  std_logic;

        pin_in  :  in  std_logic_vector(158 downto 0);
        pin_out :  out std_logic_vector(93  downto 0)
        );
end tb_timing;

architecture RTL of tb_timing is

  signal ff_in : std_logic_vector(pin_in'range);
  signal ff_out: std_logic_vector(pin_out'range);

  signal send_clk   : std_logic := '0';

  signal recv_clk_orig : std_logic := '0';

  signal recv_slow_clk   : std_logic := '0';

  signal recv_fast_clk   : std_logic := '0';
  signal recv_fast_clk90 : std_logic := '0';

  -- Common

  signal send_info_bit    : std_logic := '0';

  signal recv_info_bit    : std_logic := '0';
  signal recv_msg_strobe  : std_logic := '0';

  signal bad_bit_count    : std_logic_vector(31 downto 0);
  signal test_cycle_count : std_logic_vector(31 downto 0);

  -- Signals

  signal single_symbol_mode : std_logic;
  signal multi_signal4_mode : std_logic;

  signal send_signals   : std_logic_vector(10 downto 0) := (others => '0');

  signal send_serial      : std_logic;
  signal send_serial_sync : std_logic;

  signal recv_serial      : std_logic;

  signal recv_signals    : std_logic_vector(10 downto 0);

  -- Clock

  signal send_tick_ns    : std_logic_vector(63 downto 0) := (others => '0');
  signal send_aux_sigs   : std_logic_vector(4 downto 0) := (others => '0');

  signal recv_tick_ns_lo : std_logic_vector(31 downto  0) := (others => '0');
  signal recv_tick_ns_hi : std_logic_vector(63 downto 32) := (others => '0');
  signal recv_aux_sigs   : std_logic_vector(4 downto 0) := (others => '0');

  signal pulse_period_clks : std_logic_vector(7 downto 0);
  signal duty_low_min_clks : std_logic_vector(7 downto 0);
  signal duty_low_max_clks : std_logic_vector(7 downto 0);
  signal eight_slot     : std_logic;

  signal pulse_period_ns  : std_logic_vector(9 downto 0);
  signal use_pulse_period_ns : std_logic;
  signal clk_period_ns    : std_logic_vector(9 downto 0);

  signal message_delay_ns : std_logic_vector(21 downto 0);
  signal receive_delay_ns : std_logic_vector(15 downto 0);

  signal expect_edge   : std_logic_vector(1 downto 0);
  signal last_edges    : std_logic_vector(7 downto 0);
  signal use_auto_edge : std_logic;
  signal auto_edge     : std_logic_vector(1 downto 0);

  signal sync_status   : std_logic_vector(2 downto 0);
  signal sync_lost     : std_logic_vector(2 downto 0);
  signal bad_signals   : std_logic_vector(4 downto 0);
  signal clear_status  : std_logic;

  -- 4phase decoder

  signal saz           : std_logic_vector(2 to 3);
  signal sbz           : std_logic_vector(2 to 3);
  signal scz           : std_logic_vector(2 to 3);
  signal sdz           : std_logic_vector(2 to 3);

  signal bit_out       : std_logic_vector(0 to 1);
  signal got_bit       : std_logic_vector(0 to 1);

begin

  process(recv_fast_clk)
  begin
    if (rising_edge(recv_fast_clk)) then
      ff_in <= pin_in;
      pin_out <= ff_out;
    end if;
  end process;

  gen_slow: if (true) generate
    recv_clk_orig <= clk;

    clock_inst: entity work.clock_gen
      generic map(divide_factor => 3)
      port map(clk_in  => recv_clk_orig,
               clk     => recv_fast_clk,
               clk90   => recv_fast_clk90,
               clk_div => recv_slow_clk);

    send_clk   <= recv_fast_clk;
  end generate;
  no_gen_slow: if (false) generate
    recv_clk_orig   <= clk;
    recv_fast_clk   <= clk;
    recv_fast_clk90 <= clk90;
    recv_slow_clk   <= clk;
  end generate;

  sig: if (false) generate
    send_signals  <= ff_in(10 downto 0);
    send_info_bit <= ff_in(11);

    recv_serial   <= ff_in(12);

    single_symbol_mode <= ff_in(13);
    multi_signal4_mode <= ff_in(14);

    ff_out(0)  <= send_serial;
    ff_out(1)  <= send_serial_sync;

    ff_out(12 downto 2) <= recv_signals;
    ff_out(13) <= recv_info_bit;
    ff_out(14) <= recv_msg_strobe;

    ff_out(46 downto 15) <= bad_bit_count;
    ff_out(50 downto 47) <= test_cycle_count(3 downto 0);

    sig_send: if (true) generate
      send_sig : entity work.rataser_signals_send
        port map(
          clk           => send_clk,

          signals       => send_signals,
          info_bit      => send_info_bit,

          single_symbol_mode => single_symbol_mode,
          multi_signal4_mode => multi_signal4_mode,

          test_mode     => '0',
          fault_shift_count => (others => '0'),

          transmit      => send_serial,
          transmit_sync => send_serial_sync
          );
    end generate;

    sig_recv: if (true) generate
      recv_sig : entity work.rataser_signals_recv
        generic map(pipeline_sym24_diff_in   => true,
                    pipeline_sym24_diff_1    => true,
                    pipeline_sym24_diff_2    => true,
                    pipeline_sym24_diff_out  => true,
                    pipeline_sym24_match_out => true)
        port map(
          fast_clk      => recv_fast_clk,
          fast_clk90    => recv_fast_clk90,

          slow_clk        => recv_slow_clk,
          slow_clk_jitter => recv_slow_clk,

          receive       => recv_serial,

          signals       => recv_signals,
          info_bit      => recv_info_bit,
          msg_strobe    => recv_msg_strobe,

          single_symbol_mode => single_symbol_mode,
          multi_signal4_mode => multi_signal4_mode,

          test_mode     => '0',

          bad_bit_count    => bad_bit_count,
          test_cycle_count => test_cycle_count
          );
    end generate;

  end generate;

  clock: if (true) generate
    send_tick_ns  <= ff_in(63 downto 0);
    send_aux_sigs <= ff_in(68 downto 64);
    send_info_bit <= ff_in(69);

    pulse_period_clks <= ff_in(77 downto 70);
    duty_low_min_clks <= ff_in(85 downto 78);
    duty_low_max_clks <= ff_in(93 downto 86);
    eight_slot    <= ff_in(94);

    recv_serial   <= ff_in(95);

    clk_period_ns <= ff_in(105 downto 96);

    pulse_period_ns <= ff_in(115 downto 106);
    use_pulse_period_ns <= ff_in(116);

    expect_edge   <= ff_in(118 downto 117);
    clear_status  <= ff_in(119);
    use_auto_edge <= ff_in(120);

    message_delay_ns <= ff_in(142 downto 121);
    receive_delay_ns <= ff_in(158 downto 143);

    ff_out(0)  <= send_serial;
    ff_out(1)  <= send_serial_sync;

    ff_out(33 downto  2) <= recv_tick_ns_lo;
    ff_out(65 downto 34) <= recv_tick_ns_hi;
    ff_out(70 downto 66) <= recv_aux_sigs;
    ff_out(71) <= recv_info_bit;
    ff_out(72) <= recv_msg_strobe;

    ff_out(80 downto 73) <= last_edges;
    ff_out(83 downto 81) <= sync_status;
    ff_out(86 downto 84) <= sync_lost;
    ff_out(91 downto 87) <= bad_signals;
    ff_out(93 downto 92) <= auto_edge;

    clock_send: if (false) generate
      send_clock : entity work.rataser_clock_send
        generic map(period_bits => 6)
        port map(
          clk           => send_clk,

          tick_ns       => send_tick_ns,
          aux_sigs      => send_aux_sigs,
          info_bit      => send_info_bit,

          pulse_period_clks => pulse_period_clks(5 downto 0),
          duty_low_min_clks  => duty_low_min_clks,
          duty_low_max_clks  => duty_low_max_clks,
          eight_slot    => eight_slot,

          pulse_period_ns => pulse_period_ns,
          use_pulse_period_ns => use_pulse_period_ns,

          message_delay_ns => message_delay_ns,

          transmit      => send_serial,
          transmit_sync => send_serial_sync
          );
    end generate;

    clock_recv: if (true) generate
      recv_clock : entity work.rataser_clock_recv
        generic map(period_bits => 6,
                    num_last_edges => 4)
        port map(
          clk           => recv_fast_clk,
          clk90         => recv_fast_clk90,

          receive       => recv_serial,
          eight_slot    => '1', -- eight_slot,

          expect_edge   => expect_edge,
          last_edges    => last_edges,
          use_auto_edge => use_auto_edge,
          auto_edge     => auto_edge,

          receive_delay_ns => receive_delay_ns,

          tick_ns_lo    => recv_tick_ns_lo,
          tick_ns_hi    => recv_tick_ns_hi,
          aux_sigs      => recv_aux_sigs,
          info_bit      => recv_info_bit,
          msg_strobe    => recv_msg_strobe,

          sync_status   => sync_status,
          sync_lost     => sync_lost,
          bad_signals   => bad_signals,
          clear_status  => clear_status,

          pulse_period_clks => pulse_period_clks(5 downto 0),
          clk_period_ns => clk_period_ns
          );
    end generate;

  end generate;

  fph: if (false) generate

    saz <= ff_in(1 downto 0);
    sbz <= ff_in(3 downto 2);
    scz <= ff_in(5 downto 4);
    sdz <= ff_in(7 downto 6);

    ff_out(1 downto 0) <= bit_out;
    ff_out(3 downto 2) <= got_bit;

    fphase: entity work.rataser_4phase_deser
      port map(
        clk  => recv_fast_clk,

        saz  => saz,
        sbz  => sbz,
        scz  => scz,
        sdz  => sdz,

        bit_out => bit_out,
        got_bit => got_bit);

  end generate;

end;
