-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use std.textio.all;
use ieee.std_logic_textio.all;

use work.all;

entity tb_rataclock is
  generic (seed : integer := 1);
end entity tb_rataclock;

architecture dataflow of tb_rataclock is

  -- Should simulation continue?
  signal runsim : boolean := true;
  
  -- Clock
  signal plain_clk   : std_logic := '0';
  signal drift_clk   : std_logic := '0';
  signal jitter_clk  : std_logic := '0';

  signal send_clk   : std_logic := '0';

  signal recv_clk   : std_logic := '0';
  signal recv_clk90 : std_logic := '0';
  
  constant period : time := 10 ns;

  -- Funky init to give carry into high 32 bits after 4 cycles sent, ca 1.5 ms.
  signal send_tick_ns   : std_logic_vector(63 downto 0) :=
    "00000000" & "00000000" & "00000000" & "00001011" &
    "11111111" & "00000000" & "00000000" & "00000000";
  signal send_aux_sigs  : std_logic_vector(4 downto 0) := (others => '0');
  signal send_info_bit  : std_logic := '0';

  signal send_serial      : std_logic := '0';
  signal send_serial_sync : std_logic := '0';

  signal fault_count1     : std_logic_vector(5 downto 0) := (others => '0');
  signal fault_count2     : std_logic_vector(5 downto 0) := (others => '0');
  signal fault_count3     : std_logic_vector(7 downto 0) := (others => '0');
  signal inject_fault     : std_logic := '0';

  signal recv_serial      : std_logic := '0';

  signal recv_tick_ns_lo : std_logic_vector(31 downto  0) := (others => '0');
  signal recv_tick_ns_hi : std_logic_vector(63 downto 32) := (others => '0');
  signal recv_aux_sigs   : std_logic_vector(4 downto 0) := (others => '0');
  signal recv_info_bit   : std_logic := '0';
  signal recv_msg_strobe : std_logic := '0';

  signal clock_period_ns : std_logic_vector(9 downto 0) :=
    --conv_std_logic_vector(100,10);
    --conv_std_logic_vector(30,10);
    conv_std_logic_vector(24,10);
    --conv_std_logic_vector(10,10);

  signal expect_edge     : std_logic_vector(1 downto 0) := "01";
  signal last_edges      : std_logic_vector(7 downto 0);

  signal recv_auto_edge  : std_logic_vector(1 downto 0);

  signal sync_status     : std_logic_vector(2 downto 0);
  signal sync_lost       : std_logic_vector(2 downto 0);
  signal bad_signals     : std_logic_vector(4 downto 0);

  signal clear_status    : std_logic := '0';

begin

  send_clk <= plain_clk;
  --send_clk <= drift_clk;
  -- send_clk <= jitter_clk;
  -- send_clk <= recv_clk;

  send_clock : entity work.rataser_clock_send
    generic map(period_bits => 6)
    port map(
      clk                => send_clk,

      tick_ns            => send_tick_ns,
      aux_sigs           => send_aux_sigs,
      info_bit           => send_info_bit,

      pulse_period_clks  => "001000", -- 4
      duty_low_min_clks  => "000010", -- 1
      duty_low_max_clks  => "000100", -- 2
      eight_slot         => '1',

      pulse_period_ns    => "0000000000", -- unused
      use_pulse_period_ns => '0',

      message_delay_ns   => (others => '0'),

      transmit           => send_serial,
      transmit_sync      => send_serial_sync
      );

  recv_serial <= send_serial; -- xor inject_fault;

  recv_clock : entity work.rataser_clock_recv
    generic map(period_bits => 6,
                num_last_edges => 4)
    port map(
      clk               => recv_clk,
      clk90             => recv_clk90,

      receive           => recv_serial,
      eight_slot        => '1',

      expect_edge       => expect_edge,
      last_edges        => last_edges,
      use_auto_edge     => '0',
      auto_edge         => recv_auto_edge,

      receive_delay_ns  => (others => '0'),

      tick_ns_lo        => recv_tick_ns_lo,
      tick_ns_hi        => recv_tick_ns_hi,
      aux_sigs          => recv_aux_sigs,
      info_bit          => recv_info_bit,
      msg_strobe        => recv_msg_strobe,

      sync_status       => sync_status,
      sync_lost         => sync_lost,
      bad_signals       => bad_signals,
      clear_status      => clear_status,

      pulse_period_clks => "001010", -- 4
      clk_period_ns     => clock_period_ns
      );

  -- Update the signals to send
  process(send_clk)
  begin
    if (rising_edge(send_clk)) then
      -- For pulse_period_clks = 3 and eight_slot = 0,
      -- 1000 is a good value (for test).
      send_tick_ns <= send_tick_ns + clock_period_ns;
    end if;
  end process;

  -- We are not able to get a clean init for the hamming decoders,
  -- so issue a clear.
  process
  begin
    clear_status <= '1';
    wait for 15.0 ns;
    clear_status <= '0';
    wait;
  end process;

  -- Inject bit-faults
  process(send_clk)
  begin
    if (rising_edge(send_clk)) then
      inject_fault <= '0';

      fault_count1 <= fault_count1 - 1;
      fault_count2 <= fault_count2 - 1;
      fault_count3 <= fault_count3 - 1;
      
      if (fault_count1 = 0) then
        fault_count1 <= conv_std_logic_vector(53-1,fault_count1'length);
        inject_fault <= '1';
      end if;
      if (fault_count2 = 0) then
        fault_count2 <= conv_std_logic_vector(59-1,fault_count2'length);
        inject_fault <= '1';
      end if;
      if (fault_count3 = 0) then
        fault_count3 <= conv_std_logic_vector(251-1,fault_count3'length);
        inject_fault <= '1';
      end if;
    end if;
  end process;

  -- Drive the clock
  process
  begin
    if runsim then
      recv_clk   <= '1';
      wait for 2.0 ns;
      recv_clk90 <= '1';
      wait for 2.0 ns;
      recv_clk   <= '0';
      wait for 2.0 ns;
      recv_clk90 <= '0';
      wait for 2.0 ns;
    else
      wait;
    end if;
  end process;

  -- A drifting clock for the sender
  process
  begin
    if runsim then
      plain_clk   <= '1';
      wait for 5.0 ns;
      plain_clk   <= '0';
      wait for 5.0 ns;
    else
      wait;
    end if;
  end process;

  -- A drifting clock for the sender
  process
  begin
    if runsim then
      drift_clk   <= '1';
      wait for 5.00002 ns; -- 0.00002 ns / 10 ns = 2e-6 ; 2 ns per 1 ms
      drift_clk   <= '0';
      wait for 5.0 ns;
    else
      wait;
    end if;
  end process;

  -- Sender clock
  process
    variable x : std_logic_vector(31 downto 0);
    variable add1, add2 : time;
    --variable l : line;
  begin
    x := std_logic_vector(to_unsigned(seed, x'length));
       
    while runsim loop
      x := x xor (  x(31-13 downto 0) & "0000000000000");
      x := x xor ("00000000000000000" & x(31 downto 17));
      x := x xor (   x(31-5 downto 0) & "00000"        );

      add1 := conv_integer(x( 7 downto  4)) * 0.01 ns;
      add2 := conv_integer(x(15 downto 12)) * 0.01 ns;

      --write (l, String'("OUTPUT: "));
      --hwrite (l, x);
      --write (l, String'(" "));
      --write (l, add1);
      --write (l, String'(" "));
      --write (l, add2);
      --writeline (output, l);

      jitter_clk <= '1';
      wait for (4.925 ns + add1);
      jitter_clk <= '0';
      wait for (5 ns);
      --wait for (4.925 ns + add2);
    end loop;
    wait;
  end process;

end architecture dataflow;
