-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use std.textio.all;
use ieee.std_logic_textio.all;

use work.all;
use work.rataser_util.all;

entity tb_ratasignals is
  generic (seed  : integer := 1;
           seed2 : integer := 2;
           seed3 : integer := 3);
end entity tb_ratasignals;

architecture dataflow of tb_ratasignals is

  -- Should simulation continue?
  signal runsim : boolean := true;
  
  -- Clock
  signal drift_clk   : std_logic := '0';

  signal send_clk   : std_logic := '0';

  signal slow_clk        : std_logic := '0';
  signal slow_clk_jitter : std_logic := '0';

  signal recv_clk   : std_logic := '0';
  signal recv_clk90 : std_logic := '0';
  
  constant period : time := 10 ns;

  signal send_signals   : std_logic_vector(10 downto 0) := (others => '0');
  signal send_info_bit  : std_logic := '0';

  signal send_serial      : std_logic;
  signal send_serial_sync : std_logic;

  signal fault_count1     : std_logic_vector(5 downto 0) := (others => '0');
  signal fault_count2     : std_logic_vector(5 downto 0) := (others => '0');
  signal fault_count3     : std_logic_vector(7 downto 0) := (others => '0');
  signal inject_fault     : std_logic := '0';

  signal fault_x          : std_logic_vector(31 downto 0) :=
    std_logic_vector(to_unsigned(seed2, 32));

  signal fault_shift_count : std_logic_vector(4 downto 0) := (others => '0');
  signal invert_signal    : std_logic := '0';

  signal recv_serial      : std_logic;

  signal recv_signals    : std_logic_vector(10 downto 0);
  signal recv_info_bit   : std_logic := '0';
  signal recv_msg_strobe : std_logic := '0';

  signal bad_bit_count    : std_logic_vector(31 downto 0);
  signal test_cycle_count : std_logic_vector(31 downto 0);

  signal single_symbol_mode : std_logic := '0';
  signal multi_signal4_mode : std_logic := '1';

  signal test_mode       : std_logic := '0';

begin

  send_clk <= drift_clk;

  send_sig : entity work.rataser_signals_send
    port map(
      clk           => send_clk,

      signals       => send_signals,
      info_bit      => send_info_bit,

      single_symbol_mode => single_symbol_mode,
      multi_signal4_mode => multi_signal4_mode,

      test_mode     => test_mode,

      fault_shift_count => fault_shift_count,

      transmit      => send_serial,
      transmit_sync => send_serial_sync
      );

  recv_serial <= send_serial xor inject_fault xor invert_signal;

  recv_sig : entity work.rataser_signals_recv
    generic map(use_slow_clk_jitter => false)
    port map(
      fast_clk      => recv_clk,
      fast_clk90    => recv_clk90,

      slow_clk      => slow_clk,
      slow_clk_jitter => slow_clk_jitter,

      receive       => recv_serial,

      signals       => recv_signals,
      info_bit      => recv_info_bit,
      msg_strobe    => recv_msg_strobe,

      single_symbol_mode => single_symbol_mode,
      multi_signal4_mode => multi_signal4_mode,

      test_mode     => test_mode,

      bad_bit_count => bad_bit_count,
      test_cycle_count => test_cycle_count
      );

  -- Update the signals to send
  process(send_clk)
  begin
    if (rising_edge(send_clk)) then
      if (send_serial_sync = '1') then
        send_signals <= send_signals + 1;
      end if;
    end if;
  end process;

  -- Inject bit-faults
  process(send_clk)
  begin
    if (rising_edge(send_clk)) then
      inject_fault <= '0';

      fault_count1 <= fault_count1 - 1;
      fault_count2 <= fault_count2 - 1;
      fault_count3 <= fault_count3 - 1;
      
      if (fault_count1 = 0) then
        fault_count1 <= conv_std_logic_vector(53-1,fault_count1'length);
        inject_fault <= '1';
      end if;
      if (fault_count2 = 0) then
        fault_count2 <= conv_std_logic_vector(59-1,fault_count2'length);
        inject_fault <= '1';
      end if;
      if (fault_count3 = 0) then
        fault_count3 <= conv_std_logic_vector(251-1,fault_count3'length);
        inject_fault <= '1';
      end if;

      fault_x <= rataser_xorshift(fault_x, 5, 17, 13);

      fault_shift_count <= (others => '0');

      if (fault_x(15 downto 5) = 0) then
        fault_shift_count <= fault_x(4 downto 0);
      end if;

      if (fault_x(27 downto 16) = 0) then
        invert_signal <= not invert_signal;
      end if;

    end if;
  end process;

  -- Drive the clock
  process
  begin
    if runsim then
      recv_clk   <= '1';
      wait for 2.5 ns;
      recv_clk90 <= '1';
      wait for 2.5 ns;
      recv_clk   <= '0';
      wait for 2.5 ns;
      recv_clk90 <= '0';
      wait for 2.5 ns;
    else
      wait;
    end if;
  end process;

  -- Sender clock
  process
    variable x : std_logic_vector(31 downto 0);
    variable add1, add2 : time;
    --variable l : line;
  begin
    x := std_logic_vector(to_unsigned(seed, x'length));
       
    while runsim loop
      x := x xor (  x(31-13 downto 0) & "0000000000000");
      x := x xor ("00000000000000000" & x(31 downto 17));
      x := x xor (   x(31-5 downto 0) & "00000"        );

      add1 := conv_integer(x( 7 downto  4)) * 0.01 ns;
      add2 := conv_integer(x(15 downto 12)) * 0.01 ns;

      --write (l, String'("OUTPUT: "));
      --hwrite (l, x);
      --write (l, String'(" "));
      --write (l, add1);
      --write (l, String'(" "));
      --write (l, add2);
      --writeline (output, l);

      drift_clk   <= '1';
      wait for (4.925 ns + add1);
      drift_clk   <= '0';
      wait for (5 ns);
      --wait for (4.925 ns + add2);
    end loop;
    wait;
  end process;

  -- Slow clock
  process
    variable x : std_logic_vector(31 downto 0);
    variable add1, add2 : time;
    --variable l : line;
  begin
    x := std_logic_vector(to_unsigned(seed3, x'length));
       
    while runsim loop
      x := x xor (  x(31-13 downto 0) & "0000000000000");
      x := x xor ("00000000000000000" & x(31 downto 17));
      x := x xor (   x(31-5 downto 0) & "00000"        );

      add1 := conv_integer(x( 7 downto  4)) * 0.01 ns;
      add2 := conv_integer(x(15 downto 12)) * 0.01 ns;

      --write (l, String'("OUTPUT: "));
      --hwrite (l, x);
      --write (l, String'(" "));
      --write (l, add1);
      --write (l, String'(" "));
      --write (l, add2);
      --writeline (output, l);

      slow_clk          <= '0';
      slow_clk_jitter   <= '0';
      wait for (3*(4.925 ns + add1 - 0.5 ns));

      if (x(31) = '0') then
        slow_clk_jitter <= '1';
      end if;
      
      wait for (3*(0.5 ns));
      slow_clk          <= '1';
      wait for (3*(0.5 ns));

      if (x(31) = '1') then
        slow_clk_jitter <= '1';
      end if;

      wait for (3*(5 ns            - 0.5 ns));
      --wait for (4.925 ns + add2);
    end loop;
    wait;
  end process;

end architecture dataflow;
