-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use std.textio.all;
use ieee.std_logic_textio.all;

-- Simulate for 3 ms such that count wraps.  Then all symbols
-- are tested.  (Actually 2^18*10 ns = 2.621440 ms)

use work.rataser_sym24word8_util.all;

entity tb_ratasym24 is
  generic (seed : integer := 1);
end entity tb_ratasym24;

architecture dataflow of tb_ratasym24 is

  -- Should simulation continue?
  signal runsim : boolean := true;

  -- Clock
  signal send_clk   : std_logic := '0';

  -- Counter of data to send (2*9 bits)
  signal count : unsigned(17 downto 0) := (others => '0');

  -- Patterns
  signal pattern1 : std_logic_vector(23 downto 0) := (others => '0');
  signal pattern2 : std_logic_vector(23 downto 0) := (others => '0');

  signal downto_pattern : std_logic_vector(47 downto 0) := (others => '0');
  signal to_pattern     : std_logic_vector(0     to 47) := (others => '0');
  signal to_pattern_not : std_logic_vector(0     to 47) := (others => '0');

  signal diffp : unsigned(25*4-1 downto 0) := (others => '0');
  signal matchp : std_logic_vector(24 downto 0) := (others => '0');

  signal diffn : unsigned(25*4-1 downto 0) := (others => '0');
  signal matchn : std_logic_vector(24 downto 0) := (others => '0');

begin

  -- Update the signals to send
  process(send_clk)
  begin
    if (rising_edge(send_clk)) then
      count <= count + 1;
    end if;
  end process;

  gen_sequence(seq  => pattern1,
               data => count(8 downto 1),
               aux  => count(0));
  gen_sequence(seq  => pattern2,
               data => count(17 downto 10),
               aux  => count(9));

  downto_pattern <= pattern1 & pattern2;

  copy: for i in 0 to 47 generate
    to_pattern(i) <= downto_pattern(47-i);
  end generate;

  to_pattern_not <= not to_pattern;

  check_p: for i in 0 to 24 generate
    d0: entity work.rataser_sym24word8_diff
      generic map(pipeline_1 => false,
                  pipeline_2 => false)
      port map(clk  => send_clk,
               seq  => to_pattern(i+0 to i+23),
               diff => diffp(i*4+3 downto i*4));

    matchp(i) <= '1' when (to_integer(diffp(i*4+3 downto i*4)) <= 1) else '0';
  end generate;

  check_n: for i in 0 to 24 generate
    d0: entity work.rataser_sym24word8_diff
      generic map(pipeline_1 => false,
                  pipeline_2 => false)
      port map(clk  => send_clk,
               seq  => to_pattern_not(i+0 to i+23),
               diff => diffn(i*4+3 downto i*4));

    matchn(i) <= '1' when (to_integer(diffn(i*4+3 downto i*4)) <= 1) else '0';
  end generate;

  -- Drive the clock
  process
  begin
    if runsim then
      send_clk   <= '1';
      wait for 5.0 ns;
      send_clk   <= '0';
      wait for 5.0 ns;
    else
      wait;
    end if;
  end process;

end architecture dataflow;
