-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use std.textio.all;
use ieee.std_logic_textio.all;

use work.all;
use work.rataser_util.all;

entity tb_ratatrig is
  generic (seed  : integer := 1;
           seed2 : integer := 2;
           seed3 : integer := 3);
end entity tb_ratatrig;

architecture dataflow of tb_ratatrig is

  -- Should simulation continue?
  signal runsim : boolean := true;
  
  -- Clock
  signal drift_clk   : std_logic := '0';

  signal send_clk   : std_logic := '0';

  signal slow_clk        : std_logic := '0';
  signal slow_clk_jitter : std_logic := '0';

  signal recv_clk   : std_logic := '0';
  signal recv_clk90 : std_logic := '0';
  
  constant period : time := 10 ns;

  signal send_data      : std_logic_vector(10 downto 0) := (others => '0');
  signal send_send      : std_logic := '0';
  signal send_info_bit  : std_logic := '0';

  signal send_data_send   : std_logic := '0';
  signal send_info_new    : std_logic := '0';

  signal send_slot_period_clks : std_logic_vector(7 downto 0) :=
    "00000110";

  signal send_serial      : std_logic;
  signal send_serial_sync : std_logic;
  signal send_idle_sync   : std_logic;

  signal send_count1      : integer := 0;
  signal send_count2      : integer := 0;
  signal send_count3      : integer := 0;

  signal fault_count1     : std_logic_vector(5 downto 0) := (others => '0');
  signal fault_count2     : std_logic_vector(5 downto 0) := (others => '0');
  signal fault_count3     : std_logic_vector(7 downto 0) := (others => '0');
  signal inject_fault     : std_logic := '0';

  signal fault_x          : std_logic_vector(31 downto 0) :=
    std_logic_vector(to_unsigned(seed2, 32));

  signal fault_shift_count : std_logic_vector(4 downto 0) := (others => '0');
  signal invert_signal    : std_logic := '0';

  signal recv_serial      : std_logic;

  signal recv_data        : std_logic_vector(10 downto 0) := (others => '0');
  signal recv_msg_strobe  : std_logic := '0';

  signal recv_info_bit    : std_logic := '0';
  signal recv_info_strobe : std_logic := '0';

  signal sync_status      : std_logic;
  signal sync_lost        : std_logic;
  signal bad_bit_count    : std_logic_vector(4 downto 0);

  signal clear_status    : std_logic := '0';

  signal test_mode       : std_logic := '0';

begin

  send_clk <= drift_clk;

  send_trig : entity work.rataser_trig_send
    generic map(
      period_bits => 8
      )
    port map(
      clk           => send_clk,

      data          => send_data,
      send          => send_send,

      info_bit      => send_info_bit,

      data_send     => send_data_send,
      info_new      => send_info_new,

      slot_period_clks => send_slot_period_clks,

      transmit      => send_serial,
      transmit_sync => send_serial_sync,
      idle_sync     => send_idle_sync
      );

  recv_serial <= send_serial;
  --recv_serial <= send_serial xor inject_fault xor invert_signal;

  recv_trig : entity work.rataser_trig_recv
    generic map(period_bits => 8)
    port map(
      clk           => recv_clk,

      receive       => recv_serial,

      data          => recv_data,
      msg_strobe    => recv_msg_strobe,

      info_bit      => recv_info_bit,
      info_strobe   => recv_info_strobe,

      sync_status   => sync_status,
      sync_lost     => sync_lost,
      bad_bit_count => bad_bit_count,
      clear_status  => clear_status
      );

  -- Update the signals to send
  process(send_clk)
  begin
    if (rising_edge(send_clk)) then
      if (send_serial_sync = '1') then
        send_data <= send_data + 1;
      end if;

      send_send <= '0';
      if (send_count1 = 659) then
        send_send <= '1';
        send_count1 <= 0;
      else
        send_count1 <= send_count1 + 1;
      end if;
      if (send_count2 = 809) then
        send_send <= '1';
        send_count2 <= 0;
      else
        send_count2 <= send_count2 + 1;
      end if;
      if (send_count3 = 941) then
        send_send <= '1';
        send_count3 <= 0;
      else
        send_count3 <= send_count3 + 1;
      end if;
    end if;
  end process;

  -- Update the info signal to send
  process(send_clk)
  begin
    if (rising_edge(send_clk)) then
      if (send_info_new = '1') then
        send_info_bit <= not send_info_bit;
      end if;
    end if;
  end process;

  -- Inject bit-faults
  process(send_clk)
  begin
    if (rising_edge(send_clk)) then
      inject_fault <= '0';

      fault_count1 <= fault_count1 - 1;
      fault_count2 <= fault_count2 - 1;
      fault_count3 <= fault_count3 - 1;
      
      if (fault_count1 = 0) then
        fault_count1 <= conv_std_logic_vector(53-1,fault_count1'length);
        inject_fault <= '1';
      end if;
      if (fault_count2 = 0) then
        fault_count2 <= conv_std_logic_vector(59-1,fault_count2'length);
        inject_fault <= '1';
      end if;
      if (fault_count3 = 0) then
        fault_count3 <= conv_std_logic_vector(251-1,fault_count3'length);
        inject_fault <= '1';
      end if;

      fault_x <= rataser_xorshift(fault_x, 5, 17, 13);

      fault_shift_count <= (others => '0');

      if (fault_x(15 downto 5) = 0) then
        fault_shift_count <= fault_x(4 downto 0);
      end if;

      if (fault_x(27 downto 16) = 0) then
        invert_signal <= not invert_signal;
      end if;

    end if;
  end process;

  -- Drive the clock
  process
  begin
    if runsim then
      recv_clk   <= '1';
      wait for 2.5 ns;
      recv_clk90 <= '1';
      wait for 2.5 ns;
      recv_clk   <= '0';
      wait for 2.5 ns;
      recv_clk90 <= '0';
      wait for 2.5 ns;
    else
      wait;
    end if;
  end process;

  -- Sender clock
  process
    variable x : std_logic_vector(31 downto 0);
    variable add1, add2 : time;
    --variable l : line;
  begin
    x := std_logic_vector(to_unsigned(seed, x'length));
       
    while runsim loop
      x := x xor (  x(31-13 downto 0) & "0000000000000");
      x := x xor ("00000000000000000" & x(31 downto 17));
      x := x xor (   x(31-5 downto 0) & "00000"        );

      add1 := conv_integer(x( 7 downto  4)) * 0.01 ns;
      add2 := conv_integer(x(15 downto 12)) * 0.01 ns;

      --write (l, String'("OUTPUT: "));
      --hwrite (l, x);
      --write (l, String'(" "));
      --write (l, add1);
      --write (l, String'(" "));
      --write (l, add2);
      --writeline (output, l);

      drift_clk   <= '1';
      wait for (4.925 ns + add1);
      drift_clk   <= '0';
      wait for (5 ns);
      --wait for (4.925 ns + add2);
    end loop;
    wait;
  end process;

  -- Slow clock
  process
    variable x : std_logic_vector(31 downto 0);
    variable add1, add2 : time;
    --variable l : line;
  begin
    x := std_logic_vector(to_unsigned(seed3, x'length));
       
    while runsim loop
      x := x xor (  x(31-13 downto 0) & "0000000000000");
      x := x xor ("00000000000000000" & x(31 downto 17));
      x := x xor (   x(31-5 downto 0) & "00000"        );

      add1 := conv_integer(x( 7 downto  4)) * 0.01 ns;
      add2 := conv_integer(x(15 downto 12)) * 0.01 ns;

      --write (l, String'("OUTPUT: "));
      --hwrite (l, x);
      --write (l, String'(" "));
      --write (l, add1);
      --write (l, String'(" "));
      --write (l, add2);
      --writeline (output, l);

      slow_clk          <= '0';
      slow_clk_jitter   <= '0';
      wait for (3*(4.925 ns + add1 - 0.5 ns));

      if (x(31) = '0') then
        slow_clk_jitter <= '1';
      end if;
      
      wait for (3*(0.5 ns));
      slow_clk          <= '1';
      wait for (3*(0.5 ns));

      if (x(31) = '1') then
        slow_clk_jitter <= '1';
      end if;

      wait for (3*(5 ns            - 0.5 ns));
      --wait for (4.925 ns + add2);
    end loop;
    wait;
  end process;

end architecture dataflow;
