-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;

entity rataser_4phase_sample is
  generic (max_skew   : string := "0.5 ns");
  port (clk           : in std_logic;
        clk90         : in std_logic;
        serial_in     : in std_logic;

        -- Sample at 0, 90, 180 and 270 degrees, cycles 2 and 3.
        saz           : out std_logic_vector(2 to 3);
        sbz           : out std_logic_vector(2 to 3);
        scz           : out std_logic_vector(2 to 3);
        sdz           : out std_logic_vector(2 to 3)
  );

  -- Attribute to limit the max skew between samples.  Since the
  -- samples are taken with different flip-flops, we need the signal
  -- to arrive with not more skew than a quarter of the clock cycle.
  attribute maxskew : string;
  attribute maxskew of serial_in : signal is max_skew;
  
end rataser_4phase_sample;

architecture RTL of rataser_4phase_sample is

  signal i_saz : std_logic_vector(0 to 3) := (others => '0');
  signal i_sbz : std_logic_vector(0 to 3) := (others => '0');
  signal i_scz : std_logic_vector(0 to 3) := (others => '0');
  signal i_sdz : std_logic_vector(0 to 3) := (others => '0');

begin

  -- Sampling is based on figure 4 of Xilinx XAPP224.

  -- Xz(0) is the first sample point, a-d are at progressively 90 degrees
  -- later phase of the clock cycle.
  --
  -- Xz(1,2,3) is then the pipelines version at the next clock cycle.  Note
  -- however that those which are not already at clk but a later phase,
  -- are gradually moved to an earlier phase, such that all are at clk
  -- for Xz(3)

  process(clk)
  begin
    if (rising_edge(clk)) then
      i_saz(0) <= serial_in;
      
      i_saz(1) <= i_saz(0);
      i_sbz(1) <= i_sbz(0);
      
      i_saz(2) <= i_saz(1);
      i_sbz(2) <= i_sbz(1);
      i_scz(2) <= i_scz(1);
      
      i_saz(3) <= i_saz(2);
      i_sbz(3) <= i_sbz(2);
      i_scz(3) <= i_scz(2);
      i_sdz(3) <= i_sdz(2);
    end if;
  end process;

  process(clk90)
  begin
    if (rising_edge(clk90)) then
      i_sbz(0) <= serial_in;

      i_scz(1) <= i_scz(0);

      i_sdz(2) <= i_sdz(1);
    end if;
  end process;

  process(clk)
  begin
    if (falling_edge(clk)) then
      i_scz(0) <= serial_in;

      i_sdz(1) <= i_sdz(0);
    end if;
  end process;

  process(clk90)
  begin
    if (falling_edge(clk90)) then
      i_sdz(0) <= serial_in;
    end if;
  end process;

  -- The deserialiser (bit-slot tracker) needs two samples of each
  -- phase.  From one cycle (here 2) and the previous (3).
  
  saz <= i_saz(2 to 3);
  sbz <= i_sbz(2 to 3);
  scz <= i_scz(2 to 3);
  sdz <= i_sdz(2 to 3);
  
end RTL;  
