-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;

use work.rataser_util.all;

entity rataser_4phase_deser is
  generic(pipeline_use1 : boolean := false;
          pipeline_use2 : boolean := false;
          pipeline_use3 : boolean := false;
          pipeline_use4 : boolean := false);
  port (clk           : in std_logic;

        saz           : in std_logic_vector(2 to 3);
        sbz           : in std_logic_vector(2 to 3);
        scz           : in std_logic_vector(2 to 3);
        sdz           : in std_logic_vector(2 to 3);

        bit_out       : out std_logic_vector(0 to 1) := (others => '0');
        got_bit       : out std_logic_vector(0 to 1) := (others => '0')
  );
end rataser_4phase_deser;
 
architecture RTL of rataser_4phase_deser is

  -- Suggestion to use a specific sample point (based on edge).
  signal new_a,  new_b,  new_c,  new_d  : std_logic  := '0';
  -- Pipelined
  signal new1_or : std_logic  := '0';
  signal new1_a, new1_b, new1_c, new1_d : std_logic  := '0';
  -- Previous and prev-previous suggestion.
  signal new2_a, new2_b, new2_c, new2_d : std_logic  := '0';
  signal new3_a, new3_b, new3_c, new3_d : std_logic  := '0';
  -- Do the last three suggestions agree?
  signal newX_a, newX_b, newX_c, newX_d : std_logic  := '0';

  -- Which sample point to use.
  signal use_a,  use_b,  use_c,  use_d : std_logic := '0';

  -- Pipelined (only for testing purposes!)
  signal use_a_p1, use_b_p1, use_c_p1, use_d_p1 : std_logic := '0';
  signal use_a_p2, use_b_p2, use_c_p2, use_d_p2 : std_logic := '0';
  signal use_a_p3, use_b_p3, use_c_p3, use_d_p3 : std_logic := '0';

  -- Sample point in use (after pipeline)
  signal use_a_r,  use_b_r,  use_c_r,  use_d_r  : std_logic := '0';
  -- Previous round used sample point.
  signal prev_a_r, prev_b_r, prev_c_r, prev_d_r : std_logic := '0';

begin

  process(clk)
  begin
    if (rising_edge(clk)) then
      -- Decide which sample to use.
      -- new1_X means to suggest to use X as sample point,
      -- so edge should be half a cycle earlier.

      -- 33332222
      -- abcdabcd
      --
      -- 0011s     -- s is the suggested sample point (also required 1)
      --  0011s
      --   0011s
      --    0011s

      -- By construction, these are mutually exclusive.
      -- (sdz(3) is part of all, and when it requires a 1, then
      -- scz(3) is unique, while if 0, then saz(2) is unique.)
      -- Getting none is also possible, if there was no acceptable
      -- transition.

      new_a <= (not saz(3)) and (not sbz(3)) and scz(3) and sdz(3) and saz(2);
      new_b <= (not sbz(3)) and (not scz(3)) and sdz(3) and saz(2) and sbz(2);
      new_c <= (not scz(3)) and (not sdz(3)) and saz(2) and sbz(2) and scz(2);
      new_d <= (not sdz(3)) and (not saz(2)) and sbz(2) and scz(2) and sdz(2);

      -- If there was a new suggestion on which transition to use, then
      -- update.  But only if the same phase was detected three times
      -- in a row.  This is to avoid false edges from accidentally
      -- inserting.

      -- How fast can it cycle transitions and thereby insert extra bits
      -- on its own?  An extra bit is only inserted if previous was A and
      -- next is D.  So it has to switch A-D.  But A-D-A will in total
      -- insert just 2, as D-A inserts no bits.  Thus something else
      -- has to be inbetween, like: A-D-B-A-D-B-A, here the possibly fake
      -- Bs would each cause the A-D transitions to inject an additional
      -- bit, we would get 2-1-1-2-1-1, instead of e.g 1-1-1-1-1-1,
      -- or 2-0-1-2-0-1, i.e. 8 instead of 6.

      -- The decision to shift the new1->new2->new3 below depends on
      -- an or, and with fan-out.  To pipeline that, precalculate the or.
      new1_or <= new_a or new_b or new_c or new_d;
      new1_a <= new_a;
      new1_b <= new_b;
      new1_c <= new_c;
      new1_d <= new_d;

      -- Shift the previous and next-previous suggestion if a new one arrived.
      if (new1_or = '1') then
        new2_a <= new1_a;
        new2_b <= new1_b;
        new2_c <= new1_c;
        new2_d <= new1_d;

        new3_a <= new2_a;
        new3_b <= new2_b;
        new3_c <= new2_c;
        new3_d <= new2_d;
      end if;

      -- Calculate if the three most recent suggestions agree.
      -- (Note, inputs are mutually exclusive, so only one can be active.)
      -- Done as latch on clock - pipelined.
      newX_a <= new1_a and new2_a and new3_a;
      newX_b <= new1_b and new2_b and new3_b;
      newX_c <= new1_c and new2_c and new3_c;
      newX_d <= new1_d and new2_d and new3_d;

      -- In next pipeline stage, copy the new decision if one was formed.
      if ((newX_a or newX_b or newX_c or newX_d) = '1') then
        use_a <= newX_a;
        use_b <= newX_b;
        use_c <= newX_c;
        use_d <= newX_d;
      end if;
    end if;
  end process;

  -- Pipelines of the use decision.  This is only used for testing,
  -- to give the possibility to have a 2-bit got_bit output at the
  -- end of a sym24 pattern.
  rataser_ppl(p=>pipeline_use1,c=>clk,i=>use_a,   o=>use_a_p1);
  rataser_ppl(p=>pipeline_use1,c=>clk,i=>use_b,   o=>use_b_p1);
  rataser_ppl(p=>pipeline_use1,c=>clk,i=>use_c,   o=>use_c_p1);
  rataser_ppl(p=>pipeline_use1,c=>clk,i=>use_d,   o=>use_d_p1);

  rataser_ppl(p=>pipeline_use2,c=>clk,i=>use_a_p1,o=>use_a_p2);
  rataser_ppl(p=>pipeline_use2,c=>clk,i=>use_b_p1,o=>use_b_p2);
  rataser_ppl(p=>pipeline_use2,c=>clk,i=>use_c_p1,o=>use_c_p2);
  rataser_ppl(p=>pipeline_use2,c=>clk,i=>use_d_p1,o=>use_d_p2);

  rataser_ppl(p=>pipeline_use3,c=>clk,i=>use_a_p2,o=>use_a_p3);
  rataser_ppl(p=>pipeline_use3,c=>clk,i=>use_b_p2,o=>use_b_p3);
  rataser_ppl(p=>pipeline_use3,c=>clk,i=>use_c_p2,o=>use_c_p3);
  rataser_ppl(p=>pipeline_use3,c=>clk,i=>use_d_p2,o=>use_d_p3);

  rataser_ppl(p=>pipeline_use4,c=>clk,i=>use_a_p3,o=>use_a_r);
  rataser_ppl(p=>pipeline_use4,c=>clk,i=>use_b_p3,o=>use_b_r);
  rataser_ppl(p=>pipeline_use4,c=>clk,i=>use_c_p3,o=>use_c_r);
  rataser_ppl(p=>pipeline_use4,c=>clk,i=>use_d_p3,o=>use_d_r);

  -- The decision in a latched block, since at high speed, even the
  -- simplest logic (that follows us), may be enough to bust timing.
  process(clk)
  begin
    if (rising_edge(clk)) then
      -- By the logic above, once running, always exactly one of use_X
      -- will be set.  (As well as one of prev_X).

      -- Remember previous cycle.
      prev_a_r <= use_a_r;
      prev_b_r <= use_b_r;
      prev_c_r <= use_c_r;
      prev_d_r <= use_d_r;

      -- Last (most often only) output bit is the one chosen.
      bit_out(1) <=
        (use_a_r and saz(3)) or
        (use_b_r and sbz(3)) or
        (use_c_r and scz(3)) or
        (use_d_r and sdz(3));

      -- If we went from D to A, no bit is valid.
      got_bit(1) <=
        ((not prev_d_r) and use_a_r) or use_b_r or use_c_r or use_d_r;

      -- If we went from A to D, two bits are valid.  sdz(3) is
      -- provided above, here we give the previous.
      bit_out(0) <= saz(3);
      got_bit(0) <= prev_a_r and use_d_r;
    end if;
  end process;
  
end RTL;
