-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.rataser_records.all;
use work.rataser_sym24word8_util;

entity rataser_clock_recv is
  generic(-- Number of bits for pulse period.
          period_bits: integer;
          -- Number of edges to report.
          num_last_edges: integer;
          -- Pipeline stages to handle diff calculation.
          pipeline_sym24_diff_in   : boolean := true;
          pipeline_sym24_diff_1    : boolean := false;
          pipeline_sym24_diff_2    : boolean := false;
          pipeline_sym24_diff_out  : boolean := true;
          pipeline_sym24_match_out : boolean := false;
          pipeline_sym24_pick      : boolean := false;
          pipeline_sym24_out       : boolean := false;
          -- Pipeline stages for hamming decoder.
          pipeline_hamming_parity  : boolean := true;
          pipeline_hamming_dual    : boolean := true;
          -- Limit the skew when doing 4phase sampling without primitive.
          max_skew_4phase_sample   : string := "0.5 ns");
  port (clk:           in std_logic;
        -- 90 degree phase shifted clock (only needed for eight_slot/4phase).
        clk90:         in std_logic;

        -- Input serial signal.
        receive:       in std_logic;
        -- Use 8 pulses to decode 1 raw bit (for PLL reception).
        eight_slot:    in std_logic;

        -- The two following only used for 4phase_sample / eight_slot:
        -- Where is the edge expected?
        expect_edge:   in std_logic_vector(1 downto 0);
        -- Sequence of edges seen most recently.  (For monitoring/setup.)
        -- Two bits for each edge, most recent in 1 downto 0.
        last_edges:    out std_logic_vector(num_last_edges*2-1 downto 0);
        -- Use routine to automatically lock onto edge.
        -- NOTE! This is *NOT* repeatable when signal is at clock edge!
        use_auto_edge: in std_logic;
        -- The edge currently determined by the edge tracker.
        auto_edge:     out std_logic_vector(1 downto 0);

        -- Reception delay (in ns, added to timestamp).
        receive_delay_ns: in std_logic_vector(15 downto 0) := (others => '0');

        -- Output timestamp, low 32 bits.
        tick_ns_lo:    out std_logic_vector(31 downto  0);
        -- High 32 bits of timestamp, latch one cycle later than low word.
        tick_ns_hi:    out std_logic_vector(63 downto 32);
        -- Aux signals.
        aux_sigs:      out std_logic_vector(4 downto 0);
        -- Auxiliary information.
        info_bit:      out std_logic;
        -- Strobe signal that we have decoded a new info bit message.
        msg_strobe:    out std_logic;

        -- Report current sync status (0 = pulse, 1 = 24-slot, 2 = time).
        sync_status:   out std_logic_vector(2 downto 0);
        -- Report if any sync status has been lost since clear.
        sync_lost:     out std_logic_vector(2 downto 0);
        -- Report different kinds of protocol issues since clear.
        bad_signals:   out std_logic_vector(4 downto 0);
        -- Clear sync loss errors.
        clear_status:  in std_logic;

        -- When operated behind a PLL, the pulse length (in clock cycles)
        -- must be given.
        pulse_period_clks: in std_logic_vector(period_bits-1 downto 0);
        -- When operated behind a PLL, the frequency must be given.
        clk_period_ns: in std_logic_vector(9 downto 0);

        -- For debugging: the received pulse period value.
        sender_period_ns: out std_logic_vector(9 downto 0);
        -- For debugging: the (non-PLL) bitstream tracking per-clock increment.
        track_incr_per_clock: out std_logic_vector(15 downto 0);

        -- For debugging: the filtered min pulse length (clock cyles).
        pulse_length_min: out std_logic_vector(9 downto 0);
        -- For debugging: the filtered max pulse length (clock cyles).
        pulse_length_max: out std_logic_vector(9 downto 0);
        -- For debugging: the filtered pulse length difference (clock cyles).
        pulse_length_diff6: out std_logic_vector(10 downto 0);

        sym8_debug:    out rataser_sym8word1_debug
        );
end;

architecture RTL of rataser_clock_recv is

  signal saz            : std_logic_vector(2 to 3);
  signal sbz            : std_logic_vector(2 to 3);
  signal scz            : std_logic_vector(2 to 3);
  signal sdz            : std_logic_vector(2 to 3);

  signal sabcdz         : std_logic_vector(0 to 7);

  signal bit_val        : std_logic_vector(0 to 1) := (others => '0');
  signal got_bit        : std_logic_vector(0 to 1) := (others => '0');
  signal missed_bit     : std_logic := '0';

  signal fp_bad_period  : std_logic := '0';
  signal fp_bad_8slot   : std_logic := '0';

  signal sp_bad_period  : std_logic := '0';
  signal sp_bad_8slot   : std_logic := '0';

  signal bad_period     : std_logic := '0';
  signal bad_8slot      : std_logic := '0';

  signal fp_bit_val        : std_logic_vector(0 to 1) := (others => '0');
  signal fp_got_bit        : std_logic_vector(0 to 1) := (others => '0');
  signal fp_missed_bit     : std_logic := '0';

  signal sp_bit_val        : std_logic_vector(0 to 1) := (others => '0');
  signal sp_got_bit        : std_logic_vector(0 to 1) := (others => '0');
  signal sp_missed_bit     : std_logic := '0';

  signal mark_w8a1      : std_logic := '0';
  signal w8p            : unsigned(7 downto 0) := (others => '0');
  signal w8n            : unsigned(7 downto 0) := (others => '0');
  signal aux1p          : std_logic := '0';
  signal aux1n          : std_logic := '0';

  signal bad_bit_sym24         : std_logic := '0';
  signal bad_bit_sym24_strobed : std_logic := '0';

  signal i_info_bit   : std_logic := '0';
  signal i_msg_strobe : std_logic := '0';

  -- For timing test.
  signal delay : std_logic_vector(0 to 7) := (others => '0');

  -- Pipeline for non-phase-locked receiver, to get the input into the
  -- clock domain, anti-metastable and one cycle extra.
  signal receive_delayed : std_logic_vector(2 downto 0) := (others => '0');

  -- For faking.
  signal prev_receive : std_logic_vector(0 to 3) := (others => '0');

  signal i_sender_period_ns : std_logic_vector(9 downto 0) := (others => '0');

  -- Reset the timetracking counter to exact 0 on next bit.
  signal reset_counter_on_bit : std_logic := '0';

  signal incr_per_clock : std_logic_vector(9 downto 0) := (others => '0');
  signal incr_per_clock_carry : std_logic := '0';

  signal clk_add_ns : std_logic_vector(9 downto 0) := (others => '0');
  signal clk_add_ns_carry : std_logic := '0';

  signal fp_clk_add_ns : std_logic_vector(9 downto 0) := (others => '0');
  signal fp_clk_add_ns_carry : std_logic := '0';

  signal sp_clk_add_ns : std_logic_vector(9 downto 0) := (others => '0');
  signal sp_clk_add_ns_carry : std_logic := '0';

  signal fp_last_edges : std_logic_vector(num_last_edges*2-1 downto 0);
  signal sp_last_edges : std_logic_vector(num_last_edges*2-1 downto 0);

  signal use_edge : std_logic_vector(1 downto 0) := (others => '0');

  signal fp_track_edge : std_logic_vector(1 downto 0) := (others => '0');
  signal sp_track_edge : std_logic_vector(1 downto 0) := (others => '0');

  signal track_edge : std_logic_vector(1 downto 0) := (others => '0');

  signal fp_pulse_length_min   : std_logic_vector(7 downto 0);
  signal fp_pulse_length_max   : std_logic_vector(7 downto 0);
  signal fp_pulse_length_diff6 : std_logic_vector(10 downto 0);

  signal sp_pulse_length_min : std_logic_vector(9 downto 0);
  signal sp_pulse_length_max : std_logic_vector(9 downto 0);

  signal bad_hamming     : std_logic := '0';
  signal mismatch_time   : std_logic := '0';

  signal time_wrap_midbit : std_logic := '0';

  -- Sync status from various parts.
  signal synced_pulse : std_logic := '0';
  signal synced_sym24 : std_logic := '0';
  signal synced_time  : std_logic := '0';

  signal fp_synced_pulse : std_logic := '0';
  signal sp_synced_pulse : std_logic := '0';

  signal i_sync_status : std_logic_vector(2 downto 0) := (others => '0');
  signal i_sync_lost   : std_logic_vector(2 downto 0) := (others => '0');

  signal i_bad_signals : std_logic_vector(4 downto 0) := (others => '0');
  
begin

  -------------------------------------------------------------------
  -- Phase-locked receiver clock.

  -- Incoming data is sampled four times per clock cycle.
  sh: if (true) generate
    sample_4phase: entity work.rataser_4phase_sample
      generic map(max_skew => max_skew_4phase_sample)
      port map(clk     => clk,
               clk90   => clk90,

               serial_in => receive,

               saz     => saz,
               sbz     => sbz,
               scz     => scz,
               sdz     => sdz);
  end generate;
  -- For timing test:
  st: if (false) generate
    process(clk)
    begin
      if (rising_edge(clk)) then
        delay <= delay(1 to 7) & receive;
        saz <= delay(0 to 1);
        sbz <= delay(2 to 3);
        scz <= delay(4 to 5);
        sdz <= delay(6 to 7);
      end if;
    end process;
  end generate;

  sabcdz <=
    saz(3) & sbz(3) & scz(3) & sdz(3) &
    saz(2) & sbz(2) & scz(2) & sdz(2);

  -- Each bit has been encoded as symbols of 8 slots.
  -- Figure out both the start of each symbol, and evaluate the bits.
  decode_8bit: entity work.rataser_sym8word1_decode
    generic map(period_bits => period_bits)
    port map(clk         => clk,

             raw_val     => sabcdz,

             pulse_period_clks => pulse_period_clks,

             expect_edge => use_edge,
             last_edges  => fp_last_edges,

             bit_val     => fp_bit_val,
             got_bit     => fp_got_bit,
             missed_bit  => fp_missed_bit,

             synced      => fp_synced_pulse,

             bad_period  => fp_bad_period,
             bad_8slot   => fp_bad_8slot,

             pulse_length_min => fp_pulse_length_min,
             pulse_length_max => fp_pulse_length_max,

             pulse_length_diff6 => fp_pulse_length_diff6,

             debug   => sym8_debug
             );

  fp_clk_add_ns       <= clk_period_ns;
  fp_clk_add_ns_carry <= '0';

  edge_track: entity work.rataser_track_edge
    port map(clk          => clk,

             last_edge    => fp_last_edges(1 downto 0),

             synced_pulse => fp_synced_pulse,
             synced_sym24 => synced_sym24,
             synced_time  => synced_time,

             update_pulse => time_wrap_midbit,

             track_edge   => fp_track_edge);

  use_edge <= expect_edge when (use_auto_edge = '0') else fp_track_edge;

  -------------------------------------------------------------------
  -- Non-phase-locked receiver.

  -- Get value into clock domain.
  process(clk)
  begin
    if (rising_edge(clk)) then
      receive_delayed <= receive_delayed(1 downto 0) & receive;
    end if;
  end process;

  -- Track the frequency of the sender, and evaluate the bits.
  track_stream: entity work.rataser_track_bitstream
    generic map(frac_bits => 8)
    port map(clk            => clk,
             receive        => receive_delayed(2),
             synced_period  => sp_synced_pulse,
             pulse_period   => i_sender_period_ns,
             pattern_synced => '0',
             clear_failure  => '0',
             bit_val        => sp_bit_val(1),
             got_bit        => sp_got_bit(1),
             reset_counter_on_bit => reset_counter_on_bit,
             incr_per_clock       => incr_per_clock,
             incr_per_clock_carry => incr_per_clock_carry,
             full_incr_per_clock  => track_incr_per_clock,
             pulse_length_min     => sp_pulse_length_min,
             pulse_length_max     => sp_pulse_length_max
             );

  sp_missed_bit <= '0'; -- Not used here.
  sp_track_edge <= (others => '0'); -- Not used here.

  sp_bad_period <= '0'; -- Not used here.
  sp_bad_8slot <= '0'; -- Not used here.

  sp_last_edges(sp_last_edges'range) <= (others => '0'); -- Not used here.

  sp_clk_add_ns       <= incr_per_clock;
  sp_clk_add_ns_carry <= incr_per_clock_carry;

  -------------------------------------------------------------------

  got_bit       <= fp_got_bit    when (eight_slot = '1') else sp_got_bit;
  bit_val       <= fp_bit_val    when (eight_slot = '1') else sp_bit_val;
  missed_bit    <= fp_missed_bit when (eight_slot = '1') else sp_missed_bit;

  bad_period    <= fp_bad_period when (eight_slot = '1') else sp_bad_period;
  bad_8slot     <= fp_bad_8slot  when (eight_slot = '1') else sp_bad_8slot;

  clk_add_ns    <= fp_clk_add_ns when (eight_slot = '1') else sp_clk_add_ns;
  clk_add_ns_carry
    <= fp_clk_add_ns_carry when (eight_slot = '1') else sp_clk_add_ns_carry;

  synced_pulse  <= fp_synced_pulse when (eight_slot = '1') else sp_synced_pulse;

  last_edges    <= fp_last_edges when (eight_slot = '1') else sp_last_edges;

  track_edge    <= fp_track_edge when (eight_slot = '1') else sp_track_edge;

  pulse_length_min <=
    "00"& fp_pulse_length_min when (eight_slot = '1') else sp_pulse_length_min;
  pulse_length_max <=
    "00"& fp_pulse_length_max when (eight_slot = '1') else sp_pulse_length_max;
  pulse_length_diff6 <=
    fp_pulse_length_diff6 when (eight_slot = '1') else (others => '0');

  -------------------------------------------------------------------

  ft: if (false) generate
    -- Faking reception from period 3, short 1 and long 2
    process(clk)
    begin
      if (rising_edge(clk)) then
        -- 0 oldest
        prev_receive <= prev_receive(1 to prev_receive'high) & receive;

        got_bit(1) <= '0';
      
        if (prev_receive(0) = '0' and
            prev_receive(1) = '1') then
          got_bit(1) <= '1';
          bit_val(1) <= prev_receive(2);
        end if;
      end if;
    end process;
  end generate;


  -- If we would not use the deserialiser, and run on the same clock
  -- as the sender:
  -- got_bit <= '0' & '1';
  -- bit_val <= '0' & receive;

  decode_sym: entity work.rataser_sym24word8_decode
    generic map(pipeline_diff_in   => pipeline_sym24_diff_in,
                pipeline_diff_1    => pipeline_sym24_diff_1,
                pipeline_diff_2    => pipeline_sym24_diff_2,
                pipeline_diff_out  => pipeline_sym24_diff_out,
                pipeline_match_out => pipeline_sym24_match_out,
                pipeline_pick      => pipeline_sym24_pick,
                pipeline_out       => pipeline_sym24_out)
    port map(clk => clk,

             bit_val => bit_val,
             got_bit => got_bit,
             missed_bit => missed_bit,

             mark_w8a1      => mark_w8a1,
             mark_w8a1_2bit => open,
             out_w8p        => w8p,
             out_w8n        => w8n,
             out_aux1p      => aux1p,
             out_aux1n      => aux1n,

             mark_w1        => open,
             mark_w1_2bit   => open,

             out_w1p        => open,
             out_w1n        => open,

             mark_w16       => open,
             out_w16p       => open,
             out_w16n       => open,

             synced         => synced_sym24,
             bad_bit        => bad_bit_sym24
             );

  -- It would be enough to pipeline the Hamming decoder when we are
  -- running phase-locked.  This is the case when we likely have
  -- higher frequency requirements.
  decode_ts: entity work.rataser_timestamp_decode
    generic map(pipeline_hamming_parity => pipeline_hamming_parity,
                pipeline_hamming_dual   => pipeline_hamming_dual)
    port map(clk => clk,

             mark_w8a1   => mark_w8a1,
             in_w8p      => w8p,
             in_w8n      => w8n,
             in_aux1p    => aux1p,
             in_aux1n    => aux1n,

             receive_delay_ns => receive_delay_ns,

             tick_ns_lo  => tick_ns_lo,
             tick_ns_hi  => tick_ns_hi,
             aux_sigs    => aux_sigs,
             info_bit    => i_info_bit,
             msg_strobe  => i_msg_strobe,

             clk_add_ns               => clk_add_ns,
             clk_add_ns_carry         => clk_add_ns_carry,

             reset_add_counter_on_bit => reset_counter_on_bit,
             got_bit     => got_bit(1),

             wrap_midbit => time_wrap_midbit,

             synced      => synced_time,

             bad_hamming => bad_hamming,
             mismatch_time => mismatch_time,

             sender_period_ns => i_sender_period_ns
             );

  -- Copy of sync status info.
  i_sync_status(0) <= synced_pulse;
  i_sync_status(1) <= synced_sym24;
  i_sync_status(2) <= synced_time;

  bad_bit_sym24_strobed <= bad_bit_sym24 and mark_w8a1;

  process(clk)
  begin
    if (rising_edge(clk)) then
      -- If a sync bit is missing, we have lost it.
      i_sync_lost <= i_sync_lost or (not i_sync_status);

      i_bad_signals(0) <= i_bad_signals(0) or bad_period;
      i_bad_signals(1) <= i_bad_signals(1) or bad_8slot;
      i_bad_signals(2) <= i_bad_signals(2) or bad_bit_sym24_strobed;
      i_bad_signals(3) <= i_bad_signals(3) or bad_hamming;
      i_bad_signals(4) <= i_bad_signals(4) or mismatch_time;

      if (clear_status = '1') then
        -- We do not bother to keep whatever was reported this cycle.
        i_sync_lost <= (others => '0');
        i_bad_signals <= (others => '0');
      end if;
    end if;
  end process;

  -- signals    <= i_signals;
  info_bit    <= i_info_bit;
  msg_strobe  <= i_msg_strobe;

  sync_status <= i_sync_status;
  sync_lost   <= i_sync_lost;
  bad_signals <= i_bad_signals;

  auto_edge   <= track_edge;

  sender_period_ns <= i_sender_period_ns;

end RTL;
