-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.rataser_records.all;
use work.rataser_sym24word8_util.all;

entity rataser_clock_send is
  generic(-- Number of bits for pulse period.
          period_bits:   integer
          );
    port (clk:           in std_logic;

          -- Absolute time (the timestamp that is sent).
          tick_ns:       in std_logic_vector(63 downto 0);
          -- Auxiliary signals.
          aux_sigs:      in std_logic_vector(4 downto 0);
          -- Auxiliary information.
          info_bit:      in std_logic;

          -- Pulse period (in clk cycles, controls sender).
          pulse_period_clks:  in std_logic_vector(period_bits-1 downto 0);
          -- Minimum duty cycle of pulse pause (remaining length after long).
          duty_low_min_clks:  in std_logic_vector(period_bits-1 downto 0);
          -- Maximum duty cycle of pulse pause (remaining length after short).
          duty_low_max_clks:  in std_logic_vector(period_bits-1 downto 0);
          -- Use 8 pulses to encode 1 raw bit (for PLL reception).
          eight_slot:    in std_logic;

          -- Pulse period (in ns, encoded in message).
          pulse_period_ns: in std_logic_vector(9 downto 0) := (others => '0');
          -- Use the given pulse period.  Otherwise, calculate from time diff.
          use_pulse_period_ns: in std_logic := '0';

          -- Message delay (in ns, added to time in message).
          message_delay_ns: in std_logic_vector(21 downto 0) := (others =>'0');

          -- Output message.
          transmit:      out std_logic := '0';
          -- Oscilloscope sync signal for message.
          transmit_sync: out std_logic := '0';

          -- Trailing cycle after short pulse.
          trail_short:   out std_logic := '0';
          -- Trailing cycle after long pulse.
          trail_long:    out std_logic := '0'
          );
end;

-- Transmission is done in (16+4+x)*(2*24)*(m=8/1)*n clk cycles.
-- n = pulse_period_clks
-- m = eight_slot ? 8 : 1
-- x >= 1, to allow at least one cycle of slow data
-- Each n clk cycles, one pulse is emitted.
-- Leading edge is exactly periodic.
-- Bit data is encoded in the length (i.e. trailing edge jitters).
-- n gives the divider of the source clock.
-- If e.g. the source clock is 125 MHz, and the emitted clock signal
-- is to have a 25 MHz frequency, then n = 5.
-- m is either 1 (plain encoding) or 8 (01qQqQqQ) encoding of data.
-- Basically: use m=8 if recovering the clock with a PLL, usually
-- means that the emitted clock frequency itself is > 10-20 MHz.
-- PLLs do not like large duty cycle differences, so to be more
-- significant while keeping the duty cycle differences small,
-- each payload bit is here sent multiple times.
-- If just recovering the timestamps (without PLL), then the duty cycle
-- difference between short and long can be large anyhow, so no need to
-- stretch with m.  Reason to reduce to m = 1 is such that the aux
-- signals that are embedded should not be updated unnecessarily seldom.
-- In particular since in the no-PLL mode, we also have much larger n,
-- since we often run over worse cables.  I.e. two different use-cases,
-- that due to different functional constraints cannot use the same
-- signal anyhow.

-- Transmission is started whenever the timestamp value has
-- wrapped the next suitable power-of-two.  This is done such that
-- we know that the low part of the timestamp has mostly zeros.
-- That way, we can add an offset correction to the transmitted timestamp
-- such that the delay from us starting the transmission until the receiver
-- having gotten it does not cause a big time offset.  (This is for
-- convenience - experience is that it is helpful for DAQ operations.)

-- Also, the pattern sent is intentionally started with a fixed value
-- (0 or 1), such that the timestamp latching can occur after the wrap
-- while the pattern that is going to be sent also takes a clk cycle to
-- prepare.

-- When m = 8, then n will usually be small.  Put another way, PLLs usually
-- do not want to go below 10 MHz, i.e. 100 ns.  With m = 8 -> 800 ns.
-- (16+4+1)*(2*24) * 800 ns = 806400 ns = 0.8 ms.

-- So for maximum cycle period, we only need to consider the case m = 1
-- and e.g. n = 255.  This would for a 100 MHz sender (10 ns) mean 2.55 us
-- pulse period.

-- (16+4+1)*(2*24)*1*255 * 10 ns = 2570400 = 2.57 ms

-- Assume the transmitted timestamp to be encoded in 1 ns steps.
-- That means that the minimum message period is 2570400 < 4194304 = 2^22.
-- Since we also want to be able to send along the proper delay this
-- is causing *without* the need to carry into the actual timestamp
-- bits when the actual send offset is added, we allow offsets up to
-- 2^22-1 to be transmitted, and use bit 23 as carry.





architecture RTL of rataser_clock_send is

  -- A new timestamp is sent (after) this bit changes.
  constant tick_wrap_loc : integer := 23;

  -- We only need to latch the low part.
  -- The high part is stable during a transmission cycle.
  -- Since each cycle shall contain at least 21 groups, 4 less bits
  -- are needed, since we start at most one group after the wrap.
  signal tick_ns_latch  : unsigned(23-4-1 downto 0) := (others => '0');
  signal tick_ns_offset : unsigned(23  -1 downto 0) := (others => '0');
  -- Latch of the aux bits at the start of the group.
  signal latch_aux_sigs : std_logic_vector(4 downto 0) := (others => '0');
  signal latch_info_bit : std_logic := '0';

  -- Pulse period (in ns, encoded in message).
  -- Calculated from actual difference of timestamp at pulses.
  signal calc_pulse_period_ns : unsigned(pulse_period_ns'range) :=
    (others => '0');
  -- Timestamp at previous pulse.
  signal tick_ns_prev_pulse   : unsigned(pulse_period_ns'range) :=
    (others => '0');
  -- The pulse period to encode in the message.
  signal send_pulse_period_ns : std_logic_vector(pulse_period_ns'range) :=
    (others => '0');

  -- Data to send during filler groups.
  signal latch_slow : std_logic_vector(3 downto 0) := (others => '0');

  -- Full data buffer:
  signal data_send       : std_logic_vector(63 downto 0) := (others => '0');
  signal extra_data_send : std_logic_vector(15 downto 0) := (others => '0');

  -- Split into groups of 4 bits:
  type group_vector16 is array (0 to 15) of std_logic_vector(3 downto 0);
  type group_vector4  is array (0 to  3) of std_logic_vector(3 downto 0);

  signal data_send_groups      : group_vector16 := (others => (others => '0'));
  signal extra_data_send_groups : group_vector4 := (others => (others => '0'));

  -- The group of data to (if chosen) send this cycle:
  signal data_send_grp       : std_logic_vector(3 downto 0) := (others => '0');
  signal extra_data_send_grp : std_logic_vector(3 downto 0) := (others => '0');

  -- Within each hamming block.
  signal hamm_w11_ts_data   : std_logic_vector(3 downto 0) := (others => '0');
  signal hamm_w11_grp_mark  : std_logic_vector(1 downto 0) := (others => '0');
  signal hamm_w11_aux_sigs  : std_logic_vector(4 downto 0) := (others => '0');

  signal hamm_word11   : std_logic_vector(10 downto 0) := (others => '0');
  signal hamm_data16   : std_logic_vector(15 downto 0) := (others => '0');

  -- Latched hamming input.
  signal hamm_word11_r : std_logic_vector(10 downto 0) := (others => '0');
  -- Latched hamming output.
  signal hamm_data16_r : std_logic_vector(15 downto 0) := (others => '0');

  -- Input data payload to 24-slot sequence.
  signal sym24_data8   : std_logic_vector(7 downto 0) := (others => '0');
  signal sym24_data8_u : unsigned(7 downto 0) := (others => '0');
  -- Sequence of 24-slot pattern.
  signal sym24       : std_logic_vector(0 to 23) := (others => '0');
  -- Sequence of 24-slot pattern in 32 slots (slots % 4 == 3 unused).
  signal sym32       : std_logic_vector(0 to 31) := (others => '0');

  -- Bit to transmit, based on current count value.
  signal sym_output  : std_logic := '0';

  -- qqq0 data
  signal slot8_tmp : std_logic := '0';
  -- 10QqQqQq data
  signal slot8_out : std_logic := '0';
  -- 0001 - for debugging (to know when fixed bits appear).
  signal slot8_fix : std_logic := '0';

  -- Pipeline stage.
  signal slot8_out_r : std_logic := '0';

  signal count_clk_lt_duty_low_min : std_logic := '0';
  signal count_clk_lt_duty_low_max : std_logic := '0';

  -- Value (raw bit) to transmit (internal).
  signal i_transmit    : std_logic := '0';
  -- Scope-watching sync signal for transmit sequence;
  signal i_transmit_sync : std_logic := '0';

  -- Did we just end a short pulse (internal).
  signal i_trail_short : std_logic := '0';
  -- Did we just end a long pulse (internal).
  signal i_trail_long  : std_logic := '0';

  -- Normalise inputs to downto ranges.
  signal downto_pulse_n      : std_logic_vector(period_bits-1 downto 0);
  signal downto_duty_low_min : std_logic_vector(period_bits-1 downto 0);
  signal downto_duty_low_max : std_logic_vector(period_bits-1 downto 0);

  -- Counter of clock cycles.
  signal count_clk      : unsigned(period_bits-1 downto 0) := (others => '0');
  signal count_clk_next : unsigned(period_bits   downto 0) := (others => '0');
  signal count_clk_wrap : std_logic := '0';

  signal count_clk_max  : unsigned(period_bits-1 downto 0) := (others => '0');

  -- Counter for the 10QqQqQq pattern.
  signal count_slot8      : unsigned(2 downto 0) := (others => '0');
  signal count_slot8_next : unsigned(3 downto 0) := (others => '0');
  signal count_slot8_wrap : std_logic := '0';

  -- Counter for the 24-slot symbol (dual).
  signal count_sym_add    : unsigned(1 downto 0) := "00";
  signal count_sym        : unsigned(5 downto 0) := (others => '0');
  signal count_sym_next   : unsigned(6 downto 0) := (others => '0');
  signal count_sym_wrap   : std_logic := '0';

  -- Counter for the signal groups.
  signal count_grp      : unsigned(4 downto 0) := (others => '0');
  signal count_grp_next : unsigned(4 downto 0) := (others => '0');
  signal count_grp_wrap : std_logic := '0';

  -- What group are we emitting.
  signal grp_mark : std_logic_vector(1 downto 0) := "11"; -- filler

  -- Previous value of the wrapping location.
  signal timestamp_prev_wrap_loc : std_logic := '0';
  -- Did the timestamp wrap?
  signal timestamp_wrapped : std_logic := '0';

begin

  -- Aliases.
  downto_pulse_n      <= pulse_period_clks;
  downto_duty_low_min <= duty_low_min_clks;
  downto_duty_low_max <= duty_low_max_clks;

  -- Maximum value (start) of pulse counter is pulse period - 1.
  count_clk_max <= unsigned(downto_pulse_n) - 1;
  
  -- Counter for the pulse itself (down-count)
  count_clk_next <= ('0' & count_clk) - 1;
  count_clk_wrap <= count_clk_next(count_clk_next'high);

  -- Counter for the 10QqQqQq pattern (down-count)
  count_slot8_next <= ('0' & count_slot8) - ("00" & count_clk_wrap);
  count_slot8_wrap <= count_slot8_next(count_slot8_next'high);

  -- Count the bit-slots.  (up-count)
  -- 0, 1, 2,  4, 5, 6,  8, 9, 10,  12, ...
  -- Add 1 when 0 or 1, add 2 when 2.  (adding 1 when 3)
  count_sym_add(1) <=
    count_slot8_wrap when (count_sym(1)  = '1') else '0';
  count_sym_add(0) <=
    count_slot8_wrap when (count_sym(1) /= '1') else '0';
  count_sym_next <= ('0' & count_sym) + count_sym_add;
  count_sym_wrap <= count_sym_next(count_sym_next'high);

  -- Count the group number
  count_grp_next <= count_grp + ("00" & count_sym_wrap);
  count_grp_wrap <= '1' when (count_grp =
                              to_unsigned(19, count_grp'length)) else '0';

  -- Data generation
  data_send(22 downto  0) <= std_logic_vector(tick_ns_offset);
  data_send(63 downto 23) <= tick_ns(63 downto 23);

  -- Pick which pulse period to send.
  send_pulse_period_ns <=
    pulse_period_ns when (use_pulse_period_ns = '1') else
    std_logic_vector(calc_pulse_period_ns);

  -- Extra data (info bit etc), sent in final groups.
  extra_data_send <= send_pulse_period_ns & "00" & "000" & latch_info_bit;

  -- Assign the group members:
  edsg: for i in 0 to 3 generate
    extra_data_send_groups(i) <= extra_data_send(i * 4 + 3 downto i * 4);
  end generate;
  dsg: for i in 0 to 15 generate
    data_send_groups(i)       <=       data_send(i * 4 + 3 downto i * 4);
  end generate;

  -- Make warning-happy compiler silent.  (Should be assigned elsewhere with
  -- some data).
  latch_slow <= (others => '0');

  -- Pick the 4 bits of data to send from either extra or data.
  extra_data_send_grp <=
    extra_data_send_groups(to_integer(count_grp(1 downto 0)));
  data_send_grp <=
    data_send_groups(to_integer(count_grp(3 downto 0)));

  -- For ease-of-read, aliases.
  hamm_w11_ts_data  <=
    latch_slow when (grp_mark = "11") else
    extra_data_send_grp when (count_grp(4) = '1') else data_send_grp;
  hamm_w11_grp_mark <= grp_mark;
  hamm_w11_aux_sigs <= latch_aux_sigs;
  
  hamm_word11( 3 downto 0) <= hamm_w11_ts_data;
  hamm_word11( 5 downto 4) <= hamm_w11_grp_mark;
  hamm_word11(10 downto 6) <= hamm_w11_aux_sigs;
  
  -- Do the Hamming encoding
  hamm: entity work.rataser_hamming_encode
    port map(data   => hamm_word11_r,
             output => hamm_data16);

  -- First or second 24-slot symbol.
  sym24_data8 <=
    hamm_data16_r(15 downto 8) when (count_sym(5) = '0') else
    hamm_data16_r( 7 downto 0);

  sym24_data8_u <= unsigned(sym24_data8);

  gen_sequence(seq  => sym24,
               data => sym24_data8_u,
               aux  => count_sym(5));

  gen_sequence32(seq32 => sym32,
                 seq24 => sym24);

  sym_output <= sym32(to_integer(count_sym(4 downto 0)));

  -- The 8-slot output is moderated with inversion due to the lowest bit,
  -- and otherwise 0qqq (inversion gives 1QQQ).
  -- Since count_slot8 count downwards, we get 10QqQqQq.
  -- If eight_slot is not enabled, count_slot8 = 0 and we always emit q.
  slot8_tmp <= '0' when (count_slot8(2 downto 1) = "11") else sym_output;
  slot8_out <= slot8_tmp xor count_slot8(0);
  -- For debugging
  slot8_fix <= '1' when (count_slot8(2 downto 1) = "11") else '0';

  process(clk)
  begin
    if (rising_edge(clk)) then

      -- Pipeline stage before and after Hamming encoding.
      -- This is no problem since the first bit of the 24-slot pattern
      -- is fixed.
      hamm_word11_r <= hamm_word11;
      hamm_data16_r <= hamm_data16;

      -- Pipeline stage after symbol bit multiplexer.
      -- This is safe, since each emitted period always has at least
      -- one first clock cycle with a fixed 1.
      slot8_out_r <= slot8_out;

      -- Counter for pulse itself.
      count_clk <= count_clk_next(count_clk'range);
      if (count_clk_wrap = '1') then
        count_clk <= count_clk_max;
        -- Calculate timestamp difference.
        calc_pulse_period_ns <=
          unsigned(tick_ns(tick_ns_prev_pulse'range)) - tick_ns_prev_pulse;
        -- Remember the timestamp at the previous pulse.
        tick_ns_prev_pulse <= unsigned(tick_ns(tick_ns_prev_pulse'range));
      end if;

      -- Counter for 8-slot expansion.
      count_slot8 <= count_slot8_next(count_slot8'range);
      -- If we are not doing 01qQqQqQ, then we always set the
      -- slot8 counter to 0.  Natural wrapping take care of other reset.
      if (eight_slot = '0') then
        count_slot8 <= (others => '0');
      end if;

      -- Counter for the 24-slot symbol.
      count_sym <= count_sym_next(count_sym'range);
      -- The symbol counter wraps naturally at 64, so need no reset.

      -- Counter for the group.
      count_grp <= count_grp_next(count_grp'range);

      -- Did the timestamp wrap?
      if (tick_ns(tick_wrap_loc) /= timestamp_prev_wrap_loc) then
        timestamp_wrapped <= '1';
      end if;
      timestamp_prev_wrap_loc <= tick_ns(tick_wrap_loc);

      -- Default: no sync marker pulse.
      i_transmit_sync <= '0';

      -- What kind of group to emit cycles through the valid words.
      -- We update when suitable
      case (grp_mark) is
        when "00" =>
          -- First group ends with symbol of group.
          if (count_sym_wrap = '1') then
            grp_mark <= "01";
          end if;

        when "01" =>
          -- Timestamp data ends with full group.
          if (count_grp_wrap = '1') then
            grp_mark <= "10";
          end if;

        when "10" =>
          -- Last group ends with symbol of group.
          if (count_sym_wrap = '1') then
            grp_mark <= "11";
          end if;

        when "11" =>
          -- Filler symbol ends with symbol together with timestamp wrap.
          if (count_sym_wrap = '1' and
              timestamp_wrapped = '1') then
            -- Latch low timestamp bits
            tick_ns_latch <= unsigned(tick_ns(tick_ns_latch'range));
            -- Group counter is reset.
            count_grp <= (others => '0');
            grp_mark <= "00";
            -- The timestamp wrap has been used.
            timestamp_wrapped <= '0';
            -- Emit sync marker pulse.
            i_transmit_sync <= '1';
          end if;

        when others => -- Make compiler happy.

      end case;

      if (count_sym_wrap = '1') then
        -- Latch aux signals on each new group.
        latch_aux_sigs <= aux_sigs;
        -- Also the info bit.
        latch_info_bit <= info_bit;
      end if;

      count_clk_lt_duty_low_min <= '0';
      count_clk_lt_duty_low_max <= '0';
      -- Calculate the comparisons for the next cycle, to help timing.
      if (count_clk < unsigned(downto_duty_low_min)) then
        count_clk_lt_duty_low_min <= '1';
      end if;
      if (count_clk < unsigned(downto_duty_low_max)) then
        count_clk_lt_duty_low_max <= '1';
      end if;

      i_trail_short <= '0';
      i_trail_long  <= '0';

      if (count_clk_lt_duty_low_min = '1') then
        i_transmit <= '0';
        i_trail_long <= i_transmit;
      elsif (count_clk_lt_duty_low_max = '1') then
        i_transmit <= slot8_out_r;
        i_trail_short <= i_transmit and not slot8_out_r;
      else
        i_transmit <= '1';
      end if;

      -- Calculate the transmission time at the receiver.
      -- The result has 23 bits, so can handle a carry from
      -- message_delay_ns which only has 22 bits.
      tick_ns_offset <=
        unsigned('0' & message_delay_ns) + tick_ns_latch;
        
    end if;
  end process;

  -- Assign the actual output signals
  transmit      <= i_transmit;
  transmit_sync <= i_transmit_sync;

  trail_short   <= i_trail_short;
  trail_long    <= i_trail_long;

end RTL;
