-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.rataser_hamming_util.all;
use work.rataser_util.all;

entity rataser_hamming_decode is
  generic (pipeline_parity : boolean := false);
  port (clk     : in  std_logic := '0'; -- Only needed for pipeline.
        message : in  std_logic_vector;
        data    : out std_logic_vector;
        good    : out std_logic;
        single_error :  out std_logic;
        double_error :  out std_logic
        );
end rataser_hamming_decode;

-- Generate data from a Hamming coded message.

architecture RTL of rataser_hamming_decode is

  constant data_size   : positive := data'high + 1 - data'low;

  constant parity_bits : positive := rataser_hamming_parity_bits(data_size);
  constant msg_size    : positive := rataser_hamming_msg_size(data_size);

  -- Parity bits (in error)
  signal syndrome   : std_logic_vector(parity_bits downto 1) := (others =>'0');
  -- Pipeline
  signal syndrome_r : std_logic_vector(parity_bits downto 1) := (others =>'0');
  signal message_r  : std_logic_vector(message'range) := (others => '0');
  -- Parity bit 0 (in error)
  signal p0_bad           : std_logic := '0';

  -- The bit to be repaired (if any)
  signal repair     : std_logic_vector(2**parity_bits-1 downto 0) :=
    (others => '0');
  -- The repaired message
  signal repaired_message : std_logic_vector(msg_size-1 downto 0) :=
    (others => '0');

  -- Did we have single error?
  signal single_err       : std_logic := '0';
  -- Did we have double error?
  signal double_err       : std_logic := '0';
begin

  --  0  1  2  3  4  5  6  7  8  9 10 11 12 13  14  15
  -- p0 p1 p2 d1 p4 d2 d3 d4 p8 d5 d6 d7 d8 d9 d10 d11
  --  P  P  P  -  P  -  -  -  P  -  -  -  -  -   -   -
  --  0  1  2     3           4                        (parity bit)
  -- -1  0  1     2           3                        (i)
 
  process(message)
    variable located_at  : positive;
    variable tmp_partity : std_logic;
  begin
    get_parity_i: for i in 1 to parity_bits loop
      located_at := 2**(i-1);  -- 1, 2, 4, 8, ...
      tmp_partity := '0';
      -- Find the bits that are part of the parity
      get_parity_j: for j in 0 to msg_size-1 loop
        if ((std_logic_vector(to_unsigned(j, parity_bits)) and
             std_logic_vector(to_unsigned(located_at, parity_bits))) /=
            std_logic_vector(to_unsigned(0, parity_bits))) then
          tmp_partity := tmp_partity xor message(j);
        end if;
      end loop;
      syndrome(i) <= tmp_partity;
    end loop;
  end process;

  rataser_ppl(p=>pipeline_parity,c=>clk,i=>syndrome,o=>syndrome_r);
  rataser_ppl(p=>pipeline_parity,c=>clk,i=>message ,o=>message_r );

  -- Now, if any parity bits are non-zero, then they suggest a location
  -- to correct, i.e. to flip.  Basically, we can calculate which bit
  -- bit location needs flipping, and then xor with that.

  process(syndrome_r)
  begin
    repair <= (others => '0');
    -- This is the expensive statement in this entity.
    repair(to_integer(unsigned(syndrome_r(parity_bits downto 1)))) <= '1';
  end process;

  repaired_message <= message_r xor repair(msg_size-1 downto 0);

  p0: entity work.rataser_big_xor
    port map(inputs => message_r,
             output => p0_bad);
  
  -- It might now have repaired the parity bits.  Which does not matter,
  -- since we will not use those.  Then we also have the following rules
  -- to consider:
  --
  -- no (all-0) syndrome + p0 good  => message good,
  -- syndrome            + p0 bad   => single bit error,
  -- syndrome            + p0 good  => double bit error,
  -- no (all-0) syndrome + p0 bad   => single bit error (in p0).
  --
  -- Note: no (all-0) syndrome <=> p0        repaired (repair(0) = '1'),
  --       non-0 syndrome      <=> pX [X!=0] repaired (repair(0) = '0').

  -- So did we have any errors?
  single_err <= p0_bad;
  double_err <= (not p0_bad) and (not repair(0));

  -- Outputs:
  good <= not double_err;          -- Single-bit errors handled, so ok.
  single_error <= single_err;
  double_error <= double_err;

  -- Extract the data bits, after repair.

  process(repaired_message)
    variable data_low  : natural;
    variable data_high : natural;
    variable get_low   : natural;
    variable get_high  : natural;
  begin
    data_low := data'low;
    get_data_i: for i in 2 to parity_bits loop
      -- We shall get data, where?
      get_low  := (2**(i-1))+1;
      get_high := (2** i   )-1;
      data_high   := data_low + (get_high - get_low);

      if (data_high > data'high) then
        get_high := get_high - (data_high - data'high);
        data_high := data'high;
      end if;
      data(data_high downto data_low) <=
        repaired_message(get_high downto get_low);
      
      data_low := data_high + 1;
    end loop;
  end process;
  
end RTL;
