-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.rataser_util.all;

entity rataser_hamming_decode_dual is
  generic (pipeline_parity : boolean := false;
           pipeline_dual   : boolean := false);
  port (clk           : in  std_logic := '0'; -- Only needed for pipeline.
        -- Input messages
        message_p     : in  std_logic_vector;
        message_n     : in  std_logic_vector;
        -- Output data (when good)
        data          : out std_logic_vector;
        -- Was reconstrcution possible
        good          : out std_logic;
        -- Set if reconstruction relied on one source only
        single_source : out std_logic;
        -- Did we have any bad bit?
        bad_bit       : out std_logic
        );
end rataser_hamming_decode_dual;
  
architecture RTL of rataser_hamming_decode_dual is

  -- The decoded information from the two message versions
  signal data_p         : std_logic_vector(data'range) := (others => '0');
  signal good_p         : std_logic := '0';
  signal single_error_p : std_logic := '0';
  signal double_error_p : std_logic := '0';

  signal data_n         : std_logic_vector(data'range) := (others => '0');
  signal good_n         : std_logic := '0';
  signal single_error_n : std_logic := '0';
  signal double_error_n : std_logic := '0';

  -- Pipelined
  signal data_p_r         : std_logic_vector(data'range) := (others => '0');
  signal single_error_p_r : std_logic := '0';
  signal double_error_p_r : std_logic := '0';

  signal data_n_r         : std_logic_vector(data'range) := (others => '0');
  signal single_error_n_r : std_logic := '0';
  signal double_error_n_r : std_logic := '0';

  -- Internals
  signal data_match      : std_logic := '0';

  -- Internals/output
  signal i_data          : std_logic_vector(data'range) := (others => '0');
  signal i_good          : std_logic := '0';
  signal i_both_sources  : std_logic := '0';
  signal i_any_source    : std_logic := '0';
  signal i_single_source : std_logic := '0';
  signal i_bad_bit       : std_logic := '0';

begin

  -- The two decoders, for the positive and the negative messages.
  hamm_p: entity work.rataser_hamming_decode
    generic map(pipeline_parity => pipeline_parity)
    port map(clk          => clk,
             message      => message_p,
             data         => data_p,
             good         => good_p,
             single_error => single_error_p,
             double_error => double_error_p);
  hamm_n: entity work.rataser_hamming_decode
    generic map(pipeline_parity => pipeline_parity)
    port map(clk          => clk,
             message      => message_n,
             data         => data_n,
             good         => good_n,
             single_error => single_error_n,
             double_error => double_error_n);

  rataser_ppl(p=>pipeline_dual,c=>clk,i=>data_p        ,o=>data_p_r        );
  rataser_ppl(p=>pipeline_dual,c=>clk,i=>single_error_p,o=>single_error_p_r);
  rataser_ppl(p=>pipeline_dual,c=>clk,i=>double_error_p,o=>double_error_p_r);
  rataser_ppl(p=>pipeline_dual,c=>clk,i=>data_n        ,o=>data_n_r        );
  rataser_ppl(p=>pipeline_dual,c=>clk,i=>single_error_n,o=>single_error_n_r);
  rataser_ppl(p=>pipeline_dual,c=>clk,i=>double_error_n,o=>double_error_n_r);

  -- Do the two decoded messages match?
  data_match <= '1' when (data_p_r = data_n_r) else '0';

  -- Cases (se/de = single/double error):
  --
  --  1\2  ok  se  de
  -- ok    C   C   1
  -- se    C   C   1
  -- de    2   2   -
  --
  -- 1: use message from 1; no check possible
  -- 2: use message from 2; no check possible
  -- C: use any message, require them to match
  -- -: both messages broken

  -- Did we have any fault?
  i_bad_bit <= single_error_p_r or double_error_p_r or
               single_error_n_r or double_error_n_r;
  -- Is only one of the messages useful (but at least that one)?
  i_single_source <=
    (double_error_p_r  or double_error_n_r) and not
    (double_error_p_r and double_error_n_r);
  -- Is at least one of the messages useful?
  i_any_source <= (not double_error_p_r) or (not double_error_n_r);
  -- Can we check that the messages match?
  i_both_sources <= (not double_error_p_r) and (not double_error_n_r);
  -- Do we have an good output?  (check if possible, or simply trust)
  i_good <= (i_both_sources and data_match) or (i_single_source);

  -- Deliver the first message, unless it has dual-bit error
  i_data <= data_p_r when (double_error_n_r = '1') else data_n_r;

  -- Assign the output signals
  data <= i_data;
  good <= i_good;

  single_source <= i_single_source;

  bad_bit <= i_bad_bit;
  
end RTL;
