-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.rataser_hamming_util.all;

entity rataser_hamming_encode is
  port (data     : in  std_logic_vector;
        output   : out std_logic_vector
        );
end rataser_hamming_encode;

-- generate a Hamming coded message from the data

architecture RTL of rataser_hamming_encode is

  constant data_size   : positive := data'high + 1 - data'low;

  constant parity_bits : positive := rataser_hamming_parity_bits(data_size);
  constant msg_size    : positive := rataser_hamming_msg_size(data_size);
    
  -- We create data and parity bits separately for less confusion.
  signal data_part     : std_logic_vector(msg_size-1 downto 1) :=
    (others => '1');
  signal parity_part   : std_logic_vector(msg_size-1 downto 1) :=
    (others => '1');
  -- The non-zero message.
  signal message       : std_logic_vector(msg_size-1 downto 1) :=
    (others => '1');
  -- Overall parity.
  signal p0            : std_logic;
begin

  --  0  1  2  3  4  5  6  7  8  9 10 11 12 13  14  15
  -- p0 p1 p2 d1 p4 d2 d3 d4 p8 d5 d6 d7 d8 d9 d10 d11
  --  P  P  P  -  P  -  -  -  P  -  -  -  -  -   -   -
  --  0  1  2     3           4                        (parity bit)
  -- -1  0  1     2           3                        (i)

  -- Data is injected after each parity bit 2**(i-1), i.e.
  -- from 2**(i-1)+1 to 2**i-1, with i = 2,3,...

  process(data)
    variable data_low    : natural;
    variable data_high   : natural;
    variable insert_low  : natural;
    variable insert_high : natural;
  begin
    data_part <= (others => '0');
    data_low := data'low;
    insert_data_i: for i in 2 to parity_bits loop
      -- We shall insert data, where?
      insert_low  := (2**(i-1))+1;
      insert_high := (2** i   )-1;
      data_high   := data_low + (insert_high - insert_low);

      if (data_high > data'high) then
        insert_high := insert_high - (data_high - data'high);
        data_high := data'high;
      end if;
      data_part(insert_high downto insert_low) <=
        data(data_high downto data_low);
      
      data_low := data_high + 1;
    end loop;
  end process;

  process(data, data_part)
    variable insert_at   : natural;
    variable tmp_partity : std_logic;
  begin
    parity_part <= (others => '0');
    insert_parity_i: for i in 1 to parity_bits loop
      insert_at := 2**(i-1);  -- 1, 2, 4, 8, ...
      tmp_partity := '0';
      -- Find the bits that are to be included
      get_parity_j: for j in 1 to msg_size-1 loop
        if ((std_logic_vector(to_unsigned(j, parity_bits)) and
             std_logic_vector(to_unsigned(insert_at, parity_bits))) /=
            std_logic_vector(to_unsigned(0, parity_bits))) then
          tmp_partity := tmp_partity xor data_part(j);
        end if;
      end loop;
      parity_part(insert_at) <= tmp_partity;
    end loop;
  end process;

  message <= data_part or parity_part;

  get_p0: entity work.rataser_big_xor
    port map(inputs => message,
             output => p0);

  output <= (message & p0);
  
end RTL;
