-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.rataser_util.all;

package rataser_hamming_util is

  procedure rataser_hamming_sizes(data_size   : in positive;
                                  parity_bits : out positive;  -- excluding p0
                                  msg_size    : out positive);

  function rataser_hamming_parity_bits(data_size : positive) return positive;
  function rataser_hamming_msg_size(data_size : positive) return positive;

end rataser_hamming_util;

package body rataser_hamming_util is

  procedure rataser_hamming_sizes(data_size   : in positive;
                                  parity_bits : out positive;  -- excluding p0
                                  msg_size    : out positive) is

    variable est_parity_bits              : positive;
    variable max_data_for_est_parity_bits : positive;
    variable num_parity_bits              : positive;
    variable max_data_for_parity          : positive;
    variable silent_bits                  : natural;
  begin

    --  0  1  2  3  4  5  6  7  8  9 10 11 12 13  14  15
    -- p0 p1 p2 d1 p4 d2 d3 d4 p8 d5 d6 d7 d8 d9 d10 d11
    --  P  P  P  -  P  -  -  -  P  -  -  -  -  -   -   -
    
    -- with m parity bits (excluding the extended p0), we get a
    -- message length of 2^m, and thus n = 2^m-m-1 data bits.
    -- Knowing the number of data bits, that equation is not nice
    -- to solve.  But we can first estimate m from n = 2^estm, which
    -- will possibly be too low:
    --
    -- n:    1 2 3 4 5 6 7 8 9 10 11 12 13
    -- m:    2 3 3 3 4 4 4 4 4  4  4  5  5
    -- estm: 0 1 2 2 3 3 3 3 4  4  4  4  4

    est_parity_bits := log2o1(data_size);

    -- How much data can that take?    
    max_data_for_est_parity_bits :=
      (2 ** est_parity_bits) - est_parity_bits - 1;

    -- Then we can decide if we need another parity bit
    if (max_data_for_est_parity_bits < data_size) then
      if (data_size <= 2) then
        num_parity_bits := est_parity_bits + 2;
      else
        num_parity_bits := est_parity_bits + 1;
      end if;
    else
      num_parity_bits := est_parity_bits;
    end if;

    parity_bits := num_parity_bits;

    -- How much data can that actually take?
    max_data_for_parity :=
      (2 ** num_parity_bits) - num_parity_bits - 1;
    -- How much is then left silent?
    silent_bits := max_data_for_parity - data_size;

    -- How long is our message? (add 1 for p0)
    msg_size := data_size + num_parity_bits + 1;
  end;

  function rataser_hamming_parity_bits(data_size : positive) return positive is
    variable parity_bits : positive;
    variable msg_size    : positive;
  begin
    rataser_hamming_sizes(data_size   => data_size,
                          parity_bits => parity_bits,
                          msg_size    => msg_size);
    return parity_bits;
  end;
    
  function rataser_hamming_msg_size(data_size : positive) return positive is
    variable parity_bits : positive;
    variable msg_size    : positive;
  begin
    rataser_hamming_sizes(data_size   => data_size,
                          parity_bits => parity_bits,
                          msg_size    => msg_size);
    return msg_size;
  end;

end rataser_hamming_util;
