-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity rataser_min_max_filter is
  generic(min_cycle_bits : integer;
          max_cycle_bits : integer);
  port (clk            : in  std_logic;

        -- Input value.
        input          : in  unsigned;
        take           : in  std_logic;

        -- Output filtered average value.
        output         : out unsigned;
        -- Output the filtered values.
        output_min_max : out unsigned;
        output_max_min : out unsigned;
        -- Output has new value.
        output_new     : out std_logic
        );
end rataser_min_max_filter;

architecture RTL of rataser_min_max_filter is

  -- Filter for how many or few samples we have seen over 2^min_cycle_bits
  -- pulses.
  signal accum_min : unsigned(input'range) := (others => '0');
  signal accum_max : unsigned(input'range) := (others => '0');

  -- Second-stage filter (to avoid outliers), another factor 2^max_cycle_bits
  -- pulses.
  signal accum_min_max : unsigned(input'range) := (others => '0');
  signal accum_max_min : unsigned(input'range) := (others => '0');

  -- Second-stage filter results.
  signal hold_min_max : unsigned(input'range) := (others => '0');
  signal hold_max_min : unsigned(input'range) := (others => '0');

  -- Average of the second-stage filter.
  signal hold_avg : unsigned(input'range) := (others => '0');
  -- Temporary value for the calculation.
  signal accum_avg_calc : unsigned(input'left+1 downto 0) := (others => '0');

  -- Count how many times we have seen a pulse
  signal long_count : unsigned (min_cycle_bits+max_cycle_bits-1 downto
                                0) := (others => '0');
  signal long_count_next : unsigned (long_count'left+1 downto
                                     0) := (others => '0');

  signal wrap_min : std_logic := '0';
  signal wrap_max : std_logic := '0';

  -- Second-stage filter has a new value.
  signal hold_new : std_logic := '0';

begin

  long_count_next <= ('0' & long_count) + 1;

  wrap_min <= long_count_next(min_cycle_bits) xor long_count(min_cycle_bits);
  wrap_max <= long_count_next(long_count_next'left);

  -- Calculate the mid-point of the min and max accumulated sample counts.
  accum_avg_calc <= (('0' & accum_max_min) + accum_min_max);
    
  process(clk)
  begin
    if (rising_edge(clk)) then

      -- Default.
      hold_new <= '0';

      if (take = '1') then

        -- Filter the values, reset every minor wrap.
        if (input > accum_max or
            wrap_min = '1') then
          accum_max <= input;
        end if;
        if (input < accum_min or
            wrap_min = '1') then
          accum_min <= input;
        end if;

        -- Second-stage filter, reset every major wrap.
        if ((accum_min > accum_min_max and
             wrap_min = '1') or
            wrap_max = '1') then
          accum_min_max <= accum_min;
        end if;
        if ((accum_max < accum_max_min and
             wrap_min = '1') or
            wrap_max = '1') then
          accum_max_min <= accum_max;
        end if;

        -- Update the average when counter major wraps.
        if (wrap_max = '1') then
          hold_new <= '1';
          hold_avg <= accum_avg_calc(accum_avg_calc'left downto 1);
          hold_min_max <= accum_min_max;
          hold_max_min <= accum_max_min;
        end if;

        long_count <= long_count_next(long_count'range);
      end if;

    end if;
  end process;

  output         <= hold_avg;

  output_min_max <= hold_min_max;
  output_max_min <= hold_max_min;

  output_new     <= hold_new;

end RTL;
