-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.rataser_records.all;
use work.rataser_sym24word8_util;

entity rataser_signals_recv is
  generic(-- Actually use the slow_clk_jitter (testbench).
          use_slow_clk_jitter      : boolean := false;
          -- Pipeline stages to handle diff calculation.
          pipeline_sym24_diff_in   : boolean := false;
          pipeline_sym24_diff_1    : boolean := false;
          pipeline_sym24_diff_2    : boolean := false;
          pipeline_sym24_diff_out  : boolean := false;
          pipeline_sym24_match_out : boolean := false;
          pipeline_sym24_pick      : boolean := false;
          pipeline_sym24_out       : boolean := false
          );
  port (fast_clk:      in std_logic;
        fast_clk90:    in std_logic;

        slow_clk:      in std_logic;
        -- Should be same as slow_clk.
        -- Only give jittering clock in testbench.
        slow_clk_jitter: in std_logic := '0';

        -- Input serial signal.
        receive:       in std_logic;

        -- Output signals (only 0 if single_symbol_mode),
        -- only (3 downto 0 if multi_signal4_mode).
        signals:       out std_logic_vector(10 downto 0);
        -- Auxiliary information.
        info_bit:      out std_logic;
        -- Strobe signal that we have decoded a new message.
        msg_strobe:    out std_logic;

        -- These are to be handled internally!  From slow message.
        -- Transmission of only one signal selected.
        single_symbol_mode: in std_logic;
        -- Transmission of multiple signals recv 4 signals.
        multi_signal4_mode: in std_logic;  -- set to 0
        -- Test mode (check cycling counter)
        test_mode:     in std_logic;

        -- Counter of bad bits.  On multiple errors within the same
        -- will be counted only as 1 or 2.
        bad_bit_count: out std_logic_vector;
        -- Counter of message cycles received in test mode.
        -- Note: also count if bad bits are detected. 
        test_cycle_count: out std_logic_vector
        );
end;

architecture RTL of rataser_signals_recv is

  signal slow_clk_jitter_actual : std_logic;

  signal saz            : std_logic_vector(2 to 3);
  signal sbz            : std_logic_vector(2 to 3);
  signal scz            : std_logic_vector(2 to 3);
  signal sdz            : std_logic_vector(2 to 3);

  signal bit_val        : std_logic_vector(0 to 1);
  signal got_bit        : std_logic_vector(0 to 1);

  signal mark_w8a1_2bit : unsigned(1 downto 0);
  signal out_w8p        : unsigned(7 downto 0);
  signal out_w8n        : unsigned(7 downto 0);
  signal out_aux1p      : std_logic;
  signal out_aux1n      : std_logic;

  signal mark_w1_2bit   : unsigned(1 downto 0);
  signal out_w1p        : std_logic;
  signal out_w1n        : std_logic;

  signal sym_synced     : std_logic;
  signal sym_bad_bit    : std_logic;

  signal hamm_data_8p   : std_logic_vector(7 downto 0);
  signal hamm_data_8n   : std_logic_vector(7 downto 0);

  -- In slow clock domain:
  signal hamm_data_8p_r    : std_logic_vector(7 downto 0);
  signal hamm_data_8n_r    : std_logic_vector(7 downto 0);

  signal mark_w8a1_2bit_r  : unsigned(1 downto 0);
  signal mark_w8a1_2bit_rr : unsigned(1 downto 0);
  
  signal hamm_word4     : std_logic_vector(3 downto 0);
  signal hamm_good      : std_logic;
  signal hamm_single_source : std_logic;
  signal hamm_bad_bit   : std_logic;

  -- Since we might run at a slower clock, we detect new data when a
  -- two-bit counter has changed value.
  -- (A short strobe might be missed.)
  signal prev_mark_w8a1_2bit : unsigned(1 downto 0) := (others => '0');

  signal i_signals    : std_logic_vector(10 downto 0) := (others => '0');
  signal i_info_bit   : std_logic := '0';
  signal i_msg_strobe : std_logic := '0';

  -- Bad bit counting
  signal sym_bad_bit_latched  : std_logic := '0';
  signal hamm_bad_bit_latched : std_logic := '0';

  signal i_bad_bit_count : unsigned(bad_bit_count'range) :=
    (others => '0');

  -- Test cycle counting
  signal test_cycle : std_logic := '0';
  
  signal i_test_cycle_count : unsigned(test_cycle_count'range) :=
    (others => '0');

  -- For timing test:
  signal delay : std_logic_vector(0 to 7);

  -- For test mode:
  signal test_error   : std_logic := '0';
  signal expect_test_value : std_logic_vector(7 downto 0) := (others => '0');

begin

  sh: if (true) generate
  sample_4phase: entity work.rataser_4phase_sample
    port map(clk     => fast_clk,
             clk90   => fast_clk90,

             serial_in => receive,

             saz     => saz,
             sbz     => sbz,
             scz     => scz,
             sdz     => sdz);
  end generate;
  -- For timing test:
  st: if (false) generate
    process(fast_clk)
    begin
      if (rising_edge(fast_clk)) then
        delay <= delay(1 to 7) & receive;
        saz <= delay(0 to 1);
        sbz <= delay(2 to 3);
        scz <= delay(4 to 5);
        sdz <= delay(6 to 7);
      end if;
    end process;
  end generate;

  deser_4phase: entity work.rataser_4phase_deser
    port map(clk     => fast_clk,

             saz     => saz,
             sbz     => sbz,
             scz     => scz,
             sdz     => sdz,

             bit_out => bit_val,
             got_bit => got_bit);

  -- If we would not use the deserialiser, and run on the same clock
  -- as the sender:
  -- got_bit <= '0' & '1';
  -- bit_val <= '0' & receive;

  decode_sym: entity work.rataser_sym24word8_decode
    generic map(pipeline_diff_in   => pipeline_sym24_diff_in,
                pipeline_diff_1    => pipeline_sym24_diff_1,
                pipeline_diff_2    => pipeline_sym24_diff_2,
                pipeline_diff_out  => pipeline_sym24_diff_out,
                pipeline_match_out => pipeline_sym24_match_out,
                pipeline_pick      => pipeline_sym24_pick,
                pipeline_out       => pipeline_sym24_out)
    port map(clk => fast_clk,

             bit_val => bit_val,
             got_bit => got_bit,

             mark_w8a1      => open,
             mark_w8a1_2bit => mark_w8a1_2bit,
             out_w8p        => out_w8p,
             out_w8n        => out_w8n,
             out_aux1p      => out_aux1p,
             out_aux1n      => out_aux1n,

             mark_w1        => open,
             mark_w1_2bit   => mark_w1_2bit,

             out_w1p        => out_w1p,
             out_w1n        => out_w1n,

             mark_w16       => open,
             out_w16p       => open,
             out_w16n       => open,

             synced         => sym_synced,
             bad_bit        => sym_bad_bit
             );

  hamm_data_8p <=     std_logic_vector(out_w8p);
  hamm_data_8n <= not std_logic_vector(out_w8n);

  hh: if (true) generate
    decode_hamm: entity work.rataser_hamming_decode_dual
      port map(message_p     => hamm_data_8p_r,
               message_n     => hamm_data_8n_r,

               data          => hamm_word4,
               good          => hamm_good,
               single_source => hamm_single_source,
               bad_bit       => hamm_bad_bit);
  end generate;
  ht: if (false) generate
    -- For timing test
    hamm_word4 <=
      hamm_data_8p(3 downto 0) xor hamm_data_8p(7 downto 4) xor
      hamm_data_8n(3 downto 0) xor hamm_data_8n(7 downto 4);
    hamm_good          <= hamm_data_8p(0);
    hamm_single_source <= hamm_data_8p(1);
  end generate;  

  slow_clk_jitter_actual <=
    slow_clk_jitter when (use_slow_clk_jitter) else slow_clk;

  process(slow_clk_jitter_actual)
  begin
    if (rising_edge(slow_clk_jitter_actual)) then
      -- Latched on a jittering clock (in testbench) to
      -- test clock-domain crossing failure.
      mark_w8a1_2bit_r <= mark_w8a1_2bit;
    end if;
  end process;

  -- BL: Needed, otherwise GHDL warns about others not being set
  -- putting this here (instead of in the process) to run faster.
  i_signals(10 downto 4) <= (others => '0');

  process(slow_clk)
  begin
    if (rising_edge(slow_clk)) then
      i_msg_strobe <= '0';
      test_cycle <= '0';
      test_error <= '0';
      hamm_bad_bit_latched <= '0';
      sym_bad_bit_latched <= '0';

      -- In order for the hamming block to actually run on the slow
      -- clock, its input data must be latched.  We must also latch
      -- the marks.  Since slow_clock and fast_clock are phase-related
      -- we do not need to do an additional latch for
      -- anti-metastability.
      hamm_data_8p_r <= hamm_data_8p;
      hamm_data_8n_r <= hamm_data_8n;
      -- mark_w8a1_2bit_r <= mark_w8a1_2bit; -- latched above (jitter)
      -- For clock-domain crossing (test mode).
      mark_w8a1_2bit_rr <= mark_w8a1_2bit_r;

      if (test_mode = '0' and
          mark_w8a1_2bit_r /= prev_mark_w8a1_2bit) then
        prev_mark_w8a1_2bit <= mark_w8a1_2bit_r;

        -- For the bad bit counter (only good on mark update).
        sym_bad_bit_latched <= sym_bad_bit;

        -- Only update the output data if we managed to decode the
        -- message.
        if (hamm_good = '1') then
          i_signals(3 downto 0) <= hamm_word4;
          i_msg_strobe <= '1';
        end if;

        -- We only update the aux bit if the positive and negative copy
        -- match.
        if (out_aux1p /= out_aux1n) then
          i_info_bit <= out_aux1p;
        end if;

        if (multi_signal4_mode = '1' and
            test_mode = '0') then
          hamm_bad_bit_latched <= hamm_bad_bit;
        end if;
      end if;

      -- In test mode, we assume that the slow_clk is unrelated
      -- to fast_clk (or has very naughty multiplier/divisor) such
      -- that we have to do clock-domain crossing.

      -- That is handled by _detecting_ the new data availability by
      -- any change in two-cycles old marker bits, while all data
      -- that is latched is taken from one cycle ago (including the
      -- marker).  This relies on the fact that the marker will only
      -- by updated once per 24 fast clock cycles, and we are not
      -- operating with more than say a differnce of a factor 6.
      -- Also the prev marker is sampled from a newer (i.e. stable
      -- value, since the value we trigger on may be the first bit
      -- changing, and the second might have been just after the
      -- slow_clk latched.

      -- Use mark_w8a1_2bit_r in if-statement and a jittering
      -- slow_clk_jitter to see domain-crossing errors (spurious
      -- bad_bit_count increments).
      if (test_mode = '1' and
          mark_w8a1_2bit_rr /= prev_mark_w8a1_2bit) then
        prev_mark_w8a1_2bit <= mark_w8a1_2bit_r;

        -- We have gotten one more cycle of data.
        test_cycle <= '1';

        -- Was it bad?
        if ((hamm_data_8p_r /= expect_test_value) or
            (hamm_data_8n_r /= expect_test_value)) then
          test_error <= '1';
        end if;

        -- In an attempt to avoid double-counting errors, we only
        -- update using the current values if they match.
        -- If they mismatch, we increment previous by 1.
        if (hamm_data_8p_r = hamm_data_8n_r) then
          expect_test_value <= std_logic_vector(unsigned(hamm_data_8p_r) +
                                                to_unsigned(1,1));
        else
          expect_test_value <= std_logic_vector(unsigned(expect_test_value) +
                                                to_unsigned(1,1));
        end if;
      end if;

      -- Update the bad bit counter
      i_bad_bit_count <= i_bad_bit_count +
                         ("" & sym_bad_bit_latched) +
                         ("" & hamm_bad_bit_latched) +
                         ("" & test_error);

      i_test_cycle_count <= i_test_cycle_count +
                            ("" & test_cycle);
    end if;
  end process;

  signals    <= i_signals;
  info_bit   <= i_info_bit;
  msg_strobe <= i_msg_strobe;

  bad_bit_count    <= std_logic_vector(i_bad_bit_count);
  test_cycle_count <= std_logic_vector(i_test_cycle_count);
  
end RTL;
