-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.rataser_records.all;
use work.rataser_sym24word8_util.all;

entity rataser_signals_send is
    port (clk:           in std_logic;

          -- Input signals (only index 0 if single_symbol_mode,
          -- only (3 downto 0) if multi_signal4_mode).
          signals:       in std_logic_vector(10 downto 0);
          -- Auxiliary information.
          info_bit:      in std_logic;

          -- Transmission of only one signal selected.
          single_symbol_mode: in std_logic;
          -- Transmission of multiple signals send only 4 signals.
          multi_signal4_mode: in std_logic; -- set to 0

          -- Test mode (emit cycling counter)
          test_mode:     in std_logic := '0';

          -- For testing of faults only, yank running send counter.
          fault_shift_count: in std_logic_vector(4 downto 0) :=
            (others => '0');

          -- Output message.
          transmit:      out std_logic := '0';
          -- Oscilloscope sync signal for message.
          transmit_sync: out std_logic := '0'
          );
end;

architecture RTL of rataser_signals_send is

  -- Latched single signal to transfer.
  signal signal0_latch : std_logic := '0';
  -- Latched auxiliary information.
  signal info_bit_latch : std_logic := '0';

  -- Input payload (message) to Hamming encoder (may also be 3 downto 0).
  signal hamm_word11 : std_logic_vector(10 downto 0) := (others => '0');
  -- Output block from Hamming encoder (may also be 7 downto 0).
  signal hamm_data16 : std_logic_vector(15 downto 0) := (others => '0');

  -- Latched hamming block.
  signal hamm_data16_latch : std_logic_vector(15 downto 0) := (others => '0');

  -- Input data payload to 24-slot sequence.
  signal sym24_data8   : std_logic_vector(7 downto 0);
  signal sym24_data8_u : unsigned(7 downto 0);
  -- Sequence of 24-slot pattern.
  signal sym24       : std_logic_vector(0 to 23);
  -- Sequence of 24-slot pattern in 32 slots (slots % 4 == 3 unused).
  signal sym32       : std_logic_vector(0 to 31) := (others => '0');

  -- Bit to transmit, based on current count value.
  signal bit_out     : std_logic;

  -- Value (raw bit) to transmit (internal).
  signal i_transmit    : std_logic := '0';
  -- Scope-watching sync signal for transmit sequence;
  signal i_transmit_sync : std_logic := '0';

  -- Send counter, runs through 0-31 (or 0-63).  It always
  -- jumps past values % 4 == 3, i.e. 3, 7, 11, 15...
  -- Effectively, period is 24 or 48 cycles.

  signal count_sym        : unsigned(5 downto 0) := (others => '0');
  signal count_sym_next   : unsigned(6 downto 0) := (others => '0');
  signal count_sym_wrap24 : std_logic := '0';
  signal count_sym_wrap48 : std_logic := '0';

  -- Count bits 0..4  [0..31]  determine which bit of 24-slot pattern
  -- Count bit  5     [32]     determine first/second Hamming message
  --                           (if using 11-symbol pattern)

  -- Pipelined
  signal sym32_p     : std_logic_vector(0 to 31) := (others => '0');
  signal count_sym_p : unsigned(5 downto 0) := (others => '0');

begin
  
  -- Add 1 when we have not bit for 2, and add 2 when we are two, thus we
  -- count 0, 1, 2,  4, 5, 6,  8, 9, 10,  12, ...
  count_sym_next   <= ('0' & count_sym) + (count_sym(1) & (not count_sym(1)))
                      + unsigned(fault_shift_count);
  count_sym_wrap24 <= count_sym_next(5) xor count_sym(5);
  count_sym_wrap48 <= count_sym_next(6);

  -- We can use the same Hamming encoder, provided that the
  -- upper (unused) data bits that go into the encoder are all 0, so
  -- as to not affect the lower parity bits.

  hamm_word11(3 downto 0) <= signals(3 downto 0);
  hamm_word11(10 downto 4) <=
    (10 downto 4 => '0') when (multi_signal4_mode = '1') else
    signals(10 downto 4);  
  
  hamm: entity work.rataser_hamming_encode
    port map(data   => hamm_word11,
             output => hamm_data16);

  -- What data to pack into the symbol is given by the current mode.
  -- In single-signal mode, we always have the first signal.
  -- In multi-signal mode, it depends on if we pack 4 or 11 signals,
  -- and if 11, on the counter value.

  sym24_data8 <=
    (7 downto 0 => signal0_latch) when (single_symbol_mode = '1' and
                                        test_mode = '0') else
    hamm_data16_latch(7 downto 0) when (multi_signal4_mode = '1' or
                                        test_mode = '1' or
                                        count_sym(5) = '0') else
    hamm_data16_latch(15 downto 8);

  sym24_data8_u <= unsigned(sym24_data8);

  -- Generate the transmit sequence, which is used by the multiplexer.
  -- This generation is really just assigning signals to slots, so pure
  -- routing.

  -- TODO: in 11-signal-mode (2 sym24 groups), we need the aux bit to
  -- mark which group is which.  Thus one of the 11 bits need to be
  -- used for the info bit.

  gen_sequence(seq  => sym24,
               data => sym24_data8_u,
               aux  => info_bit_latch);

  gen_sequence32(seq32 => sym32,
                 seq24 => sym24);

  -- Select the bit to transmit (select with counter value 0..31).
  -- Pipelined.  (Note: virtex6 only has 16:1 multiplexers, so
  -- will be a two-stage operation.)
  bit_out <= sym32_p(to_integer(count_sym_p(4 downto 0)));

  process(clk)
  begin
    if (rising_edge(clk)) then

      -- Pipeline
      sym32_p     <= sym32;
      count_sym_p <= count_sym;

      -- Update the counter
      count_sym <= count_sym_next(count_sym'range);

      -- When we are operating in one-signal mode, we latch the
      -- signals every time the current counter value is two.
      if (count_sym(1) = '1') then
        signal0_latch <= signals(0);
      end if;
      if (count_sym_wrap24 = '1') then
        info_bit_latch <= info_bit;
      end if;
      -- The data for multi-signal mode is latched when we wrap
      -- the appropriate counter        
      if ((multi_signal4_mode = '1' and count_sym_wrap24 = '1') or
          count_sym_wrap48 = '1') then
        hamm_data16_latch <= hamm_data16;
      end if;
      -- In test mode, no Hamming code, just an incrementing counter.
      if (count_sym_wrap24 = '1' and
          test_mode = '1') then
        hamm_data16_latch <=
          std_logic_vector(unsigned(hamm_data16_latch) + to_unsigned(1,1));
      end if;

      -- Latch the value to transmit
      i_transmit <= bit_out;
      if (multi_signal4_mode = '1') then
        i_transmit_sync <= count_sym_wrap24;
      else
        i_transmit_sync <= count_sym_wrap48;
      end if;
    end if;
  end process;

  -- Assign the actual output signals
  transmit      <= i_transmit;
  transmit_sync <= i_transmit_sync;

end RTL;
