-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.rataser_records.all;
use work.rataser_sym24word8_util.all;
use work.rataser_util.all;

-- Cycle 0 - Get new input bits [0,1,2],
--           calculate mismatch distance.
-- Cycle 1 - Mismatch distance <= 1 -> reset counter,
--           compare next_counter == wrap to distance <= 1.

-- got_bit(0 to 1) is one of "00", "01", or "11", for 0, 1 or 2 bits.
-- bit_val(0 to 1) is thus   "xx", "xb", or "bb", with b bit, x unused.

-- Consider the situation where the new bits just complete the pattern:
--
-- Cycle 0:
--   got_bit(1) is 1.
--   bit_val(1) contains the last needed bit.
--   got_bit(0) / bit_val(0) may be set (if we got 2 bits).
--   combined ends with the completed pattern.
--   diff1 (based on combined to last bit) is 0.
--   match1 (based on diff1, based on combined) is 1.
--
--   count_next is 0 (+ carry).
--   count_wrap is 1 (carry of count_next).
--
-- Cycle 1:
--   count is 0 (set to 0 when starting sync tracking).

-- Consider the situation where the new bits are two, and overshoots
-- the complete pattern with one bit:
--
-- Cycle 0:
--   got_bit(1) is 1.
--   bit_val(1) contains the first bit of the next pattern.
--   got_bit(0) is 1.
--   bit_val(0) contains the last needed bit.
--   combined ends with first bit of the next pattern.
--   diff0 (based on combined to second-last bit) is 0.
--   match0 (based on diff0, based on combined) is 1.
--
--   count_next is 1 (+ carry).
--   count_wrap is 1 (carry of count_next).
--
-- Cycle 1:
--   count is 1 (set to 1 when starting sync tracking).

-- Generally xxx1 (diff1, match1) means with end using bit_val(1),
-- and xxx0 (diff0, match0) means with end using bit_val(0).

entity rataser_sym24word8_decode is
  generic(-- Pipeline stages to handle diff calculation.
          pipeline_diff_in   : boolean := false;
          pipeline_diff_1    : boolean := false;
          pipeline_diff_2    : boolean := false;
          pipeline_diff_out  : boolean := false;
          pipeline_match_out : boolean := false;
          -- Pipeline the output latching.
          pipeline_pick      : boolean := false;
          -- Pipeline output registers.
          pipeline_out       : boolean := false;
          -- Report a sequence of pure 16 (plain+inv) data after mark.
          puredata_seq : boolean := false);
  port (clk            : in std_logic;

        -- Input from bit decoder.
        bit_val        : in std_logic_vector(0 to 1);
        got_bit        : in std_logic_vector(0 to 1);
        missed_bit     : in std_logic := '0';

        -- Output to next stage: full symbol -> data word.
        -- Two marks are provided: strobe and an updating 2-bit counter.
        -- (The updating 2-bit counter is needed when used in another
        -- clock domain, since the strobe can be missed.)
        -- Values are latched and held until next update.
        -- Note: the aux bit is marked here, even if another mode is used
        -- to decode the payload.
        mark_w8a1      : out std_logic;
        mark_w8a1_2bit : out unsigned(1 downto 0);
        -- Only use output values at strobe / after 2-bit counter update.
        -- (Output values may also be updated when reception lock is
        -- not in sync).
        out_w8p        : out unsigned(7 downto 0);
        out_w8n        : out unsigned(7 downto 0);
        out_aux1p      : out std_logic;
        out_aux1n      : out std_logic;

        -- Output to next stage: single word bit (every three cycles).
        -- Values are latched and held until next update.
        mark_w1        : out std_logic;
        mark_w1_2bit   : out unsigned(1 downto 0);
        -- Only use output values at strobe / after 2-bit counter update.
        out_w1p        : out std_logic;
        out_w1n        : out std_logic;

        -- Output of full 32-bit data word in case the special
        -- pattern detection is in use, which signals that the
        -- symbol sequence is interrupted by pure (pair-inverted) data.
        mark_w16       : out std_logic;
        -- Only use output values at strobe / after 2-bit counter update.
        out_w16p       : out unsigned(15 downto 0);
        out_w16n       : out unsigned(15 downto 0);

        -- General synchronisation information:
        -- Continuous signal telling if we are good.
        synced        : out std_logic;
        -- Strobed by mark_w8a1/mark_w8a1_2bit.
        bad_bit       : out std_logic
        );

end rataser_sym24word8_decode;


architecture RTL of rataser_sym24word8_decode is

  -- Invert the input signal bits.
  signal invert_signal : std_logic := '0';
  -- Inverted bits (if inverting).
  signal inv_bit     : std_logic_vector(0 to 1) := (others => '0');

  -- Historical data (including new), 0 is oldest, 32 is newest.
  -- Symbol pattern only contain 24 bits, but for puredata_seq mode
  -- we return 32 most recent bits (thus need 33, in case we got 2).
  signal combined : std_logic_vector(0 to 32) := (others => '0');

  -- Cycle 1.
  signal combined_prev : std_logic_vector(0 to 32) := (others => '0');

  -- Pipeline stages.
  signal combined_p0 : std_logic_vector(0 to 32) := (others => '0');
  signal got_bit_p0  : std_logic_vector(0 to  1) := (others => '0');

  signal combined_p1 : std_logic_vector(0 to 32) := (others => '0');
  signal got_bit_p1  : std_logic_vector(0 to  1) := (others => '0');

  signal combined_p2 : std_logic_vector(0 to 32) := (others => '0');
  signal got_bit_p2  : std_logic_vector(0 to  1) := (others => '0');

  signal combined_p3 : std_logic_vector(0 to 32) := (others => '0');
  signal got_bit_p3  : std_logic_vector(0 to  1) := (others => '0');

  -- The stage used in the locked-loop.
  signal combined_r : std_logic_vector(0 to 32) := (others => '0');
  signal got_bit_r  : std_logic_vector(0 to  1) := (others => '0');

  -- Extra stage for use in the final value-pick latch.
  signal got_bit_rr : std_logic_vector(0 to  1) := (others => '0');

  -- How well do we match
  signal diff0_p2   : unsigned(3 downto 0) := (others => '0');
  signal diff1_p2   : unsigned(3 downto 0) := (others => '0');
  -- Pipeline stage at end of diff calculator.
  signal diff0_p3   : unsigned(3 downto 0) := (others => '0');
  signal diff1_p3   : unsigned(3 downto 0) := (others => '0');
  -- Is it considered a match?
  signal match0_p3  : std_logic := '0';
  signal match1_p3  : std_logic := '0';
  -- Did we find bad bits?
  signal badbit0_p3 : std_logic := '0';
  signal badbit1_p3 : std_logic := '0';
  -- Combination of mismatch or bad bit in pattern.
  signal bad_bit0_p : std_logic := '0';
  signal bad_bit1_p : std_logic := '0';
  -- Pipeline for value-pick.
  signal bad_bit0_pr : std_logic := '0';
  signal bad_bit1_pr : std_logic := '0';

  -- Pipelined, for used in locked-loop.
  signal match0_r   : std_logic := '0';
  signal match1_r   : std_logic := '0';
  signal badbit0_r  : std_logic := '0';
  signal badbit1_r  : std_logic := '0';

  -- Symbol sequence counter, locked-loop main actor.
  signal count       : unsigned(6 downto 0) := (others => '0');
  -- Number of bits we got, add to sequence counter.
  signal count_add_got_bits : unsigned(1 downto 0) := (others => '0');
  -- Extra addition to jump past count % 4 == 3.
  signal count_add_jump : std_logic := '0';
  -- Next value of sequence counter (incl carry)
  signal count_next  : unsigned(7 downto 0) := (others => '0');
  signal count_wrap  : std_logic := '0';
  signal count_wrap2 : std_logic := '0';
  signal count_wrap7 : std_logic := '0';

  -- For debugging.
  signal got0_countlow0 : std_logic := '0';
  signal got0_countlow1 : std_logic := '0';
  signal got0_countlow2 : std_logic := '0';
  signal got0_countlow3 : std_logic := '0';

  signal got2_countlow0 : std_logic := '0';
  signal got2_countlow1 : std_logic := '0';
  signal got2_countlow2 : std_logic := '0';
  signal got2_countlow3 : std_logic := '0';

  -- Pipelined for value-pick.
  signal count_next_r  : unsigned(7 downto 0) := (others => '0');
  signal count_wrap_r  : std_logic := '0';
  signal count_wrap2_r : std_logic := '0';

  -- Are we counting to get 32 bits of pure data?
  signal puredata32_pending : std_logic := '0';

  -- State of in-sync-tracking up-down-counter.
  signal match_sync : rataser_sync_count_iface :=
    rataser_sync_count_iface_zero;

  -- Pipelined for value-pick
  signal match_sync_synced_r : std_logic := '0';

  -- Decoded payload words
  signal w8sp0      : unsigned(7 downto 0) := (others => '0');
  signal w8sp1      : unsigned(7 downto 0) := (others => '0');
  signal w8sn0      : unsigned(7 downto 0) := (others => '0');
  signal w8sn1      : unsigned(7 downto 0) := (others => '0');
  signal a1sp0      : std_logic := '0';
  signal a1sp1      : std_logic := '0';
  signal a1sn0      : std_logic := '0';
  signal a1sn1      : std_logic := '0';

  signal w1sp0      : std_logic := '0';
  signal w1sp1      : std_logic := '0';
  signal w1sn0      : std_logic := '0';
  signal w1sn1      : std_logic := '0';

  signal w16sp0     : unsigned(15 downto 0) := (others => '0');
  signal w16sp1     : unsigned(15 downto 0) := (others => '0');
  signal w16sn0     : unsigned(15 downto 0) := (others => '0');
  signal w16sn1     : unsigned(15 downto 0) := (others => '0');

  -- Pipelined for value-pick
  signal w8sp0_r    : unsigned(7 downto 0) := (others => '0');
  signal w8sp1_r    : unsigned(7 downto 0) := (others => '0');
  signal w8sn0_r    : unsigned(7 downto 0) := (others => '0');
  signal w8sn1_r    : unsigned(7 downto 0) := (others => '0');
  signal a1sp0_r    : std_logic := '0';
  signal a1sp1_r    : std_logic := '0';
  signal a1sn0_r    : std_logic := '0';
  signal a1sn1_r    : std_logic := '0';

  signal w1sp0_r    : std_logic := '0';
  signal w1sp1_r    : std_logic := '0';
  signal w1sn0_r    : std_logic := '0';
  signal w1sn1_r    : std_logic := '0';

  -- Latched payload words and markers
  signal i_mark_w8a1      : std_logic := '0';
  signal i_mark_w8a1_2bit : unsigned(1 downto 0) := (others => '0');
  signal i_out_w8p        : unsigned(7 downto 0) := (others => '0');
  signal i_out_w8n        : unsigned(7 downto 0) := (others => '0');
  signal i_out_aux1p      : std_logic := '0';
  signal i_out_aux1n      : std_logic := '0';

  signal i_mark_w1        : std_logic := '0';
  signal i_mark_w1_2bit   : unsigned(1 downto 0) := (others => '0');
  signal i_out_w1p        : std_logic := '0';
  signal i_out_w1n        : std_logic := '0';

  signal i_mark_w16       : std_logic := '0';
  signal i_out_w16p       : unsigned(15 downto 0) := (others => '0');
  signal i_out_w16n       : unsigned(15 downto 0) := (others => '0');

  signal i_synced         : std_logic := '0';
  signal i_bad_bit        : std_logic := '0';

  -- Pipelined
  signal ir_mark_w8a1      : std_logic := '0';
  signal ir_mark_w8a1_2bit : unsigned(1 downto 0) := (others => '0');
  signal ir_out_w8p        : unsigned(7 downto 0) := (others => '0');
  signal ir_out_w8n        : unsigned(7 downto 0) := (others => '0');
  signal ir_out_aux1p      : std_logic := '0';
  signal ir_out_aux1n      : std_logic := '0';

  signal ir_mark_w1        : std_logic := '0';
  signal ir_mark_w1_2bit   : unsigned(1 downto 0) := (others => '0');
  signal ir_out_w1p        : std_logic := '0';
  signal ir_out_w1n        : std_logic := '0';

  signal ir_mark_w16       : std_logic := '0';
  signal ir_out_w16p       : unsigned(15 downto 0) := (others => '0');
  signal ir_out_w16n       : unsigned(15 downto 0) := (others => '0');

  signal ir_synced         : std_logic := '0';
  signal ir_bad_bit        : std_logic := '0';

begin

  inv_bit <= bit_val xor (bit_val'range => invert_signal);

  combined <=
    (combined_prev(2 to 32) & inv_bit(0 to 1)) when (got_bit(0) = '1') else
    (combined_prev(1 to 32) & inv_bit(1     )) when (got_bit(1) = '1') else
    (combined_prev(0 to 32));
  -- 0 is the oldest, 32 is the newest item.

  -- Pipeline stage before the diff calculation
  rataser_ppl(p=>pipeline_diff_in,c=>clk,i=>combined   ,o=>combined_p0);
  rataser_ppl(p=>pipeline_diff_in,c=>clk,i=>got_bit    ,o=>got_bit_p0 );

  -- Even if we align the bits with that 32 is the newest, we must
  -- also calculate the difference of the 1-less shifted pattern, in
  -- case we got 2 bits, and 1 was enough to fill the symbol.
  -- The 0 in diff0 means it refers to the situation when
  -- got_bit(0)/bit_val(0) is the final bit of the pattern.
  d0: entity work.rataser_sym24word8_diff
    generic map(pipeline_1 => pipeline_diff_1,
                pipeline_2 => pipeline_diff_2)
    port map(clk  => clk,
             seq  => combined_p0(8 to 31),
             diff => diff0_p2);
  d1: entity work.rataser_sym24word8_diff
    generic map(pipeline_1 => pipeline_diff_1,
                pipeline_2 => pipeline_diff_2)
    port map(clk  => clk,
             seq  => combined_p0(9 to 32),
             diff => diff1_p2);
  -- The diff values above are delivered (optionally) two pipeline stages
  -- after the input.  Thus p0 for combined and p2 for diffx.

  -- Pipeline stages corresponding to those within the diff calculation.
  rataser_ppl(p=>pipeline_diff_1 ,c=>clk,i=>combined_p0,o=>combined_p1);
  rataser_ppl(p=>pipeline_diff_1 ,c=>clk,i=>got_bit_p0 ,o=>got_bit_p1 );

  rataser_ppl(p=>pipeline_diff_2 ,c=>clk,i=>combined_p1,o=>combined_p2);
  rataser_ppl(p=>pipeline_diff_2 ,c=>clk,i=>got_bit_p1 ,o=>got_bit_p2 );

  -- Pipeline stage after the diff calculation
  rataser_ppl(p=>pipeline_diff_out,c=>clk,i=>combined_p2,o=>combined_p3);
  rataser_ppl(p=>pipeline_diff_out,c=>clk,i=>got_bit_p2 ,o=>got_bit_p3 );
  rataser_ppl(p=>pipeline_diff_out,c=>clk,i=>diff0_p2   ,o=>diff0_p3   );
  rataser_ppl(p=>pipeline_diff_out,c=>clk,i=>diff1_p2   ,o=>diff1_p3   );

  -- For matching, accept one piece of junk.
  -- (The patterns have a design difference of 2 to any other valid
  -- pattern (however shifted), including inverted.)
  match0_p3 <= '1' when (to_integer(diff0_p3) <= 1) else '0';
  match1_p3 <= '1' when (to_integer(diff1_p3) <= 1) else '0';
  -- This just keeps track of the extra (match-allowed) bit being bad.
  badbit0_p3 <= diff0_p3(0);
  badbit1_p3 <= diff1_p3(0);

  -- Pipeline stage after the match calculation, into locked loop.
  rataser_ppl(p=>pipeline_match_out,c=>clk,i=>combined_p3,o=>combined_r);
  rataser_ppl(p=>pipeline_match_out,c=>clk,i=>got_bit_p3 ,o=>got_bit_r );
  rataser_ppl(p=>pipeline_match_out,c=>clk,i=>match0_p3  ,o=>match0_r  );
  rataser_ppl(p=>pipeline_match_out,c=>clk,i=>match1_p3  ,o=>match1_r  );
  rataser_ppl(p=>pipeline_match_out,c=>clk,i=>badbit0_p3 ,o=>badbit0_r );
  rataser_ppl(p=>pipeline_match_out,c=>clk,i=>badbit1_p3 ,o=>badbit1_r );

  -- Did we have any bad bit detected?
  bad_bit0_p <= (not match0_r) or badbit0_r;
  bad_bit1_p <= (not match1_r) or badbit1_r;

  -- Decode the payload data bits.  Normal 8-bit data + 1-bit aux.
  extract_sv(seq => combined_r(8 to 31),
             sp  => w8sp0,
             sn  => w8sn0,
             vp  => a1sp0,
             vn  => a1sn0);
  extract_sv(seq => combined_r(9 to 32),
             sp  => w8sp1,
             sn  => w8sn1,
             vp  => a1sp1,
             vn  => a1sn1);

  -- Single-signal data, always from last three bits.
  extract_1s(combined_r(29 to 31),
             unsigned(count(4 downto 2)),
             sp  => w1sp0,
             sn  => w1sn0);
  extract_1s(combined_r(30 to 32),
             unsigned(count(4 downto 2)),
             sp  => w1sp1,
             sn  => w1sn1);

  -- Direct-coded 16-bit data.
  decode_16_0: for i in 0 to 15 generate
    -- Extract from 0 to 31
    w16sp0(i) <= combined_r(i*2 + 0); -- 0, 2, 4, ..., 30
    w16sn0(i) <= combined_r(i*2 + 1); -- 1, 3, 5, ..., 31
    -- Extract from 1 to 32
    w16sp1(i) <= combined_r(i*2 + 1); -- 1, 3, 5, ..., 31
    w16sn1(i) <= combined_r(i*2 + 2); -- 2, 4, 6, ..., 32
  end generate;

  -- For debugging.
  got0_countlow0 <=
    '1' when (got_bit_r(1) = '0' and count(1 downto 0) = "00") else '0';
  got0_countlow1 <=
    '1' when (got_bit_r(1) = '0' and count(1 downto 0) = "01") else '0';
  got0_countlow2 <=
    '1' when (got_bit_r(1) = '0' and count(1 downto 0) = "10") else '0';
  got0_countlow3 <=
    '1' when (got_bit_r(1) = '0' and count(1 downto 0) = "11") else '0';
  
  got2_countlow0 <=
    '1' when (got_bit_r(0) = '1' and count(1 downto 0) = "00") else '0';
  got2_countlow1 <=
    '1' when (got_bit_r(0) = '1' and count(1 downto 0) = "01") else '0';
  got2_countlow2 <=
    '1' when (got_bit_r(0) = '1' and count(1 downto 0) = "10") else '0';
  got2_countlow3 <=
    '1' when (got_bit_r(0) = '1' and count(1 downto 0) = "11") else '0';
  
  -- Should we add 2, 1 or 0 to the bit-counter?  Add 10 or 01 or 00.
  count_add_got_bits <= (got_bit_r(0) & (got_bit_r(1) and not got_bit_r(0)));
  -- We run the counter for 24 cycles, but counting up to 32, by jumping
  -- past all count % 4 == 3.  This requires that we make an
  -- additional adding in case the current value is 2, or if the current
  -- value is 1 and we are about to add 2.  But only if we actually add.
  count_add_jump <=
    (got_bit_r(1) and not puredata32_pending)
    when ( count(1 downto 0) = "10" or
          (count(1 downto 0) = "01" and got_bit_r(0) = '1')) else '0';
  -- The above is a table with four inputs.  Let's hope the compiler
  -- can figure that out to something efficient.
  -- Do the adding.
  count_next <= ('0' & count) + count_add_got_bits + ("00" & count_add_jump);
  count_wrap  <= count_next(5) xor count(5);
  count_wrap2 <= count_next(2) xor count(2);
  count_wrap7 <= count_next(7);

  -- Keep track of our sync status.
  match_csync: entity work.rataser_sync_count
    port map(clk => clk,
             i   => match_sync.i,
             o   => match_sync.o);

  -- Pipeline stage before the value-pick latches.  To break the dependence
  -- on the count update calculation, and fan-out.
  rataser_ppl(p=>pipeline_pick,c=>clk,i=>got_bit_r  ,o=>got_bit_rr   );
  rataser_ppl(p=>pipeline_pick,c=>clk,i=>bad_bit0_p ,o=>bad_bit0_pr  );
  rataser_ppl(p=>pipeline_pick,c=>clk,i=>bad_bit1_p ,o=>bad_bit1_pr  );
  rataser_ppl(p=>pipeline_pick,c=>clk,i=>count_next ,o=>count_next_r );
  rataser_ppl(p=>pipeline_pick,c=>clk,i=>count_wrap ,o=>count_wrap_r );
  rataser_ppl(p=>pipeline_pick,c=>clk,i=>count_wrap2,o=>count_wrap2_r);
  rataser_ppl(p=>pipeline_pick,c=>clk,
              i=>match_sync.o.synced,o=>match_sync_synced_r);
  rataser_ppl(p=>pipeline_pick,c=>clk,i=>w8sp0      ,o=>w8sp0_r      );
  rataser_ppl(p=>pipeline_pick,c=>clk,i=>w8sp1      ,o=>w8sp1_r      );
  rataser_ppl(p=>pipeline_pick,c=>clk,i=>w8sn0      ,o=>w8sn0_r      );
  rataser_ppl(p=>pipeline_pick,c=>clk,i=>w8sn1      ,o=>w8sn1_r      );
  rataser_ppl(p=>pipeline_pick,c=>clk,i=>a1sp0      ,o=>a1sp0_r      );
  rataser_ppl(p=>pipeline_pick,c=>clk,i=>a1sp1      ,o=>a1sp1_r      );
  rataser_ppl(p=>pipeline_pick,c=>clk,i=>a1sn0      ,o=>a1sn0_r      );
  rataser_ppl(p=>pipeline_pick,c=>clk,i=>a1sn1      ,o=>a1sn1_r      );
  rataser_ppl(p=>pipeline_pick,c=>clk,i=>w1sp0      ,o=>w1sp0_r      );
  rataser_ppl(p=>pipeline_pick,c=>clk,i=>w1sp1      ,o=>w1sp1_r      );
  rataser_ppl(p=>pipeline_pick,c=>clk,i=>w1sn0      ,o=>w1sn0_r      );
  rataser_ppl(p=>pipeline_pick,c=>clk,i=>w1sn1      ,o=>w1sn1_r      );

  process(clk)
  begin
    if (rising_edge(clk)) then
      -- Mark strobes are only set for one cycle
      i_mark_w8a1 <= '0';
      i_mark_w1   <= '0';
      i_mark_w16  <= '0';

      -- Previous value
      combined_prev <= combined;

      -- Always update the count (unless reset below)
      count <= count_next(count'range);

      -- Default no-input to sync-tracking.
      match_sync.i <= rataser_sync_count_in_zero;
      
      -- Note: tentative sync is reported also when we are in sync state.
      if (match_sync.o.tentative_synced = '1') then
        -- If we are currently synced, then we shall match
        -- when the counter wraps.  If not, we are bad.

        if (count_wrap = '1') then
          -- We did wrap.  Did we wrap at the right spot?
          -- Since we wrapped, got_bit_r(1) must be 1.

          -- If we got 2 bits, and we overshot by one bit and prev match.
          -- Or (either got 1 or 2 bits), and did not overshoot and match cur.
          if ((got_bit_r(0) = '1' and
               count_next(0) = '1' and match0_r = '1') or
              (count_next(0) = '0' and match1_r = '1')) then
            -- All is well.
            match_sync.i.up <= '1';
          else
            -- We have mismatched the pattern.
            match_sync.i.down <= '1';
          end if;
        end if;

        if (missed_bit = '1') then
          -- If we miss a bit, we are not going to make it!
          -- (The sync counter however does not like revoking directly,
          -- as that would require a reset also of its counter, so
          -- just degrade the sync status.)
          match_sync.i.down <= '1';
        end if;
      else
        -- If we are not yet in sync, then we restart
        -- the counter when we hit a match.

        if (got_bit_r(1) = '1' and match1_r = '1') then
          -- Used both when we got 1 or 2 bits, last ended the match.
          match_sync.i.tentative <= '1';
          count <= (others => '0');
        elsif (got_bit_r(0) = '1' and match0_r = '1') then
          -- Used when we got 2 bits, but first ended the match.
          match_sync.i.tentative <= '1';
          -- Count starts at 1, since we are one bit ahead already.
          count <= (0 => '1', others => '0');
        end if;

        if (count_wrap7 = '1') then
          -- We have seen what should amount to 4 cycles with no
          -- match.  Try inverted pattern.
          -- (After we switch, it takes one cycle until history
          -- has been inverted.  Then we need to scan for two cycles,
          -- since alignment is unknown.  Thus 3 cycles minimum,
          -- testing for 4 is cheaper.)
          invert_signal <= not invert_signal;
        end if;
      end if;

      -- Instead of an enclosing if-statement to only do the remaining
      -- updates and latches when we are in sync, only set the strobes
      -- and add to the 2-bit counters when synced.  Latch whenever
      -- counter wraps.
      -- This reduces the conditional logic for the latches below.
      if (count_wrap_r = '1') then

        -- Regardless of if we have a match or not, we latch
        -- the extracted symbol word, when in sync and wrapping
        -- the count.

        i_mark_w8a1      <= match_sync_synced_r;
        i_mark_w8a1_2bit <= i_mark_w8a1_2bit + ("0" & match_sync_synced_r);

        -- If we did a double-bit add, and have shifted in one bit
        -- beyond the sequence, then use the less shifted set of values.
        if (got_bit_rr(0) = '1' and count_next_r(0) = '1') then
          i_out_w8p     <= w8sp0_r;
          i_out_w8n     <= w8sn0_r;
          i_out_aux1p   <= a1sp0_r;
          i_out_aux1n   <= a1sn0_r;
        else
          i_out_w8p     <= w8sp1_r;
          i_out_w8n     <= w8sn1_r;
          i_out_aux1p   <= a1sp1_r;
          i_out_aux1n   <= a1sn1_r;
        end if;

        -- Report 1 if we had >= 1 bad bit.
        -- Note: we do not detect all bit flaws, some dual-bit
        -- faults can go undetected.
        if (got_bit_rr(0) = '1' and count_next_r(0) = '1') then
          i_bad_bit <= bad_bit0_pr;
        else
          i_bad_bit <= bad_bit1_pr;
        end if;
      end if;

      -- Extracting the fast updating signal has multiplexed
      -- from the correct location (depending on the count value).
      if (count_wrap2_r = '1') then
        i_mark_w1        <= match_sync_synced_r;
        i_mark_w1_2bit   <= i_mark_w1_2bit + ("0" & match_sync_synced_r);
          
        if (got_bit_rr(0) = '1' and count_next_r(0) = '1') then
          i_out_w1p     <= w1sp0_r;
          i_out_w1n     <= w1sn0_r;
        else
          i_out_w1p     <= w1sp1_r;
          i_out_w1n     <= w1sn1_r;
        end if;
      end if;

      -- TODO: this should be puredata32_pending_r?
      -- But cannot, since it is also being reset.
      -- Likely need to split into pre-cycle logic?
      -- But runs at slow enough clock to not need the _r pipelining?
      if (puredata32_pending = '1') then
        -- We are no longer pending.
        puredata32_pending <= '0';
        
        i_mark_w16       <= match_sync.o.synced;

        if (got_bit_r(0) = '1' and count_next(0) = '1') then
          i_out_w16p    <= w16sp0;
          i_out_w16n    <= w16sn0;
        else
          i_out_w16p    <= w16sp1;
          i_out_w16n    <= w16sn1;
        end if;
      end if;

    end if;
  end process;

  i_synced <= match_sync.o.synced;

  rataser_ppl(p=>pipeline_out,c=>clk,i=>i_mark_w8a1     ,o=>ir_mark_w8a1     );
  rataser_ppl(p=>pipeline_out,c=>clk,i=>i_mark_w8a1_2bit,o=>ir_mark_w8a1_2bit);
  rataser_ppl(p=>pipeline_out,c=>clk,i=>i_out_w8p       ,o=>ir_out_w8p       );
  rataser_ppl(p=>pipeline_out,c=>clk,i=>i_out_w8n       ,o=>ir_out_w8n       );
  rataser_ppl(p=>pipeline_out,c=>clk,i=>i_out_aux1p     ,o=>ir_out_aux1p     );
  rataser_ppl(p=>pipeline_out,c=>clk,i=>i_out_aux1n     ,o=>ir_out_aux1n     );

  rataser_ppl(p=>pipeline_out,c=>clk,i=>i_mark_w1       ,o=>ir_mark_w1       );
  rataser_ppl(p=>pipeline_out,c=>clk,i=>i_mark_w1_2bit  ,o=>ir_mark_w1_2bit  );
  rataser_ppl(p=>pipeline_out,c=>clk,i=>i_out_w1p       ,o=>ir_out_w1p       );
  rataser_ppl(p=>pipeline_out,c=>clk,i=>i_out_w1n       ,o=>ir_out_w1n       );

  rataser_ppl(p=>pipeline_out,c=>clk,i=>i_mark_w16      ,o=>ir_mark_w16      );
  rataser_ppl(p=>pipeline_out,c=>clk,i=>i_out_w16p      ,o=>ir_out_w16p      );
  rataser_ppl(p=>pipeline_out,c=>clk,i=>i_out_w16n      ,o=>ir_out_w16n      );

  rataser_ppl(p=>pipeline_out,c=>clk,i=>i_synced        ,o=>ir_synced        );
  rataser_ppl(p=>pipeline_out,c=>clk,i=>i_bad_bit       ,o=>ir_bad_bit       );

  -- Assign to output signals
  mark_w8a1      <= ir_mark_w8a1;
  mark_w8a1_2bit <= ir_mark_w8a1_2bit;
  out_w8p        <= ir_out_w8p;
  out_w8n        <= ir_out_w8n;
  out_aux1p      <= ir_out_aux1p;
  out_aux1n      <= ir_out_aux1n;

  mark_w1        <= ir_mark_w1;
  mark_w1_2bit   <= ir_mark_w1_2bit;
  out_w1p        <= ir_out_w1p;
  out_w1n        <= ir_out_w1n;

  mark_w16       <= ir_mark_w16;
  out_w16p       <= ir_out_w16p;
  out_w16n       <= ir_out_w16n;
  
  synced         <= ir_synced;
  bad_bit        <= ir_bad_bit;
  
end RTL;
