-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.rataser_records.all;

package rataser_sym24word8_util is

  -- These constants were in a record (declared in rataser_records),
  -- but various synthesis tools had different issues with it.
  -- Therefore they are now declared here instead, even though they
  -- logically really belong together.  Since we do not intend to
  -- change symbol layout, this is not so bad.

  type rataser_int_array is array (integer range <>) of integer;

  -- Slots for 8 bit values, positive and inverted
  constant rataser_sym24_sp : rataser_int_array(0 to 7) :=
    (0=> 1, 1=> 3, 2=> 7, 3=> 9, 4=>13, 5=>15, 6=>19, 7=>21 );
  constant rataser_sym24_sn : rataser_int_array(0 to 7) :=
    (0=> 2, 1=> 4, 2=> 8, 3=>10, 4=>14, 5=>16, 6=>20, 7=>22 );
  -- Slots for 1 aux value, positive and inverted
  constant rataser_sym24_vp : integer :=  6;
  constant rataser_sym24_vn : integer := 17;
  -- Slots for 3 fixed 0 and 1 values
  constant rataser_sym24_f0 : rataser_int_array(0 to 2) :=(0=>5, 1=>11, 2=>18);
  constant rataser_sym24_f1 : rataser_int_array(0 to 2) :=(0=>0, 1=>12, 2=>23);

  -- Leaving the record initialisation as a comment.
  -- constant sym : rataser_sym24def :=
  --   (sp => (0=> 1, 1=> 3, 2=> 7, 3=> 9, 4=>13, 5=>15, 6=>19, 7=>21 ),
  --    sn => (0=> 2, 1=> 4, 2=> 8, 3=>10, 4=>14, 5=>16, 6=>20, 7=>22 ),
  --    vp =>                    6,
  --    vn =>                                        17,
  --    f0 => (       0=> 5,        1=>11,               2=>18        ),
  --    f1 => (0=> 0,                      1=>12,               2=>23));

  -- 1aAbB0vcCdD01eEfFV0gGhH1

  -- 0         1         2
  -- 012345678901234567890123
  -- 1aAbB0vcCdD01eEfFV0gGhH1
  -- 1....0-....01....-0....1

  --  a b  vc d   e f   g h
  -- 101010001010101011001011 -- pattern with all data and aux=0

  procedure extract_sv (signal seq : in std_logic_vector(0 to 23);
                        signal sp  : out unsigned(7 downto 0);
                        signal sn  : out unsigned(7 downto 0);
                        signal vp  : out std_logic;
                        signal vn  : out std_logic);

  procedure extract_1s(signal seq : in std_logic_vector(0 to 2);
                       offset     : in unsigned(2 downto 0);
                       signal sp  : out std_logic;
                       signal sn  : out std_logic);

  procedure gen_sequence(signal seq  : out std_logic_vector(0 to 23);
                         signal data : in unsigned(7 downto 0);
                         signal aux  : in std_logic);

  procedure gen_sequence32(signal seq32 : out std_logic_vector(0 to 31);
                           signal seq24 : in  std_logic_vector(0 to 23));

end rataser_sym24word8_util;

package body rataser_sym24word8_util is

  -- Extract the embedded signal and aux bits,
  -- both positive and negative versions.
  procedure extract_sv (signal seq : in std_logic_vector(0 to 23);
                        signal sp  : out unsigned(7 downto 0);
                        signal sn  : out unsigned(7 downto 0);
                        signal vp  : out std_logic;
                        signal vn  : out std_logic) is

    variable tmp_sp : unsigned(7 downto 0);
    variable tmp_sn : unsigned(7 downto 0);
    variable tmp_vp : std_logic;
    variable tmp_vn : std_logic;
  begin
    for i in 0 to 7 loop
      tmp_sp(i) := seq(rataser_sym24_sp(i));
      tmp_sn(i) := seq(rataser_sym24_sn(i));
    end loop;
    tmp_vp := seq(rataser_sym24_vp);
    tmp_vn := seq(rataser_sym24_vn);

    sp <= tmp_sp;
    sn <= tmp_sn;
    vp <= tmp_vp;
    vn <= tmp_vn;
  end procedure;

  -- Extract one pair of signal bits (positive and negative)
  -- from the 3 most recent (given by offset) raw bit slots.
  procedure extract_1s(signal seq : in std_logic_vector(0 to 2);
                       offset     : in unsigned(2 downto 0);
                       signal sp  : out std_logic;
                       signal sn  : out std_logic) is

    variable int_offset  : integer;
    -- variable int_offset0 : integer;
    variable tmp_sp : std_logic;
    variable tmp_sn : std_logic;
  begin
    int_offset  := to_integer(offset);
    -- int_offset0 := to_integer(offset(0));

    -- TODO: Can try with a much simpler picker thanks to the simple
    -- pattern (but then not using rataser_sym24_sp/sn).
    -- For even slot, pick bits 1 and 2, for odd pick bits 0 and 1.
    -- tmp_sp := seq(1 - int_offset0);
    -- tmp_sn := seq(2 - int_offset0);
    tmp_sp := seq(rataser_sym24_sp(int_offset) - 3 * int_offset);
    tmp_sn := seq(rataser_sym24_sn(int_offset) - 3 * int_offset);

    sp <= tmp_sp;
    sn <= tmp_sn;
  end procedure;
    
  procedure gen_sequence(signal seq  : out std_logic_vector(0 to 23);
                         signal data : in unsigned(7 downto 0);
                         signal aux  : in std_logic) is
    variable tmp_seq : std_logic_vector(0 to 23);
  begin
    for i in 0 to 7 loop
      tmp_seq(rataser_sym24_sp(i)) :=     data(i);
      tmp_seq(rataser_sym24_sn(i)) := not data(i);
    end loop;
    tmp_seq(rataser_sym24_vp) :=     aux;
    tmp_seq(rataser_sym24_vn) := not aux;
    for i in 0 to 2 loop
      tmp_seq(rataser_sym24_f0(i)) := '0';
      tmp_seq(rataser_sym24_f1(i)) := '1';
    end loop;
    seq <= tmp_seq;
  end procedure;

  procedure gen_sequence32(signal seq32 : out std_logic_vector(0 to 31);
                           signal seq24 : in  std_logic_vector(0 to 23)) is
    variable tmp_seq32 : std_logic_vector(0 to 31);
  begin
    for i in 0 to 7 loop
      tmp_seq32(i*4+0) := seq24(i*3+0);
      tmp_seq32(i*4+1) := seq24(i*3+1);
      tmp_seq32(i*4+2) := seq24(i*3+2);
      tmp_seq32(i*4+3) := seq24(i*3+2); -- Note: never used, set as
      -- duplicate of previous to possibly simplify later multiplexer.
    end loop;
    seq32 <= tmp_seq32;
  end procedure;
  
end;

