-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.rataser_records.all;

entity rataser_sym8word1_decode is
  generic(-- Number of bits for pulse period.
          period_bits: integer
          );
    port (clk:      in std_logic;

          -- Input samples (4 samples each clock cycle), 0 oldest, 3 newest
          raw_val:  in std_logic_vector(0 to 7);

          -- Pulse period (in clk cycles).
          pulse_period_clks: in std_logic_vector(period_bits-1 downto 0);

          -- Where is the edge expected?  (0-3)
          expect_edge:   in std_logic_vector(1 downto 0);
          -- Sequence of edges seen most recently.  (For monitoring/setup.)
          -- Two bits for each edge, most recent in 1 downto 0.
          last_edges:    out std_logic_vector;

          -- Continuous signal telling if we are good.
          synced:   out std_logic;
          -- The period tracker did not find a correct edge.
          bad_period: out std_logic;
          -- The 8-slot identifier found a change in 8-slot index.
          bad_8slot: out std_logic;

          bit_val:  out std_logic_vector(0 to 1);
          got_bit:  out std_logic_vector(0 to 1);
          missed_bit: out std_logic;

          -- The filtered min pulse length (4x clock cyles) (for debugging).
          pulse_length_min     : out std_logic_vector;
          -- The filtered max pulse length (4x clock cyles) (for debugging).
          pulse_length_max     : out std_logic_vector;
          -- The filtered pulse group of 6 length difference (4x clock cyles).
          pulse_length_diff6   : out std_logic_vector;

          debug:    out rataser_sym8word1_debug
          );
end;

-- We receive a stream of bits.  Each symbol consist of an 8-slot
-- pattern of 10QqQqQq.  q is the payload bit, and Q its inverse.
-- We need to find out where the pattern starts.
--
-- The first thing is the edge tracker, i.e. the mechanism
-- by which we decide which clock cycles contain the starting
-- edge.  This occurs periodically, and we are only willing
-- to accept a 25% phase shift.  I.e. it must occur within the
-- expected period, +/- one phase movement.  Note that the input
-- signal always has a pulse, such that edges always shall
-- be present.
--
-- The length of each pulse is given as a count of the number of
-- high samples, starting at one sample before the nominial, and
-- continuing as far as samples are high.  Gaps of single samples
-- low are passed (but not counted).  Two low samples stop the
-- counting.
--
-- Note that all six QqQqQq slots are summed up, with the contribution
-- if the inverse slots subtracted.
--
-- The difference between a long and short sample is determined by
-- an average of repeated min-max and max-min filters of the smallest
-- and largest pulse length counts.  These run over a large number
-- of pulses.  The payload pattern guarantees that every five bits,
-- at least one short and one long pattern is present.  (At most
-- four bits in a row are the same.)
--
-- To determine which of any 8 consequtive slots is the start slot,
-- for each option (i.e. 8 slots), a sum is calculated of 24 such
-- slots, i.e., 0, 8, 16, 24, 32, ..., 184 for slot 0, and
-- 1, 9, 17, 25, 33, ..., 185 for slot 1 and so on.  Since the 24-bit
-- payload pattern is DC-balanced, the maximum imbalance that can
-- occur is that the partial group end of three bits at the beginning
-- of such an 24-slot sequence does not have the same payload bit as
-- the corresponding partial group beginning at the end of the
-- 24-slot sequence.  The negative contribution of the next slot
-- (slot + 1) % 8 is also included in each sum.
--
-- What is searched for is however the 10 group.  That group will
-- have the largest contribution since they all add up.  The 0Q and
-- q1 groups will have small contributions due to the 0 and 1 slots,
-- and average contributions from the Q and q.  Also all qQ and Qq
-- will have average contributions from the payload bits.

-- More specifically (worst case, payload flip):
--
-- Rewritten with L=A+E and S=A-E, where A is the average of L and S,
-- and E the excess for a long pulse above the average.
--
-- 10:    24 long             - 24 short           =  24 L - 24 S =  48 E
-- 0Q:    24 short            - 11 long - 13 short = -11 L + 11 S = -22 E
-- q1:    13 long  + 11 short - 24 long            = -11 L + 11 S = -22 E
-- qQ/Qq: 25 long             - 23 short           =  25 L - 23 S =   2 E
--
-- As can be seen, there is a very large margin for a decision
-- for which offset of 8 has the largest value, i.e. the 10 pattern.

architecture RTL of rataser_sym8word1_decode is

  signal prev_old_val : std_logic_vector(0 to 3)  := (others => '0');
  signal combined     : std_logic_vector(0 to 11) := (others => '0');

  -- Pipelined:
  signal combined_r  : std_logic_vector(combined'range) := (others => '0');
  signal combined_rr: std_logic_vector(combined'range) := (others => '0');
  
  -- Do we have an edge?
  signal has_edge : std_logic_vector(0 to 5) := (others => '0');

  -- For monitoring of the prevailing edge:

  -- Edge seen this cycle?
  signal mon_has_edge : std_logic := '0';
  -- Which edge (if any) was seen this cycle (quadrant code).
  signal mon_edge : std_logic_vector(1 downto 0) := (others => '0');
  -- Pipelined:
  signal mon_has_edge_r : std_logic := '0';
  signal mon_edge_r : std_logic_vector(1 downto 0) := (others => '0');
  -- Which edge was seen most recently (quadrant code).
  signal mon_last_edge : std_logic_vector(1 downto 0) := (others => '0');
  -- Shift register of most recently seen edges.
  signal i_last_edges : std_logic_vector(last_edges'range) := (others => '0');

  -- Is there an edge, that matches with expectation?
  signal good_edge : std_logic := '0';

  -- Pipelined:
  signal good_edge_r  : std_logic := '0';
  signal good_edge_rr : std_logic := '0';

  -- State of in-sync-tracking counter.
  signal period_sync : rataser_sync_count_iface :=
    rataser_sync_count_iface_zero;

  -- Which raw samples contribute to pulse-high count?
  signal cnt_hs : std_logic_vector(0 to 3) := (others => '0');
  -- Carry-over from last round towards contribution.
  -- We only continue if bits were high in last round.
  signal cnt_hs_last : std_logic_vector(2 to 3) := (others => '0');
  -- Count of number of high samples.
  signal samples_high  : unsigned(2 downto 0) := (others => '0');

  -- Rearranges samples at the start of a pulse, to simplify counting.
  signal start_hs : std_logic_vector(0 to 6) := (others => '0');
  -- Which raw samples contribute to the start pulse-high count?
  signal start_cnt_hs   : std_logic_vector(0 to 6) := (others => '0');
  -- Pipelined:
  signal start_cnt_hs_r : std_logic_vector(0 to 6) := (others => '0');
  -- Counted samples to carry over for next decision.
  signal start_cnt_hs_carry : std_logic_vector(2 to 3) := (others => '0');
  -- Count of number of start pulse high samples.
  signal start_samples_high : unsigned(2 downto 0) := (others => '0');

  -- Bit-extended number to allow easy creation of negated number.
  signal extend_samples_high       : unsigned(8+2 downto 0) := (others => '0');
  signal extend_start_samples_high : unsigned(8+2 downto 0) := (others => '0');

  -- Negated numbers
  signal extend_samples_high_xor       :
    unsigned(8+2 downto 0) := (others =>'0');
  signal extend_start_samples_high_xor :
    unsigned(8+2 downto 0) := (others =>'0');

  -- Pipelined;
  signal extend_samples_high_r :
    unsigned(extend_samples_high'range) := (others =>'0');
  signal extend_samples_high_xor_r :
    unsigned(extend_samples_high_xor'range) := (others =>'0');
  signal extend_start_samples_high_r :
    unsigned(extend_start_samples_high'range) := (others =>'0');
  signal extend_start_samples_high_xor_r :
    unsigned(extend_start_samples_high_xor'range) := (others =>'0');

  -- How much high bits have we seen for this pulse group (of 6)?
  signal accum_high : unsigned(8+2 downto 0) := (others => '0');

  -- Filter for how many or few samples we have seen over many pulses.
  signal accum_min : unsigned(8+2 downto 0) := (others => '0');
  signal accum_max : unsigned(8+2 downto 0) := (others => '0');

  -- Second-stage filter (to avoid outliers).
  signal accum_min_max : unsigned(8+2 downto 0) := (others => '0');
  signal accum_max_min : unsigned(8+2 downto 0) := (others => '0');

  -- How much high bits have we seen for this pulse (for debug)?
  signal sum_high : unsigned(5+2 downto 0) := (others => '0');

  -- Filter for how many or few samples we have seen over many pulses.
  signal sum_min : unsigned(5+2 downto 0) := (others => '0');
  signal sum_max : unsigned(5+2 downto 0) := (others => '0');

  -- Filter result, useful for debugging.
  signal debug_min : unsigned(5+2 downto 0) := (others => '0');
  signal debug_max : unsigned(5+2 downto 0) := (others => '0');

  -- Average of the second-stage filter.
  signal accum_avg : unsigned(8+2 downto 0) := (others => '0');
  -- Temporary value for the calculation.
  signal accum_avg_calc : unsigned(8+2+1 downto 0) := (others => '0');

  -- Difference of the second-stage filter.
  signal debug_diff : unsigned(8+2 downto 0) := (others => '0');
  -- Temporary value for the calculation.
  signal accum_diff_calc : unsigned(8+2 downto 0) := (others => '0');

  -- Cumulative high-low bit count for the 8 pulses in a 24-sequence.
  -- These are used to figure out which of the 8 pulses is the start pulse.
  -- Could be array (but this allows easy gtkwave view).
  signal long_accum_high0 : unsigned(8+2 downto 0) := (others => '0');
  signal long_accum_high1 : unsigned(8+2 downto 0) := (others => '0');
  signal long_accum_high2 : unsigned(8+2 downto 0) := (others => '0');
  signal long_accum_high3 : unsigned(8+2 downto 0) := (others => '0');
  signal long_accum_high4 : unsigned(8+2 downto 0) := (others => '0');
  signal long_accum_high5 : unsigned(8+2 downto 0) := (others => '0');
  signal long_accum_high6 : unsigned(8+2 downto 0) := (others => '0');
  signal long_accum_high7 : unsigned(8+2 downto 0) := (others => '0');

  -- The maximum high-low bit count seen for any of the 8 shifted starts.
  signal long_accum_high_max : unsigned(8+2 downto 0) := (others => '0');
  -- Which of the 8 shifted starts is the max so far.
  signal long_accum_high_ind : unsigned(2 downto 0) := (others => '0');

  -- Which of the 8 shifted starts is the current selection?
  signal long_accum_ind      : unsigned(2 downto 0) := (others => '0');

  -- Time to make a new decision on which 8-shift to use as start.
  signal pick_long_accum_ind : std_logic := '0';

  -- Counter for cycling through 24*8 pulses, to sample an equal
  -- number of long and short pulses.
  signal long_count      : unsigned(5+3-1 downto 0) := (others => '0');
  signal long_count_next : unsigned(5+3   downto 0) := (others => '0');
  signal long_count_wrap : std_logic := '0';

  -- Prepare values, for min-max update.
  signal long_count_below_8    : std_logic := '0';
  signal long_count_mod64_0to7 : std_logic := '0';

  -- Should the accum_high value be cleared?
  signal accum_high_clear    : std_logic_vector(2 downto 0) := "000";
  -- Add or subtract the number of high slots seen?
  signal accum_high_xor      : std_logic := '0';

  -- When we clear, how should we update the min-max values?
  signal accum_high_min_max_long_below_8    : std_logic := '0';
  signal accum_high_min_max_long_mod64_0to7 : std_logic := '0';

  -- Normalise inputs to downto ranges.
  signal downto_pulse_n    : std_logic_vector(period_bits-1 downto 0);

  -- Counter for pulse period.
  signal count_period      : unsigned(period_bits-1 downto 0) := (others=>'0');
  signal count_period_next : unsigned(period_bits   downto 0) := (others=>'0');
  signal count_period_wrap : std_logic := '0';

  signal count_period_max  : unsigned(period_bits-1 downto 0) := (others=>'0');

  -- Pipelined:
  signal count_period_wrap_r : std_logic := '0';

  -- For debugging.
  signal start_slot : std_logic := '0';

  -- Output values.
  signal i_bit_val : std_logic := '0';
  signal i_got_bit : std_logic := '0';
  signal i_missed_bit : std_logic := '0';

  signal i_synced  : std_logic := '0';

  signal i_bad_period  : std_logic := '0';
  signal i_bad_8slot  : std_logic := '0';

  signal i_debug : rataser_sym8word1_debug :=
    (accum_min_max  => (others => '0'),
     accum_max_min  => (others => '0'),
     accum_avg      => (others => '0'),
     accum_this     => (others => '0'),
     accum_high_max => (others => '0'),
     accum_high_ind => (others => '0'));

begin
  -- Pipeline already the input data.
  process (clk)
  begin
    if (rising_edge(clk)) then
      -- Include the new data values with the last of the previous.
      combined <= prev_old_val & raw_val;
      -- Latch (oldest) raw samples for next cycle.
      prev_old_val <= raw_val(0 to 3);
    end if;
  end process;

  -- Calculate if an edge is present.
  -- combined       0123456789ab
  has_edge(0) <= -- ...0011.....
        '1' when (combined(3 to  6) = "0011") else '0';
  has_edge(1) <= -- ....0011....
        '1' when (combined(4 to  7) = "0011") else '0';
  has_edge(2) <= -- .....0011...
        '1' when (combined(5 to  8) = "0011") else '0';
  has_edge(3) <= -- ......0011..
        '1' when (combined(6 to  9) = "0011") else '0';
  has_edge(4) <= -- .......0011.
        '1' when (combined(7 to 10) = "0011") else '0';
  has_edge(5) <= -- ........0011
        '1' when (combined(8 to 11) = "0011") else '0';

  -- Note: several of has_edge can be set!
  -- (But with a nicely looking signal, they will not be.)
  -- Not any however - only 0 is independent of 4 and 5, and 1 is
  -- independent of 5.  But none of 1, 2, 3, 4 can be set at the same
  -- time.

  -- This is for monitoring:
  mon_has_edge <= has_edge(1) or has_edge(2) or has_edge(3) or has_edge(4);
  -- 1 -> "00", 2 -> "01", 3 -> "10", 4 -> "11"
  mon_edge(0)  <= has_edge(2) or has_edge(4);
  mon_edge(1)  <= has_edge(3) or has_edge(4);

  -- For reception:

  -- expect_edge is one of 0, 1, 2, 3.  This is set by the user,
  -- based on statistics that we deliver on which edge occurs
  -- most often.  Since the input signal is the same that drives
  -- the clock PLL, it should be phase-stable and not move around.

  -- With 0, we expect has_edge at 0, 1, or 2.
  -- With 1, we expect has_edge at 1, 2, or 3.
  -- With 2, we expect has_edge at 2, 3, or 4.
  -- With 3, we expect has_edge at 3, 4, or 5.

  good_edge <=
    '1' when
    ((expect_edge = "00" and has_edge(0 to 2) /= "000") or
     (expect_edge = "01" and has_edge(1 to 3) /= "000") or
     (expect_edge = "10" and has_edge(2 to 4) /= "000") or
     (expect_edge = "11" and has_edge(3 to 5) /= "000"))
    else '0';

  -- Pipeline.
  process (clk)
  begin
    if (rising_edge(clk)) then
      good_edge_r <= good_edge;
      combined_r  <= combined;

      -- Second stage
      good_edge_rr   <= good_edge_r;
      combined_rr    <= combined_r;
      start_cnt_hs_r <= start_cnt_hs;
    end if;
  end process;

  -- Note: cnt_hs already runs directly from the second stage pipeline.
  -- This is needed, since the cnt_hs_last feedback is present.
  -- The start_cnt_hs however (which is more complicated, but has
  -- no feedback input) can be pipelined (start_cnt_hs_r) and therefore
  -- uses combined_r as input.

  -- To determine if a pulse is long or short, we count the number of
  -- samples that are high.

  -- Which samples to count?
  -- Once we are within the pulse, we count as long as the pulse is high,
  -- and allow single missing slots.
  cnt_hs(0) <= combined_rr( 8) and (cnt_hs_last(2) or cnt_hs_last(3));
  cnt_hs(1) <= combined_rr( 9) and (cnt_hs_last(3) or cnt_hs(0));
  cnt_hs(2) <= combined_rr(10) and (cnt_hs(0)      or cnt_hs(1));
  cnt_hs(3) <= combined_rr(11) and (cnt_hs(1)      or cnt_hs(2));

  -- Evaluate the number of high samples.
  samples_high <= -- Add in groups of (1+1+1) + 1
    (unsigned'("0" & cnt_hs(0)) +
     unsigned'("0" & cnt_hs(1)) +
     unsigned'("0" & cnt_hs(2))) +
    unsigned'("00" & cnt_hs(3));

  -- For the first slot, the samples to include in the count
  -- depend on where the edge is _supposed_ to be.  To be fair,
  -- we always start to count in the slot before the nominal slot,
  -- i.e. where an allowed edge can be.
  start_hs <=
    combined_r(5 to 11)         when expect_edge = "00" else
    combined_r(6 to 11) & "0"   when expect_edge = "01" else
    combined_r(7 to 11) & "00"  when expect_edge = "10" else
    combined_r(8 to 11) & "000"; --  expect_edge = "11"

  -- The above has shifted the bits according to this pattern:
  -- 0 (00) .....xxxxxxx    -> xxxxxxx
  -- 1 (01) ......xxxxxx0   -> xxxxxx0
  -- 2 (10) .......xxxxx00  -> xxxxx00
  -- 3 (11) ........xxxx000 -> xxxx000

  -- Example: for expect_edge = 3 (11), the four last bits (xxxx)
  -- should be either of 1111, 0111 or 0011.  (0111 is the nominal.)

  -- The logic becomes rather simple.  We start by accepting any bits
  -- within the start region (except not (0) unless (1) or (2) is also
  -- set, and (1) requires (2) or (3) - both to disallow double-lone
  -- bits).

  -- Truth table for four first bits of start_hs -> start_cnt_hs:
  -- I.e. bits remain if they are to be counted.
  -- 0000 -> 0000  ;  0100 -> 0000  ;  1000 -> 0000   ;  1100 -> 1100 (*)
  -- 0001 -> 0001  ;  0101 -> 0101  ;  1001 -> 0001 # ;  1101 -> 1101
  -- 0010 -> 0010  ;  0110 -> 0110  ;  1010 -> 1010   ;  1110 -> 1110
  -- 0011 -> 0011  ;  0111 -> 0111  ;  1011 -> 1011   ;  1111 -> 1111
  --
  -- #: a bit doubtful if this should be counted.  But since it is
  -- the same for all / independent of expect_edge it is ok.  Just
  -- means that the pulse is extra late.

  start_cnt_hs(0) <= start_hs(0) and (start_hs(1) or start_hs(2));
  start_cnt_hs(1) <= start_hs(1) and (start_hs(2) or start_hs(3) or
                                      start_hs(0)); -- last for case (*) above
  start_cnt_hs(2) <= start_hs(2);
  start_cnt_hs(3) <= start_hs(3);
  -- The last three bits of hs are counted in the usual way, i.e.
  -- as long as any of the two previous were counted.
  start_cnt_hs(4) <= start_hs(4) and (start_cnt_hs(2) or start_cnt_hs(3));
  start_cnt_hs(5) <= start_hs(5) and (start_cnt_hs(3) or start_cnt_hs(4));
  start_cnt_hs(6) <= start_hs(6) and (start_cnt_hs(4) or start_cnt_hs(5));

  -- Evaluate the number of high samples.
  start_samples_high <= -- Add in groups of (1+1+1) + (1+1+1) + 1
    ((unsigned'("0") &
     (unsigned'("0" & start_cnt_hs_r(0)) +
      unsigned'("0" & start_cnt_hs_r(1)) +
      unsigned'("0" & start_cnt_hs_r(2)))) +
    (unsigned'("0") &
     (unsigned'("0" & start_cnt_hs_r(3)) +
      unsigned'("0" & start_cnt_hs_r(4)) +
      unsigned'("0" & start_cnt_hs_r(5)))) +
    unsigned'("00" & start_cnt_hs_r(6)));

  -- Prepare carry-over samples for next round.  This adjusts back for
  -- start_hs being shifted.  I.e. start_cnt_hs_carry is always from
  -- combined_r(10 to 11).
  start_cnt_hs_carry <=
    start_cnt_hs_r(5 to 6) when expect_edge = "00" else
    start_cnt_hs_r(4 to 5) when expect_edge = "01" else
    start_cnt_hs_r(3 to 4) when expect_edge = "10" else
    start_cnt_hs_r(2 to 3); --  expect_edge = "11"

  -- Extend.  To allow easy conversion to a negative number.
  extend_start_samples_high <= (7 downto 0 => '0') & start_samples_high;
  extend_samples_high       <= (7 downto 0 => '0') & samples_high;

  -- Negative of the contributions.  Actually, -x-1, but that is ok,
  -- since we always will add the same amount of negative numbers,
  -- i.e. just biasing the result.
  extend_start_samples_high_xor <=
    (extend_start_samples_high_xor'range => not accum_high_xor) xor
    extend_start_samples_high;
  extend_samples_high_xor <=
    (extend_samples_high_xor'range => accum_high_xor) xor
    extend_samples_high;

  -- Alias:
  downto_pulse_n <= pulse_period_clks;

  -- Maximum value (start) of pulse counter is pulse period - 1.
  count_period_max <= unsigned(downto_pulse_n) - 1;

  -- Counter for the pulse itself (down-count).
  count_period_next <= ('0' & count_period) - 1;
  count_period_wrap <= count_period_next(count_period_next'high);

  -- Count of number of pulses within a full 8x24 slot message.
  long_count_next <= ("0" & long_count) - 1;
  long_count_wrap <= long_count_next(long_count_next'high);

  -- Prepare values
  long_count_below_8    <= '1' when (long_count < 8            ) else '0';
  long_count_mod64_0to7 <= '1' when (long_count(5 downto 3) = 0) else '0';

  -- Calculate the mid-point of the min and max accumulated sample counts.
  accum_avg_calc <= (('0' & accum_max_min) + accum_min_max);

  -- Calculate the difference (for debugging).
  accum_diff_calc <= accum_max_min - accum_min_max;

  -- Keep track of our sync status.
  period_csync: entity work.rataser_sync_count
    port map(clk => clk,
             i   => period_sync.i,
             o   => period_sync.o);

  process (clk)
  begin
    if (rising_edge(clk)) then
      -- For monitoring of the prevailing edge:
      mon_has_edge_r <= mon_has_edge;
      mon_edge_r     <= mon_edge;

      if (mon_has_edge_r = '1') then
        mon_last_edge <= mon_edge_r;

        i_last_edges(1 downto 0) <= mon_edge_r;
        i_last_edges(last_edges'left downto 2) <=
          i_last_edges(last_edges'left-2 downto 0);
      end if;

      -- Default is no output.
      i_got_bit <= '0';
      i_missed_bit <= '0';

      -- Always update the count (unless reset further down).
      count_period <= count_period_next(count_period'range);
      if (count_period_wrap = '1') then
        count_period <= count_period_max;
      end if;

      -- Default input to sync-tracker.
      period_sync.i <= rataser_sync_count_in_zero;
      -- Default:
      i_bad_period <= '0';

      -- Note: tentative sync is reported also when we are in sync state.
      if (period_sync.o.tentative_synced = '1') then
        -- If we are currently synced, then we shall match
        -- when the counter wraps.  If not, we are bad.
        if (count_period_wrap = '1') then
          if (good_edge_rr = '1') then
            period_sync.i.up <= '1';
          else
            period_sync.i.down <= '1';
            i_bad_period <= '1';
          end if;
        end if;
      else
        -- If we are not yet in sync, then we restart
        -- the counter when we hit a match.
        if (good_edge_rr = '1') then
          period_sync.i.tentative <= '1';
          count_period <= count_period_max;
        end if;
      end if;

      -- When we wrap the pulse period counter, many updates happen!
      if (count_period_wrap = '1') then
        -- The carry-over samples which decide what to count in
        -- the next cycle.
        cnt_hs_last <= start_cnt_hs_carry;
      else
        cnt_hs_last <= cnt_hs(2 to 3);
      end if;

      -- Pipeline:
      count_period_wrap_r             <= count_period_wrap;
      extend_samples_high_r           <= extend_samples_high;
      extend_samples_high_xor_r       <= extend_samples_high_xor;
      extend_start_samples_high_r     <= extend_start_samples_high;
      extend_start_samples_high_xor_r <= extend_start_samples_high_xor;

      -- Defaults:
      pick_long_accum_ind <= '0';
      start_slot <= '0';

      -- We run the 8-fold logic with input values that have been
      -- pipelined one stage.  This decouples, and also reduces the
      -- large fanout of count_period_wrap.
      if (count_period_wrap_r = '1') then
        -- Accumulate high samples count.
        accum_high  <= accum_high + extend_start_samples_high_xor_r;

        -- Figure out how long the pulse is (for debugging).
        sum_high <= (others => '0');
        sum_high(start_samples_high'range) <=
          extend_start_samples_high_r(start_samples_high'range);

        -- Shift the long counters, and accumulate.
        long_accum_high0 <=
          long_accum_high7 +      extend_start_samples_high_r;
        -- Subtract from the next one.
        long_accum_high1 <=
          long_accum_high0 + (not extend_start_samples_high_r);
        -- Shift the data of the long counters around.
        long_accum_high2      <= long_accum_high1;
        long_accum_high3      <= long_accum_high2;
        long_accum_high4      <= long_accum_high3;
        long_accum_high5      <= long_accum_high4;
        long_accum_high6      <= long_accum_high5;
        long_accum_high7      <= long_accum_high6;

        -- Down-count of how many pulses we have seen.
        long_count <= long_count_next(long_count'range);

        -- We always compare to find the maximum each round.
        -- Reset (i.e. always take value) whenever we start at number 7.
        -- Thus the remaining possibilities will be checked
        -- at numbers 6 to 0.
        if (long_accum_high6 > long_accum_high_max or
            long_count(2 downto 0) = 7) then
          long_accum_high_max <= long_accum_high6;
          long_accum_high_ind <= long_count(2 downto 0);
        end if;

        -- Clear the long counters on the last round.
        -- Clear = set to mid-value, such that subtraction does not wrap.
        if (long_count < 8) then
          long_accum_high7 <= (long_accum_high7'high => '1', others => '0');
        end if;

        -- When we have seen a full cycle of 24 * 8 pulses,
        -- pick the largest value seen so far.
        if (long_count_wrap = '1') then
          -- The actual index must be picked the next cycle,
          -- since the comparison above might have selected this cycle.
          pick_long_accum_ind <= '1';
          -- Reset the long cycle counter.
          long_count <= to_unsigned(191, long_count'length); -- 24 * 8 = 192
        end if;

        -- Delay the clear of the accumulator by two wraps.
        accum_high_clear <= "0" & accum_high_clear(2 downto 1);
        -- Invert if we add or subtract.
        accum_high_xor <= not accum_high_xor;

        -- Are we at the pulse currently chosen by the long counter
        -- pulse investigation?
        if (long_count(2 downto 0) = long_accum_ind) then
          accum_high_clear <= "100"; -- Initiate the clear,
                                     -- to happen in two wraps.
          accum_high_xor <= '0';     -- Initiate which pulses are subtracted.
          -- Keep info for min-max recording.
          -- Cannot do in delayed routine, since long_count values
          -- no longer are in expected ranges.
          accum_high_min_max_long_below_8    <= long_count_below_8;
          accum_high_min_max_long_mod64_0to7 <= long_count_mod64_0to7;
        end if;

        -- Bit decision on cycle after we are at the spot marked.
        if (accum_high_clear(2) = '1') then
          start_slot <= '1';        -- For debugging

          -- The output bit is 1 if the number of samples was higher
          -- than the average of the min and max values seen.
          if (accum_high > accum_avg) then
            i_bit_val <= '1';
          else
            i_bit_val <= '0';
          end if;
          -- We only report bits if we think we have the edge synced.
          i_got_bit <= period_sync.o.synced;
          -- If we do not think we are in sync, then we have just lost.
          i_missed_bit <= not period_sync.o.synced;

          i_debug.accum_this(accum_high'range) <=
            std_logic_vector(accum_high);

          -- Filter the values, reset every eight rounds.
          -- Since the protocol may have 4 bits same in a row, it is
          -- necessary to look at more than 4 bits!
          if (accum_high > accum_max or
              accum_high_min_max_long_mod64_0to7 = '1') then
            accum_max <= accum_high;
          end if;
          if (accum_high < accum_min or
              accum_high_min_max_long_mod64_0to7 = '1') then
            accum_min <= accum_high;
          end if;

          -- Second-stage filter every eight rounds, reset every 24.
          if ((accum_min > accum_min_max and
               accum_high_min_max_long_mod64_0to7 = '1') or
              accum_high_min_max_long_below_8 = '1') then
            accum_min_max <= accum_min;
          end if;
          if ((accum_max < accum_max_min and
               accum_high_min_max_long_mod64_0to7 = '1') or
              accum_high_min_max_long_below_8 = '1') then
            accum_max_min <= accum_max;
          end if;

          -- For the debug of pulse length, we just do a simple filter.
          -- User can multi-sample.
          if (sum_high > sum_max or
              accum_high_min_max_long_below_8 = '1') then
            sum_max <= sum_high;
          end if;
          if (sum_high < sum_min or
              accum_high_min_max_long_below_8 = '1') then
            sum_min <= sum_high;
          end if;

          -- Update the average every 24 pulses.
          if (accum_high_min_max_long_below_8 = '1') then
            accum_avg <= accum_avg_calc(accum_avg_calc'left downto 1);

            debug_diff <= accum_diff_calc;

            debug_min <= sum_min;
            debug_max <= sum_max;

            i_debug.accum_min_max(accum_min_max'range) <=
              std_logic_vector(accum_min_max);
            i_debug.accum_max_min(accum_max_min'range) <=
              std_logic_vector(accum_max_min);
            i_debug.accum_avg(accum_avg'range) <=
              std_logic_vector(accum_avg); -- prev cycle
          end if;
        end if;
      else
        -- Not wrapping pulse period - inside the pulse!
        accum_high  <= accum_high + extend_samples_high_xor_r;

        -- Also count for debugging.
        sum_high <= sum_high + extend_samples_high_r(samples_high'range);

        -- Reset the value during normal collection.
        -- The delay of two wraps with the accum_high_clear array make
        -- the accumulation skip the fixed 10 slots.
        if (accum_high_clear(0) = '1') then
          -- Reset to a value of half the range,
          -- such that subtractions will not wrap the effective value,
          -- while treating it as unsigned.
          accum_high <= (accum_high'high => '1', others => '0');
        end if;

        -- Accumulate into the two ongoing long summations.
        long_accum_high0 <=
          long_accum_high0 +      extend_samples_high_r;  -- Add.
        long_accum_high1 <=
          long_accum_high1 + (not extend_samples_high_r); -- Subtract.
      end if;

      -- Default:
      i_bad_8slot <= '0';

      -- Should we have a sync tracking counter on long_accum_ind?
      -- No, it is easier just just extend to n*24*8 and let the
      -- better margin avoid mistakes.
      if (pick_long_accum_ind = '1') then
        long_accum_ind <= long_accum_high_ind;

        -- If we changed index, something is strange.
        if (long_accum_ind /= long_accum_high_ind) then
          i_bad_8slot <= '1';
        end if;

        i_debug.accum_high_max(long_accum_high_max'range) <=
          std_logic_vector(long_accum_high_max);
        i_debug.accum_high_ind(long_accum_high_ind'range) <=
          std_logic_vector(long_accum_high_ind);
      end if;
    end if;
  end process;

  i_synced <= period_sync.o.synced;

  bit_val <= '0' & i_bit_val;
  got_bit <= '0' & i_got_bit;
  missed_bit <= i_missed_bit;

  last_edges <= i_last_edges;

  synced <= i_synced;

  bad_period <= i_bad_period;
  bad_8slot  <= i_bad_8slot;

  debug <= i_debug;

  pulse_length_min <= std_logic_vector(debug_min);
  pulse_length_max <= std_logic_vector(debug_max);

  pulse_length_diff6 <= std_logic_vector(debug_diff);

end RTL;
