-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.rataser_records.all;
use work.rataser_symburst_util.all;
use work.rataser_util.all;

entity rataser_symburst_decode is
  generic(-- Pipeline stages to handle diff calculation.
          pipeline_diff_in   : boolean := false;
          pipeline_diff_1    : boolean := false;
          pipeline_diff_2    : boolean := false;
          pipeline_diff_out  : boolean := false;
          pipeline_match_out : boolean := false;
          -- Pipeline the output latching.
          pipeline_pick      : boolean := false;
          -- Pipeline output registers
          pipeline_out       : boolean := false);
  port (clk            : in std_logic;

        -- Input from bit decoder.
        bit_val        : in std_logic;
        got_bit        : in std_logic;

        -- Output of full 32-bit data word after a burst marker.
        mark_w16       : out std_logic;
        -- Only use output values at strobe.
        out_w16p       : out unsigned(15 downto 0);
        out_w16n       : out unsigned(15 downto 0);

        -- Output of info bit.
        mark_aux       : out std_logic;
        out_aux1p      : out std_logic;
        out_aux1n      : out std_logic;

        -- General synchronisation information:
        -- Continuous signal telling if we are good.
        synced        : out std_logic;
        -- Marks when a bad bit has been detected.
        bad_bit       : out std_logic
        );

end rataser_symburst_decode;


architecture RTL of rataser_symburst_decode is

  -- Invert the input signal bits.
  signal invert_signal : std_logic := '0';
  -- Inverted bits (if inverting).
  signal inv_bit       : std_logic := '0';

  -- Historical data (including new), 0 is oldest, 31 is newest.
  -- Symbol pattern only contain 24 bits, but for puredata_seq mode
  -- we return 32 most recent bits (thus need 33, in case we got 2).
  signal combined : std_logic_vector(0 to 32) := (others => '0');

  -- Cycle 1
  signal combined_prev : std_logic_vector(0 to 32) := (others => '0');

  -- Pipeline stages
  signal combined_p0 : std_logic_vector(0 to 32) := (others => '0');
  signal got_bit_p0  : std_logic := '0';

  signal combined_p1 : std_logic_vector(0 to 32) := (others => '0');
  signal got_bit_p1  : std_logic := '0';

  signal combined_p2 : std_logic_vector(0 to 32) := (others => '0');
  signal got_bit_p2  : std_logic := '0';

  signal combined_p3 : std_logic_vector(0 to 32) := (others => '0');
  signal got_bit_p3  : std_logic := '0';

  -- The stage used in the locked-loop.
  signal combined_r : std_logic_vector(0 to 32) := (others => '0');
  signal got_bit_r  : std_logic := '0';

  -- How well do we match
  signal diff_syma_p2  : unsigned(3 downto 0) := (others => '0');
  signal diff_symb_p2  : unsigned(3 downto 0) := (others => '0');
  signal diff_sym_p2   : unsigned(4 downto 0) := (others => '0');
  signal diff_symar_p2 : unsigned(3 downto 0) := (others => '0');
  signal diff_idle_p2  : unsigned(3 downto 0) := (others => '0');
  -- Pipeline stage at end of diff calculator.
  signal diff_sym_p3   : unsigned(4 downto 0) := (others => '0');
  signal diff_symar_p3 : unsigned(3 downto 0) := (others => '0');
  signal diff_symb_p3  : unsigned(3 downto 0) := (others => '0');
  signal diff_idle_p3  : unsigned(3 downto 0) := (others => '0');
  -- Is it considered a match?
  signal match_sym_p3 : std_logic := '0';
  signal inv_symar_p3 : std_logic := '0';
  signal inv_symb_p3  : std_logic := '0';
  signal inv_idle_p3  : std_logic := '0';
  -- Did we find bad bits?
  signal bad_symar_p3 : std_logic := '0';
  signal bad_symb_p3  : std_logic := '0';
  signal bad_idle_p3  : std_logic := '0';

  -- Pipelined, for used in locked-loop.
  signal match_sym_r : std_logic := '0';
  signal inv_symar_r : std_logic := '0';
  signal inv_symb_r  : std_logic := '0';
  signal inv_idle_r  : std_logic := '0';
  signal bad_symar_r : std_logic := '0';
  signal bad_symb_r  : std_logic := '0';
  signal bad_idle_r  : std_logic := '0';

  -- Symbol sequence counter, locked-look main actor.
  signal count      : unsigned(10 downto 0) := (others => '0');
  -- Number of bits we got, add to sequence counter.
  signal count_add : unsigned(4 downto 0) := (others => '0');
  -- Extra addition to jump past 42..63.
  signal count_add_jump : std_logic := '0';
  -- Next value of sequence counter (incl carry)
  signal count_next : unsigned(11 downto 0) := (others => '0');
  signal count_wrap8      : std_logic := '0';
  signal count_wrap128    : std_logic := '0';
  signal count_wrap128x16 : std_logic := '0';

  -- Pipelined for value-pick.
  signal count_next_r : unsigned(7 downto 0) := (others => '0');
  signal count_wrap_r : std_logic := '0';
  signal count_wrap2_r : std_logic := '0';

  -- Locations where a data transmission can start.
  signal can_start : std_logic := '0';

  -- Point where we accept a transmission as starting.
  signal match_point_symar : std_logic := '0';
  signal match_point_symb  : std_logic := '0';
  signal match_point_idle  : std_logic := '0';

  -- Number of bits gotten of data.
  signal count_data : unsigned(4 downto 0) := (others => '0');
  -- Next value of data counter.
  signal count_data_next : unsigned(5 downto 0) := (others => '0');
  signal count_data_wrap2  : std_logic := '0';
  signal count_data_wrap32 : std_logic := '0';

  -- Is a data transmission ongoing?
  signal data_transmission     : std_logic := '0';
  signal data_aux_transmission : std_logic := '0';

  -- Are we counting to get 32 bits of pure data?
  signal puredata32_pending : std_logic := '0';

  -- State of in-sync-tracking counter.
  signal match_sync : rataser_sync_count_iface :=
    rataser_sync_count_iface_zero;

  -- Decoded payload words
  signal w16sp      : unsigned(15 downto 0) := (others => '0');
  signal w16sn      : unsigned(15 downto 0) := (others => '0');

  -- Latched payload words and markers
  signal i_mark_w16       : std_logic := '0';
  signal i_out_w16p       : unsigned(15 downto 0) := (others => '0');
  signal i_out_w16n       : unsigned(15 downto 0) := (others => '0');

  signal i_mark_aux       : std_logic := '0';
  signal i_out_aux1p      : std_logic := '0';
  signal i_out_aux1n      : std_logic := '0';

  signal i_synced         : std_logic := '0';
  signal i_bad_bit        : std_logic := '0';

  -- Pipelined
  signal ir_mark_w16       : std_logic := '0';
  signal ir_out_w16p       : unsigned(15 downto 0) := (others => '0');
  signal ir_out_w16n       : unsigned(15 downto 0) := (others => '0');

  signal ir_mark_aux       : std_logic := '0';
  signal ir_out_aux1p      : std_logic := '0';
  signal ir_out_aux1n      : std_logic := '0';

  signal ir_synced         : std_logic := '0';
  signal ir_bad_bit        : std_logic := '0';

begin

  inv_bit <= bit_val xor invert_signal;

  combined <=
    (combined_prev(1 to 32) & inv_bit) when (got_bit = '1') else
    (combined_prev(0 to 32));
  -- 0 is the oldest, 31 is the newest item.

  -- Pipeline stage before the diff calculation
  rataser_ppl(p=>pipeline_diff_in,c=>clk,i=>combined   ,o=>combined_p0);
  rataser_ppl(p=>pipeline_diff_in,c=>clk,i=>got_bit    ,o=>got_bit_p0 );

  -- Diff against last 8 bits of match pattern, 8 bits ago.
  d0_symb8_15: entity work.rataser_symburst_diff
    generic map(pipeline_1 => pipeline_diff_1,
                pipeline_2 => pipeline_diff_2)
    port map(clk  => clk,
             seq  => combined_p0(17 to 24),
             ptn  => sym_sync16(0 to 7),
             diff => diff_syma_p2);
  -- Diff against last 8 bits of match pattern, most recent bits.
  d0_symb0_7: entity work.rataser_symburst_diff
    generic map(pipeline_1 => pipeline_diff_1,
                pipeline_2 => pipeline_diff_2)
    port map(clk  => clk,
             seq  => combined_p0(25 to 32),
             ptn  => sym_sync16(8 to 15),
             diff => diff_symb_p2);
  -- Diff against first 8 bits of match pattern, most recent bits.
  d0_syma0_7: entity work.rataser_symburst_diff
    generic map(pipeline_1 => pipeline_diff_1,
                pipeline_2 => pipeline_diff_2)
    port map(clk  => clk,
             seq  => combined_p0(25 to 32),
             ptn  => sym_sync16(0 to 7),
             diff => diff_symar_p2);
  -- Diff against 8 bits of idle pattern.
  d0_idle0_7: entity work.rataser_symburst_diff
    generic map(pipeline_1 => pipeline_diff_1,
                pipeline_2 => pipeline_diff_2)
    port map(clk  => clk,
             seq  => combined_p0(25 to 32),
             ptn  => "10101010",
             diff => diff_idle_p2);

  -- The diff values above are delivered (optionally) two pipeline stages
  -- after the input.  Thus p0 for combined and p2 for diffx.

  -- Pipeline stages within the diff calculation
  rataser_ppl(p=>pipeline_diff_1 ,c=>clk,i=>combined_p0,o=>combined_p1);
  rataser_ppl(p=>pipeline_diff_1 ,c=>clk,i=>got_bit_p0 ,o=>got_bit_p1 );

  rataser_ppl(p=>pipeline_diff_2 ,c=>clk,i=>combined_p1,o=>combined_p2);
  rataser_ppl(p=>pipeline_diff_2 ,c=>clk,i=>got_bit_p1 ,o=>got_bit_p2 );

  -- Add the contributions of the two parts of the 16-bit symbol.
  diff_sym_p2 <= ("0" & diff_syma_p2) + ("0" & diff_symb_p2);

  -- Pipeline stage after the diff calculation
  rataser_ppl(p=>pipeline_diff_out,c=>clk,i=>combined_p2  ,o=>combined_p3  );
  rataser_ppl(p=>pipeline_diff_out,c=>clk,i=>got_bit_p2   ,o=>got_bit_p3   );
  rataser_ppl(p=>pipeline_diff_out,c=>clk,i=>diff_sym_p2  ,o=>diff_sym_p3  );
  rataser_ppl(p=>pipeline_diff_out,c=>clk,i=>diff_symar_p2,o=>diff_symar_p3);
  rataser_ppl(p=>pipeline_diff_out,c=>clk,i=>diff_symb_p2 ,o=>diff_symb_p3 );
  rataser_ppl(p=>pipeline_diff_out,c=>clk,i=>diff_idle_p2 ,o=>diff_idle_p3 );

  -- For matching, accept one piece of junk.
  -- (The patterns have a design difference of 2 to any other valid
  -- pattern, including inverted.)
  match_sym_p3 <= '1' when (to_integer(diff_sym_p3) <= 1) else '0';
  -- For start of burst, whenever more than half the bits have been inverted.
  inv_symar_p3 <= '1' when (to_integer(diff_symar_p3) > 4) else '0';
  inv_symb_p3  <= '1' when (to_integer(diff_symb_p3)  > 4) else '0';
  inv_idle_p3  <= '1' when (to_integer(diff_idle_p3)  > 4) else '0';

  -- If the difference is not 0 or 8, then some bit is bad.
  bad_symar_p3 <= '1' when (diff_symar_p3(2 downto 0) /= "000") else '0';
  bad_symb_p3  <= '1' when (diff_symb_p3 (2 downto 0) /= "000") else '0';
  bad_idle_p3  <= '1' when (diff_idle_p3 (2 downto 0) /= "000") else '0';
  
  -- Pipeline stage after the match calculation, into locked loop.
  rataser_ppl(p=>pipeline_match_out,c=>clk,i=>combined_p3 ,o=>combined_r );
  rataser_ppl(p=>pipeline_match_out,c=>clk,i=>got_bit_p3  ,o=>got_bit_r  );
  rataser_ppl(p=>pipeline_match_out,c=>clk,i=>match_sym_p3,o=>match_sym_r);
  rataser_ppl(p=>pipeline_match_out,c=>clk,i=>inv_symar_p3,o=>inv_symar_r);
  rataser_ppl(p=>pipeline_match_out,c=>clk,i=>inv_symb_p3 ,o=>inv_symb_r );
  rataser_ppl(p=>pipeline_match_out,c=>clk,i=>inv_idle_p3 ,o=>inv_idle_r );
  rataser_ppl(p=>pipeline_match_out,c=>clk,i=>bad_symar_p3,o=>bad_symar_r);
  rataser_ppl(p=>pipeline_match_out,c=>clk,i=>bad_symb_p3 ,o=>bad_symb_r );
  rataser_ppl(p=>pipeline_match_out,c=>clk,i=>bad_idle_p3 ,o=>bad_idle_r );

  -- Direct-coded 16-bit data.
  decode_16_0: for i in 0 to 15 generate
    -- Extract from 1 to 32.
    w16sp(i) <= combined_r(i*2 + 1); -- 1, 3, 5, ..., 31
    w16sn(i) <= combined_r(i*2 + 2); -- 2, 4, 6, ..., 32
  end generate;

  -- Add 23 when we start from 41, otherwise 1.
  count_add(4 downto 1) <=
    "1011" when (got_bit_r = '1' and
                 count_add_jump = '1') else
    "0000";
  count_add(0) <= got_bit_r;
  -- The above is a table with four inputs.  Let's hope the compiler
  -- can figure that out to something efficient.
  -- Do the adding.
  count_next <= ('0' & count) + count_add;
  count_wrap8      <= count_next(3) xor count(3);
  count_wrap128    <= count_next(7) xor count(7);
  count_wrap128x16 <= count_next(11);

  -- Are we at a position where we can expect a start pattern?
  -- Not after a jump, since the previous two bits are the aux bits.
  can_start <= count_wrap8 and not count_add_jump;

  -- Inverse pattern for sync pattern at end of counter range, i.e.
  -- right before jump to 120 and 128.
  match_point_symar <= '1' when (count(6 downto 3) =  "1110") else '0';
  match_point_symb  <= '1' when (count(6 downto 3) =  "1111") else '0';
  match_point_idle  <= '1' when (count(6 downto 4) /= "111" ) else '0';

  -- Counter for data reception.
  count_data_next   <= ('0' & count_data) + ("" & got_bit_r);
  count_data_wrap2  <= count_data_next(1) xor count_data(1);
  count_data_wrap32 <= count_data_next(5);

  -- Keep track of our sync status.
  match_csync: entity work.rataser_sync_count
    generic map(maxvalue_bits => 5, -- Maxvalue  2^5-1 = 31.
                increment_bit => 3) -- Increment 2^3   = 8 on match
    port map(clk => clk,
             i   => match_sync.i,
             o   => match_sync.o);

  process(clk)
  begin
    if (rising_edge(clk)) then
      -- Mark strobes are only set for one cycle
      i_mark_w16  <= '0';
      i_mark_aux  <= '0';
      i_bad_bit   <= '0';

      -- Previous value.
      combined_prev <= combined;

      -- Always update the count (unless reset below).
      count <= count_next(count'range);

      count_add_jump <= '0';
      if (count(5 downto 0) = "011001") then -- 25 (41-16)
        count_add_jump <= '1';
      end if;

      count_data <= count_data_next(count_data'range);

      -- Default values.
      match_sync.i <= rataser_sync_count_in_zero;
      
      -- Note: tentative sync is reported also when we are in sync state.
      if (match_sync.o.tentative_synced = '1') then
        -- If we are currently synced, then we shall match
        -- when the counter wraps.  If not, we are bad.

        -- Note: for overall sync we do not care about transmissions.
        -- Transmissions will mismatch the sync pattern.
        -- But at least every 8 cycles should have a sync pattern,
        -- So it is anough with our up-counter adding 8 for each match.

        if (count_wrap128 = '1') then
          -- We did wrap.  Did we wrap at the right spot?
          -- Since we wrapped, got_bit_r(1) must be 1.

          -- If we got 2 bits, and we overshot by one bit and prev match.
          -- Or (either got 1 or 2 bits), and did not overshoot and match cur.
          if (match_sym_r = '1') then
            -- All is well.
            match_sync.i.up <= '1';
          else
            -- We have mismatched the pattern.
            match_sync.i.down <= '1';
          end if;
        end if;
      else
        -- If we are not yet in sync, then we restart
        -- the counter when we hit a match.

        if (got_bit_r = '1' and match_sym_r = '1') then
          match_sync.i.tentative <= '1';
          -- Report one up, such that we can handle the 7 cycles
          -- of down due to missing sync patterns in case of
          -- transmissions.
          match_sync.i.up <= '1';
          -- Count starts at 0.
          count <= (others => '0');
        end if;

        if (count_wrap128x16 = '1') then
          -- We have seen what should amount to 16 cycles with no
          -- match.  Try inverted pattern.
          -- (After we switch, it takes some time until history
          -- has been inverted.  Then we need to scan for 8 cycles,
          -- since alignment is unknown.  Thus 9 cycles minimum,
          -- testing for 16 is cheaper.)
          invert_signal <= not invert_signal;
        end if;
      end if;

      -- Data transmission start?
      -- We cannot start (again) while a transmission is ongoing,
      -- or just at the end of it (aux data).
      if (got_bit_r = '1' and
          can_start = '1' and
          data_transmission = '0' and
          data_aux_transmission = '0' and
          match_sync.o.synced = '1') then

        if ((match_point_symar = '1' and inv_symar_r = '1') or
            (match_point_symb  = '1' and inv_symb_r  = '1') or
            (match_point_idle  = '1' and inv_idle_r  = '1')) then
          data_transmission <= '1';
          count_data <= (others => '0');
        end if;

        -- At a spot which can start a transmission,
        -- we can check for bad bits.

        if ((match_point_symar = '1' and bad_symar_r = '1') or
            (match_point_symb  = '1' and bad_symb_r  = '1') or
            (match_point_idle  = '1' and bad_idle_r  = '1')) then
          i_bad_bit <= '1';
        end if;
      end if;

      if (data_transmission = '1' and
          count_data_wrap32 = '1') then
        -- Data transmission is done.
        data_transmission <= '0';
        data_aux_transmission <= '1';
        -- Report data.
        i_mark_w16 <= '1'; -- We only get here if we are synced.
      end if;

      if (data_aux_transmission = '1' and
          count_data_wrap2 = '1') then
        -- Data aux transmission is done.
        data_aux_transmission <= '0';

        -- We can provide the value.
        i_mark_aux  <= '1';
        i_out_aux1p <= combined_r(31);
        i_out_aux1n <= combined_r(32);
      end if;

      -- The idle pattern provides the aux bits when it is jumping,
      -- but not when a transmission is ongoing.
      if (got_bit_r = '1' and
          count_add_jump = '1' and
          data_transmission = '0') then
        i_mark_aux  <= '1';
        i_out_aux1p <= combined_r(31);
        i_out_aux1n <= combined_r(32);
      end if;

    end if;
  end process;

  -- The data values can be picked directly.
  -- The mark is latched/set in the previous cycle, but we do not
  -- get data bits every cycle, so will not have changed.
  i_out_w16p <= w16sp;
  i_out_w16n <= w16sn;

  i_synced <= match_sync.o.synced;

  rataser_ppl(p=>pipeline_out,c=>clk,i=>i_mark_w16      ,o=>ir_mark_w16      );
  rataser_ppl(p=>pipeline_out,c=>clk,i=>i_out_w16p      ,o=>ir_out_w16p      );
  rataser_ppl(p=>pipeline_out,c=>clk,i=>i_out_w16n      ,o=>ir_out_w16n      );

  rataser_ppl(p=>pipeline_out,c=>clk,i=>i_mark_aux      ,o=>ir_mark_aux      );
  rataser_ppl(p=>pipeline_out,c=>clk,i=>i_out_aux1p     ,o=>ir_out_aux1p     );
  rataser_ppl(p=>pipeline_out,c=>clk,i=>i_out_aux1n     ,o=>ir_out_aux1n     );

  rataser_ppl(p=>pipeline_out,c=>clk,i=>i_synced        ,o=>ir_synced        );
  rataser_ppl(p=>pipeline_out,c=>clk,i=>i_bad_bit       ,o=>ir_bad_bit       );

  -- Assign to output signals

  mark_w16       <= ir_mark_w16;
  out_w16p       <= ir_out_w16p;
  out_w16n       <= ir_out_w16n;

  mark_aux       <= ir_mark_aux;
  out_aux1p      <= ir_out_aux1p;
  out_aux1n      <= ir_out_aux1n;
  
  synced         <= ir_synced;
  bad_bit        <= ir_bad_bit;
  
end RTL;
