-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.rataser_records.all;
use work.rataser_util.all;
use work.rataser_sym24word8_util.all;

entity rataser_symburst_diff is
  generic(pipeline_1 : boolean;
          pipeline_2 : boolean);
  port(clk  : in std_logic;

       seq  : in std_logic_vector(0 to 7);
       ptn  : in std_logic_vector(0 to 7);
       diff : out unsigned(3 downto 0));
end rataser_symburst_diff;

architecture RTL of rataser_symburst_diff is

  type uint_1downto0_array is array (integer range <>) of unsigned(1 downto 0);
  type uint_2downto0_array is array (integer range <>) of unsigned(2 downto 0);
    
  signal ms : std_logic_vector(0 to 7);
  -- Summing up the mismatches adder tree
  -- ((3+3+1)+1)=8
  signal d3   : uint_1downto0_array(0 to 1);

  -- For pipelining
  signal ms_p : std_logic_vector(0 to 7);
  signal d3_p : uint_1downto0_array(0 to 1);

  -- Next summing stage
  signal d7_p : uint_2downto0_array(0 to 0);

  -- For pipelining
  signal ms_q : std_logic_vector(0 to 7);
  signal d7_q : uint_2downto0_array(0 to 0);

  -- Final summing stage  
  signal d15_q : unsigned(3 downto 0);

begin
    -- Evaluate the mismatches
  ms_gen: for i in 0 to 7 generate
    ms(i) <= seq(i) xor ptn(i);
  end generate;

  -- First summing up in triples (yield values 0-3)
  d3(0) <= ((("00" + ("" & ms(0))) + ("" & ms(1))) + ("" & ms(2)));
  d3(1) <= ((("00" + ("" & ms(3))) + ("" & ms(4))) + ("" & ms(5)));

  -- Pipeline stage
  rataser_ppl(p=>pipeline_1,c=>clk,i=>d3(0),o=>d3_p(0));
  rataser_ppl(p=>pipeline_1,c=>clk,i=>d3(1),o=>d3_p(1));
  rataser_ppl(p=>pipeline_1,c=>clk,i=>ms   ,o=>ms_p);

  -- Then summing up the triples, plus two more 1-bit contributions
  d7_p(0) <= ("0" & d3_p(0)) + ("0" & d3_p(1)) + ("00" & ms_p(6));

  -- Pipeline stage
  rataser_ppl(p=>pipeline_2,c=>clk,i=>d7_p(0),o=>d7_q(0));
  rataser_ppl(p=>pipeline_2,c=>clk,i=>ms_p   ,o=>ms_q);

  -- And the final sum
  d15_q   <= ("0" & d7_q(0)) + ("000" & ms_q(7));

  diff <= d15_q;
end RTL;
  
  
