-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.rataser_records.all;

-- We keep track of sync state, by incrementing a counter
-- each time something good happens (i.up), and decrementing it
-- when something bad happens (or lack of something good) (i.down).
-- The counter is saturating, i.e. does not wrap.
-- When we would have wrapped over the top, we declare
-- sync state, and when below zero, desync.

-- To exclusively use powers of two, the number of bits are given.
-- Use at most with increment_bits < maxvalue_bits.
-- Decrements are always by 1.

-- The i.tentative synced marks that outside world found something
-- which made it move from random search to attempt at tracking.
-- It is set by outside, but cleared by us, if attempting to
-- decrement below 0.  Since we usually start at 0, this means
-- that next report better be positive!  Or we will loose the
-- tentative state again.

-- Note: tentative sync is reported by us while we are in sync state.

-- Giving i.up and i.down simultaneously makes no sense.  But
-- i.down will win.  We decrement and mark sync lost if reaching 0.

entity rataser_sync_count is
  generic(maxvalue_bits  : integer := 2;  -- maxvalue  = 2^2-1 = 3
          increment_bit  : integer := 0); -- increment = 2^0   = 1
  port (clk            : in  std_logic;
        i              : in  rataser_sync_count_in;
        o              : out rataser_sync_count_out
        );
end rataser_sync_count;

architecture RTL of rataser_sync_count is

  constant inc_bit  : integer := increment_bit;
  
  signal count      : unsigned(maxvalue_bits-1 downto 0) := (others => '0');
  signal count_add  : unsigned(maxvalue_bits-1 downto 0) := (others => '0');
  signal count_next : unsigned(maxvalue_bits-1+1 downto 0) := (others => '0');
  signal count_wrap : std_logic := '0';

  signal i_synced           : std_logic := '0';
  signal i_tentative_synced : std_logic := '0';

    -- For debugging (gtkwave dump)
  signal debug_in_up        : std_logic;
  signal debug_in_down      : std_logic;
  signal debug_in_tentative : std_logic;

begin

  -- The match counter increment (adding 2^increment_bit, or -1 (all 1)).
  count_add(count_add'high  downto increment_bit+1) <= (others => i.down);
  count_add(increment_bit)                          <= (i.up or i.down);
  count_add(increment_bit-1 downto count_add'low  ) <= (others => i.down);
  -- Evaluate the adding.
  count_next <= ('0' & count) + count_add;
  -- We wrap (either direction) if the high bit is set.
  count_wrap <= count_next(count_next'high) xor i.down;

  process(clk)
  begin
    if (rising_edge(clk)) then

      if (i.tentative = '1') then
        i_tentative_synced   <= '1';
      end if;

      if (count_wrap = '1') then
        if (i.up = '1') then
          -- We would reach > maxval, so now claim to be synced.
          i_synced <= '1';
        end if;
        if (i.down = '1') then
          -- We have had enough misfortune in a row - declare desync!
          -- This overrides the handling of i.up just above.
          i_synced <= '0';
          i_tentative_synced <= '0';
        end if;
      else
        -- Only update the match counter if we are not wrapping it.
        count <= count_next(count'range);
        -- Note, if up-counting more than 1, then we will not reach the
        -- maximum value possible.  That is however not a problem,
        -- since we are still marked sync, and to be desynced, we must
        -- fall the entire way to the bottom.  If this fall is a few
        -- steps less does not matter.
      end if;

    end if;
  end process;

  o.synced           <= i_synced;
  o.tentative_synced <= i_tentative_synced;

  debug_in_up        <= i.up;
  debug_in_down      <= i.down;
  debug_in_tentative <= i.tentative;
  
end RTL;

