-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.rataser_records.all;
use work.rataser_util.all;

entity rataser_timestamp_decode is
  generic(pipeline_hamming_parity : boolean := false;
          pipeline_hamming_dual   : boolean := false);
  port (clk            : in std_logic;

        -- Input from symbol/word decoder
        mark_w8a1      : in std_logic;
        in_w8p         : in unsigned(7 downto 0);
        in_w8n         : in unsigned(7 downto 0);
        in_aux1p       : in std_logic;
        in_aux1n       : in std_logic;

        -- Input from bitstream tracker.
        clk_add_ns       : in std_logic_vector(9 downto 0);
        clk_add_ns_carry : in std_logic;

        reset_add_counter_on_bit : out std_logic;
        got_bit          : in std_logic;

        -- Reception delay (in ns, added to timestamp).
        receive_delay_ns: in std_logic_vector(15 downto 0);

        -- Output timestamp, low 32 bits.
        tick_ns_lo:    out std_logic_vector(31 downto  0);
        -- High 32 bits of timestamp, latch one cycle later than low word.
        tick_ns_hi:    out std_logic_vector(63 downto 32);
        -- Aux signals.
        aux_sigs:      out std_logic_vector(4 downto 0);
        -- Auxiliary information.
        info_bit:      out std_logic;
        -- Strobe signal that we have decoded a new info bit message.
        msg_strobe:    out std_logic;

        -- Continuous signal telling if we are good.
        synced:        out std_logic;
        -- Something was not perfect with a hamming block.
        bad_hamming:   out std_logic;
        -- An timestamp message did not match current counter.
        mismatch_time: out std_logic;

        -- Pulse that happens once in a while (once per expected message).
        wrap_midbit:   out std_logic;

        -- Pulse period reported from sender.
        sender_period_ns : out std_logic_vector(9 downto 0)
        );
end rataser_timestamp_decode;

-- The data is encoded into (15,11) Hamming blocks (16 bits of data).
-- This is provided by dual words of 8-bit data.  To know which is
-- first block, that has aux bit set to 0.

-- Each Hamming block delivers 11 data bits.
-- 4 are used as timestamp and aux information (timestamp speed).
-- 2 bits are used to mark the data.
--   "00" - first timestamp word
--   "01" - timestamp word
--   "10" - last timestamp word
--   "11" - filler data (since timestamps start irregularly)
-- Remaining 5 bits are used to carry aux signals (updated with each
-- Hamming block).

-- We simply get a full 64-bit timestamp sent to us.

-- We need not be particularly fast in doing things - we will
-- not get data bits very often.

-- Since the input aux bits (first/second word of data for Hamming block)
-- have no forward error correction, we must slowly sync to them,
-- i.e. be able to survive a few broken.

-- That sync is roughly modeled after the symbol sync tracker of the
-- sym24word8 decoder.

architecture RTL of rataser_timestamp_decode is

  signal expect_aux : std_logic := '0';

  signal aux_sync : rataser_sync_count_iface;
  
  -- The first part of the Hamming input data.
  signal prev_word8p  : unsigned(7 downto 0) := (others => '0');
  signal prev_word8n  : unsigned(7 downto 0) := (others => '0');
  -- The full input to the Hamming decoder.
  signal full_word16p : std_logic_vector(15 downto 0) := (others => '0');
  signal full_word16n : std_logic_vector(15 downto 0) := (others => '0');

  -- Output from the hamming decoder.
  signal hamm_word11        : std_logic_vector(10 downto 0) := (others => '0');
  signal hamm_good          : std_logic := '0';
  signal hamm_bad_bit       : std_logic := '0';
  signal hamm_single_source : std_logic := '0';
  -- Already pipelined.
  signal got_hamming          : std_logic := '0';
  -- Pipelined.
  signal got_hamming_r        : std_logic := '0';
  signal got_hamming_rr       : std_logic := '0';
  -- Pilelined after decoder.
  signal hamm_rr_word11        : std_logic_vector(10 downto 0) :=(others=>'0');
  signal hamm_rr_good          : std_logic := '0';
  signal hamm_rr_bad_bit       : std_logic := '0';
  signal hamm_rr_single_source : std_logic := '0';

  -- Decomposed Hamming data.
  signal hamm_rr_w11_ts_data   : std_logic_vector(3 downto 0) := (others=>'0');
  signal hamm_rr_w11_grp_mark  : std_logic_vector(1 downto 0) := (others=>'0');
  signal hamm_rr_w11_aux_sigs  : std_logic_vector(4 downto 0) := (others=>'0');

  signal timestamp_sync : rataser_sync_count_iface;

  -- Keep track of the data block progress.
  signal data_block_ongoing    : std_logic := '0';
  signal data_block_count      : unsigned(4 downto 0) := (others => '0');
  signal data_block_count_next : unsigned(5 downto 0);
  signal data_block_count_wrap : std_logic := '0';
  signal data_block_complete   : std_logic := '0';
  -- Data reception shift register.
  signal data_shift_reg : std_logic_vector(64+16-1 downto 0) := (others =>'0');

  -- Aliases from the data reception shift register.
  signal timestamp_recv  : std_logic_vector(63 downto 0);
  signal extra_data_recv : std_logic_vector(15 downto 0);

  -- Semi-alias (partially latched) value of received timestamp,
  -- with reception delay added.
  signal timestamp_recv_off : std_logic_vector(63 downto 0);

  -- Alias of the alias.
  signal extra_data_recv_period_ns : std_logic_vector( 9 downto 0);

  -- Delayed actions.
  signal on_bit_take_timestamp  : std_logic := '0';
  signal on_bit_check_timestamp : std_logic := '0';
  signal check_timestamp        : std_logic := '0';
  signal check_timestamp2       : std_logic := '0';
  signal check_timestamp3       : std_logic := '0';

  -- The timestamp counter.  Split into low and high word, since high
  -- word is added in the next cycle.
  -- 32nd bit of low word is carried into next cycle.
  -- Init crazy value for easier simulation inspection.
  signal timestamp_counter_lo : unsigned(32 downto 0) :=
    "0" & "00000000" & "11111111" & "00000000" & "11111111";
  signal timestamp_counter_hi : unsigned(63 downto 32) :=
    "00000000" & "11111111" & "00000000" & "11111111";

  -- Calculation of how we are matching the received timestamp.
  -- Split into two words since high word is handled in the next cycle.
  signal timestamp_diff_lo : unsigned(32 downto  0) := (others => '0');
  signal timestamp_diff_hi : unsigned(64 downto 32) := (others => '0');
  -- The low word carry bit must be latched.
  signal timestamp_diff_lo_carry : std_logic := '0';
  -- Is the difference < 0 ?
  signal timestamp_less      : std_logic := '0';
  signal timestamp_less_prev : std_logic := '0';
  signal timestamp_match     : std_logic := '0';

  -- Are we matching the (new) reported sender period.
  signal sender_period_match : std_logic := '0';

  -- The sender sends a new timestamp every time this bit changes.
  constant tick_wrap_loc : integer := 23;
  -- Previous value of the wrapping location.
  signal timestamp_prev_wrap_loc : std_logic := '0';

  -- Make sure timestamp is progressing.
  signal force_add_carry : std_logic := '0';

  -- Output signals.
  signal i_aux_signals : std_logic_vector(4 downto 0) := (others => '0');

  signal i_info_bit   : std_logic := '0';
  signal i_msg_strobe : std_logic := '0';

  signal i_sender_period_ns : std_logic_vector(9 downto 0) := (others => '0');

  signal i_reset_add_counter_on_bit : std_logic := '0';

  signal i_synced  : std_logic := '0';

  signal i_bad_hamming : std_logic := '0';
  signal i_mismatch_time : std_logic := '0';

  signal i_wrap_midbit : std_logic := '0';
  
begin

  aux_csync: entity work.rataser_sync_count
    port map(clk => clk,
             i   => aux_sync.i,
             o   => aux_sync.o);

  -- Prepare data for the Hamming decoder
  full_word16p <=     std_logic_vector(prev_word8p & in_w8p);
  full_word16n <= not std_logic_vector(prev_word8n & in_w8n);

  -- Decode the hamming data, and apply corrections from having
  -- two copies: positive and negative data.
  decode_hamm: entity work.rataser_hamming_decode_dual
    generic map(pipeline_parity => pipeline_hamming_parity,
                pipeline_dual   => pipeline_hamming_dual)
    port map(clk       => clk,
             message_p => full_word16p,
             message_n => full_word16n,
             
             data      => hamm_word11,
             good      => hamm_good,
             bad_bit   => hamm_bad_bit,
             single_source => hamm_single_source);

  -- Delay for the pipeline in the Hamming decoder.
  rataser_ppl(p=>pipeline_hamming_parity,
              c=>clk,i=>got_hamming,o=>got_hamming_r);
  rataser_ppl(p=>pipeline_hamming_dual,
              c=>clk,i=>got_hamming_r,o=>got_hamming_rr);

  -- For ease-of-read, split the decoded word into its parts
  hamm_rr_w11_ts_data  <= hamm_rr_word11( 3 downto 0);
  hamm_rr_w11_grp_mark <= hamm_rr_word11( 5 downto 4);
  hamm_rr_w11_aux_sigs <= hamm_rr_word11(10 downto 6);

  -- The timestamp sync counter is decremented by 1 for each cycle,
  -- when we would expect also one increment.
  -- For each match, we add 4.  Max value is 15.
  -- So it takes an initial value and four matches to hit the roof.  
  ts_csync: entity work.rataser_sync_count
    generic map(maxvalue_bits => 4, -- Max value is 15.
                increment_bit => 2) -- Increment by 4 for each hit
    port map(clk => clk,
             i   => timestamp_sync.i,
             o   => timestamp_sync.o);

  data_block_count_next <= ('0' & data_block_count) - 1;
  data_block_count_wrap <=
    data_block_count_next(data_block_count_next'high);

  timestamp_recv  <= data_shift_reg(63    downto  0);
  extra_data_recv <= data_shift_reg(64+15 downto 64);

  -- Another alias, high bits are not added,
  -- bits 22 downto 0 are set further down (far below).
  timestamp_recv_off(63 downto 23) <= timestamp_recv(63 downto 23);

  -- For convenience, pick out the members.
  extra_data_recv_period_ns <= extra_data_recv(15 downto 6);

  -- Difference between local timestamp counter and received value.
  timestamp_diff_lo <= ('0' & timestamp_counter_lo(31 downto  0)) -
                       unsigned(timestamp_recv_off(31 downto  0));
  timestamp_diff_hi <= ('0' & timestamp_counter_hi(63 downto 32)) -
                       unsigned(timestamp_recv_off(63 downto 32)) -
                       ("" & timestamp_diff_lo_carry);

  -- We had a match if we were less, but no longer are.
  -- I.e. were < and are now >=.
  timestamp_match <= timestamp_less_prev and not timestamp_less;

  -- Do we have the period value set which is reported.
  -- If not, and the issue persists (i.e. it was not a bad received
  -- value, we'll loose lock eventually).
  sender_period_match <=
    '1' when (i_sender_period_ns = extra_data_recv_period_ns) else '0';

  process(clk)
  begin
    if (rising_edge(clk)) then

      aux_sync.i <= rataser_sync_count_in_zero;

      -- Aux bit is delivered with each 8-bit data word.
      if (mark_w8a1 = '1') then
        if (aux_sync.o.tentative_synced = '1') then
          if ((in_aux1p =     expect_aux) and
              (in_aux1n = not expect_aux)) then
            -- Both aux bits match the expectation, good.
            aux_sync.i.up <= '1';
          else
            -- We take any mismatching aux bits, and outright
            -- wrong aux bits as a bad sign.  This is necessary such
            -- that a lack of good bits eventually declares non-sync.
            aux_sync.i.down <= '1';
          end if;
          -- In any case, expect next to be opposite.
          expect_aux <= not expect_aux;
        else -- Not even in tentative sync.
          -- The two aux bits must match for us to trust them for a start.
          if (in_aux1p /= in_aux1n) then
            aux_sync.i.tentative <= '1';
            expect_aux <= in_aux1n; -- Next one should be opposite.
          end if;
        end if;
      end if;

      got_hamming <= '0';

      if (mark_w8a1 = '1') then
        if (aux_sync.o.synced = '1') then
          if (expect_aux = '0') then
            -- Latch the first part of the Hamming input
            prev_word8p <= in_w8p;
            prev_word8n <= in_w8n;
          else
            -- We got the second part of the Hamming input
            -- use the decoded data!
            -- To allow better timing, we just latch it,
            -- and deal with it next cycle.
            got_hamming <= '1';           
          end if;
        end if;
      end if;

      -- We always latch the hamming data.
      -- It will only be used when needed though.
      hamm_rr_word11        <= hamm_word11;
      hamm_rr_good          <= hamm_good;
      hamm_rr_bad_bit       <= hamm_bad_bit;
      hamm_rr_single_source <= hamm_single_source;

      -- Default:
      i_bad_hamming <= '0';

      if (got_hamming_rr = '1') then
        -- Shift the timestamp data into the shift register,
        -- regardless of quality.  Least-significant bits come first.
        data_shift_reg <=
          hamm_rr_w11_ts_data &
          data_shift_reg(data_shift_reg'high downto 4);

        if (hamm_rr_good = '1') then
          case (hamm_rr_w11_grp_mark) is
            when "00" =>
              -- First word of timestamp sequence; (re)start.
              data_block_ongoing <= '1';
              -- Reset expected blocks until end.
              data_block_count <=
                to_unsigned(18, data_block_count'length);
              
            when "01" =>
              -- Increment the block count no matter if we are
              -- ongoing.  If we are not ongoing, we will not be used
              -- afterwards.
              data_block_count <=
                data_block_count_next(data_block_count'range);
              if (data_block_count_wrap = '1') then
                -- We should have ended here.
                data_block_ongoing <= '0';
              end if;

            when "10" =>
              -- We shall have received 19 timestamp blocks + 1 extra
              -- with this.  I.e. 20 total.
              if (data_block_ongoing = '1' and
                  data_block_count_wrap = '1') then
                -- Deal with the data next cycle.
                data_block_complete <= '1';
              end if;
              data_block_ongoing <= '0';

            when "11" =>
              -- Filler data.
              data_block_ongoing <= '0';

            when others => -- Make compiler happy.

          end case;
        else -- hamm_rr_good = '0'
          -- We contain bad data
          data_block_ongoing <= '0';
        end if;

        i_bad_hamming <= hamm_rr_bad_bit;

        -- Update the aux signals if the data is deemed good.
        if (hamm_rr_good = '1') then
          i_aux_signals <= hamm_rr_w11_aux_sigs;
        end if;
      end if;

      timestamp_sync.i <= rataser_sync_count_in_zero;

      i_wrap_midbit <= '0';

      -- Remember previous value of sender new-message bit.
      timestamp_prev_wrap_loc <= timestamp_counter_lo(tick_wrap_loc);

      if (timestamp_prev_wrap_loc /= timestamp_counter_lo(tick_wrap_loc)) then
        -- So we slowly decrease the sync counter when this occurs.
        timestamp_sync.i.down <= '1';
        i_wrap_midbit <= '1';
      end if;

      force_add_carry <= '0';
      -- In order for timestamp to always progress a little bit, even
      -- if we have a broken (0) clk_add_ns value, force add 1.
      -- This is necessary such that the timestamp_sync.i.down is set
      -- above once in a while, to eventually loose the synced mark in
      -- case signal is lost.
      if (unsigned(clk_add_ns) = 0) then
        force_add_carry <= '1';
      end if;

      -- Add to the timestamp counter.
      -- High bit of destination is carry, so not included as a source.
      timestamp_counter_lo <=
        ('0' & timestamp_counter_lo(31 downto  0)) +
        unsigned(clk_add_ns) + ("" & (clk_add_ns_carry or force_add_carry));
      -- The high word is added in the following cycle.
      -- Only need to add carry from low word.
      timestamp_counter_hi <=
        timestamp_counter_hi + ("" & timestamp_counter_lo(32));

      i_msg_strobe <= '0';

      i_reset_add_counter_on_bit <= '0';

      -- Due to reset_add_counter_on_bit acting in the bitstream tracker,
      -- on the next bit that *that* routine gets, we first wait until
      -- the next bit here, such that we have more margin against the
      -- pipelining between the bitstream tracker and us.  Thus, the
      -- on_bit_..._timestamp will act on the second gotten bit.
      if (got_bit = '1' and
          data_block_complete = '1') then
        -- Do not do this again.
        data_block_complete <= '0';

        -- We have received a full timestamp.

        -- If we are not in sync, not even tentatively, ask the
        -- bitstream tracker to reset its counter on next bit.
        -- This is necessary such that we can rely on that we are
        -- counting in sync with it - i.e. in the cycle when the next
        -- bit is gotten after a complete message, we also increase
        -- the timestamp to the correct point.

        -- We always take or check the timestamp on the next bit.
        -- By doing it on the bit, we know that the timestamp updates
        -- shall just have crossed the boundary of timestamp being
        -- below the reference value in the cycle before, and being
        -- at or above in the got-bit cycle.

        if (timestamp_sync.o.tentative_synced = '0') then
          if (sender_period_match = '1') then
            i_reset_add_counter_on_bit <= '1';
            on_bit_take_timestamp <= '1';
          end if;

          -- Sender tells what the period between each each pulse is.
          -- We copy the value when we get a complete block, but
          -- are not presumed to be in sync.
          i_sender_period_ns <= extra_data_recv_period_ns;
        else
          on_bit_check_timestamp <= '1';
        end if;

        -- Latch the info bit, and strobe its reception
        i_info_bit   <= extra_data_recv(0);
        i_msg_strobe <= '1';
      end if;

      if (got_bit = '1' and
          on_bit_take_timestamp = '1') then
        -- Do not do this again.
        on_bit_take_timestamp <= '0';

        -- We are not synced at all, copy timestamp.
        -- High word in principle next cycle.
        -- But there will be no carry, since the transmit cycle has
        -- some bit 0 in the middle of the low word (tick_wrap_loc).
        timestamp_counter_lo <= unsigned('0'&timestamp_recv_off(31 downto  0));
        timestamp_counter_hi <= unsigned(    timestamp_recv_off(63 downto 32));
        -- Set tentatively synced, i.e. look for more matches
        -- But we only consider things good when period value matches.
        timestamp_sync.i.tentative <= sender_period_match;
        -- Since we are down'ing the sync count once per 4-bit
        -- block, we must yank it up such that it has a chance
        -- to still be tentative when we hit the next sync check.
        timestamp_sync.i.up        <= sender_period_match;
        timestamp_sync.i.down <= '0'; -- Would have taken priority.
      end if;

      -- We will declare to be in sync, once some time-stamps have
      -- matched.  Non-matching timestamps may be due to bad
      -- bit reception, so we do not discount them.

      -- We however must discount when time elapses without
      -- new timestamps being received.  I.e. when we should
      -- have received timestamps, but do not.  This is handled
      -- by the down-counting each time the local timestamp
      -- counter passes the location where a timestamp message
      -- would be started by the sender.

      check_timestamp <= '0';

      if (got_bit = '1' and
          on_bit_check_timestamp = '1') then
        -- Do not do this again.
        on_bit_check_timestamp <= '0';

        -- The addition in this cycle should be included, so checking
        -- happens in the next cycle...
        check_timestamp <= '1';
      end if;

      -- One cycle delay, due to timestamp_less being latched.
      check_timestamp2 <= check_timestamp;
      -- One more cycle delay, due to timestamp add being split into
      -- low and high 32 bits.
      check_timestamp3 <= check_timestamp2;

      -- Default.
      i_mismatch_time <= '0';

      if (check_timestamp3 = '1') then
        -- If the local timestamp counter is in sync with the received
        -- values, then is shall have been less then the received value
        -- in the previous cycle, and be at or more than the received
        -- value in this cycle.

        if (timestamp_match = '1') then
          timestamp_sync.i.up   <= sender_period_match;
          timestamp_sync.i.down <= '0'; -- Would have taken priority.
        end if;

        -- If timestamp or period is wrong, we have a problem.
        i_mismatch_time <=
          (not timestamp_match) or (not sender_period_match);
      end if;

      -- Carry from previous cycle.
      timestamp_diff_lo_carry <= timestamp_diff_lo(32);
      -- Is the local timestamp counter less than the received value?
      -- Latched, to decouple the match logic from the diff calculation.
      timestamp_less <= timestamp_diff_hi(64);
      -- Was the local timestamp counter less then received last cycle.
      timestamp_less_prev <= timestamp_less;

      -- The receive value is held stable several cycles before it is
      -- used, so the additional delay due to this latched add value
      -- is harmless.
      -- On the sender side, the user can contribute a 22-bit value, and
      -- internally is contributed at most an 18-bit value.  Thus we
      -- could in principle have a 21-bit value, without the risk of
      -- overflowing above the 23 bits of result.
      timestamp_recv_off(22 downto 0) <=
        std_logic_vector(unsigned(timestamp_recv(22 downto 0)) +
                         unsigned(receive_delay_ns));
    end if;
  end process;

  i_synced <= timestamp_sync.o.synced;

  tick_ns_lo <= std_logic_vector(timestamp_counter_lo(31 downto 0));
  tick_ns_hi <= std_logic_vector(timestamp_counter_hi);
  aux_sigs   <= i_aux_signals;

  info_bit   <= i_info_bit;
  msg_strobe <= i_msg_strobe;

  sender_period_ns <= i_sender_period_ns;

  reset_add_counter_on_bit <= i_reset_add_counter_on_bit;

  synced <= i_synced;

  bad_hamming <= i_bad_hamming;
  mismatch_time <= i_mismatch_time;

  wrap_midbit <= i_wrap_midbit;

end RTL;


-- Sync between timestamp and got_bit / clk_add_ns{,_carry}:
--
-- Cycle '0':
--  Input:
--   got_bit                  : from outside '1', with accompanying
--   clk_add_ns               : from outside
--
-- On initial lock (on_bit_take_timestamp is set, so used
-- for next got_bit).
--
-- Cycle '0':
--  Input:
--   got_bit                  : from outside '1'
--   timestamp_counter        : initialised
--
-- When checking (on_bit_check_timestamp is set, so used
-- for a following got_bit):
--
-- Cycle '0':
--  Input:
--   got_bit                  : from outside '1'
--   clk_add_ns               : from outside
--   timestamp_counter        : previous
--  Latched
--   timestamp_counter       <= [timestamp_counter, clk_add_ns{,_carry}]
--   check_timestamp         <= '1'
--
-- Cycle '1':
--  Input:
--   check_timestamp          : '1'
--   timestamp_counter        : previous
--  Combinatorial:
--   timestamp_diff_lo        = [timestamp_counter]
--  Latched:
--   timestamp_diff_lo_carry <= [timestamp_diff_lo]
--   check_timestamp2        <= '1'
--
-- Cycle '2':
--  Input:
--   check_timestamp2         : '1'
--   timestamp_diff_lo_carry  :
--  Combinatorial:
--   timestamp_diff_hi        = [timestamp_diff_lo_carry, timestamp_counter]
--  Latched:
--   timestamp_less          <= [timestamp_diff_hi]
--   check_timestamp3        <= '1'
--
-- Cycle '3':
--  Input:
--   check_timestamp3         : '1'
--  Combinatorial:
--   timestamp_match          = based on timestamp_less
--   timestamp_match          = used to decide if timestamp is good
--  Latched:
--   i_mismatch_time         <= [timestamp_match] (is result)
