-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

-- The leading-edge of bit-slots are detected with a 0011-filter.
--
-- We track bitslots with a clock cycle period counter, i.e.
-- we determine how many clock cycles are used for each bit slot.
--
-- The tracking is driven by 0-1-transitions.  They come most often
-- when the bit pattern is ...010101..., i.e. one transition for
-- every two bitslots.  The least often is with a ...00110011...
-- pattern.
--
-- We are willing to phase-adjust (yank) the track counter by
-- half-a-unit each time a transition is detected too early or too
-- late.  Transitions may be 6 bit-slots apart.  Thus, the period
-- estimate must be good to 0.5/6 = 0.0833 units.  For most of the idle
-- pattern, transitions occur every two bits however, and for data
-- transmission the minimum is two and maximum four, with an average
-- of 2.67.
--
-- The initial guess is given by a min-max filter.  This explicitly
-- looks for the rapid sequence 01010101...01.  This comes for at least
-- every eighth period of the idle pattern, and then has at least
-- 20 one-zero transitions.  We want 17 transitions (16 periods).
-- This gives the time for 32 bit-slots.  We could expect the start
-- sample point to just be late, and the last to just be early.
-- That would give an error of -2/32 = 0.0625.  This should not happen;
-- the start and end samples would not be off in different directions.
--
-- With 84 slots, we can at most have 42 transitions each idle pattern.
-- Having the fast transition sequence at least every 8 patterns, means
-- 336 transitions between fast patterns.  It thus makes sense to search
-- for the minimum over 512 transitions.
--
-- After the initial guess, a better estimate is obtained by averaging
-- the counts of many 010101 sequences.




entity rataser_track_bitslots is
  generic(period_bits  : integer;
          frac_bits    : integer);
  port (clk      : in  std_logic;

        -- Input serial signal.
        receive  : in  std_logic;

        bit_val  :  out std_logic;
        got_bit  :  out std_logic;

        pattern_synced : in std_logic
        );
end rataser_track_bitslots;

architecture RTL of rataser_track_bitslots is

  -- Number of bits in estimation counter.
  -- Up to 16 transitions ago, which is 32 slots.
  constant est_counter_bits : integer := 5 + period_bits;

  -- Number of bits in the fractional counter.
  -- Additional bits come from how long the initial estimate it.
  constant counter_bits : positive := period_bits + frac_bits;

  -- Number of bits to describe the value to add each cycle.
  -- Two bits less than the period, since we do not handle signals
  -- faster than 4 cycles anyhow.
  -- (Empirically, locking does not work below x.yy clock cycles anyhow.)
  constant vpcs_bits    : positive := 1 + frac_bits;

  -- Previous receive value;
  signal last_receive : std_logic_vector(2 downto 0) := (others => '0');

  -- Have we just seen a leading-edge transition.
  signal le_transition : std_logic := '0';

  -- Is the current sample a 0 or 1.
  signal sample_high   : std_logic := '0';

  -- Transitions was seen at the expected location.
  signal had_good_trans  : std_logic := '0';
  -- Transition was seen early.
  signal had_early_trans : std_logic := '0';
  -- Transition was seen late.
  signal had_late_trans  : std_logic := '0';

  -- Add .5, 1 or 1.5 ?
  signal add_full : std_logic := '0';
  signal add_half : std_logic := '0';

  -- Count fractions of the bit received (and 16 periods)
  signal counter      : unsigned(counter_bits-1 downto 0) := (others => '0');
  -- Next value of counter
  signal counter_next : unsigned(counter_bits downto 0)   := (others => '0');
  -- Next value of counter, with target period subtracted.
  signal counter_next_wrap : unsigned(counter_bits downto 0) :=
    (others => '0');
  signal counter_next_wrap_half : unsigned(counter_bits downto 0) :=
    (others => '0');

  -- The actual addition.
  signal counter_add : unsigned(vpcs_bits downto 0) := (others => '0');

  -- Helper values to calculate wrap.
  signal counter_sub_period : unsigned(counter'range) := (others => '0');
  signal counter_sub_half_period : unsigned(counter'left-1 downto 0) :=
    (others => '0');

  signal counter_next_above_half : std_logic := '0';
  signal counter_next_above_half_prev : std_logic := '0';

  -- Did counter wrap?
  signal counter_next_wrapped : std_logic := '0';
  -- Did counter wrap?
  signal counter_next_wrapped_half : std_logic := '0';

  -- Target period.
  signal target_period : unsigned(counter'range) := (others => '0');

  -- Continuously counting counter, for initial period estimate.
  signal plain_counter : unsigned(est_counter_bits - 1 downto 0) :=
    (others => '0');

  -- Counter value for each transition up to 16 transitions ago.
  type counter_vector16 is array (0 to 16) of
    unsigned(plain_counter'range);
  signal plain_counter_prev_le : counter_vector16 :=
    (others => (others => '0'));

  -- Difference of counter since 16 transitions ago.
  signal plain_diff_16le : unsigned(plain_counter'range) :=
    (others => '0');
  -- Inject the the differnce into the estimation filter.
  signal take_diff_16le : std_logic := '0';

  -- Filtered estimate of the period for 16 rapid transitions.
  signal estimate_diff_16le : unsigned(plain_counter'range) :=
    (others => '0');
  -- New estimate available.
  signal new_estimate : std_logic := '0';

  -- Dummy signals.
  signal dummy_out     : unsigned(plain_counter'range) := (others => '0');
  signal dummy_max_min : unsigned(plain_counter'range) := (others => '0');

  -- Output values.
  signal i_bit_val : std_logic := '0';
  signal i_got_bit : std_logic := '0';

begin

  -- To not get too many spurious leading edges, we have a stricter
  -- requirement than just seeing a 0-1 transition.
  le_transition <= '1' when (last_receive = "011") else '0';

  -- Do a majority decision on three adjacent samples.
  -- Note: sample is taken half-into the bit-slot.
  sample_high <= '1' when (last_receive = "111" or
                           last_receive = "011" or
                           last_receive = "101" or
                           last_receive = "110") else '0';

  -- The next counter value is the current value, plus the addition
  -- of the full and/or half value-per-clock.
  counter_add(frac_bits    ) <= add_full;
  counter_add(frac_bits - 1) <= add_half;
  counter_add(frac_bits - 2 downto 0) <= (others => '0');
  -- Do the addition.
  counter_next <= ('0' & counter) + counter_add;

  -- In case we wrap (are larger than the target period), the new
  -- value should have the target subtracted.
  counter_sub_period <= target_period;
  -- Do the subtraction.
  counter_next_wrap <= counter_next - counter_sub_period;
  -- Wrapping is determined, if the subtracted value did not become negative.
  counter_next_wrapped <=
    not counter_next_wrap(counter_next_wrap'left);

  -- Similarly, we have passed half-the target value, in case subtraction
  -- of half the target value does not become negative.
  counter_sub_half_period <= target_period(target_period'left downto 1);
  -- Do the subtraction.
  counter_next_wrap_half <= counter_next - counter_sub_half_period;
  -- Are we above?
  counter_next_above_half <=
    not counter_next_wrap_half(counter_next_wrap_half'left);
  -- We may be above for several cycles, so only mark when passing.
  counter_next_wrapped_half <=
    counter_next_above_half and not counter_next_above_half_prev;


  -- Update the filter whenever a transition has been seen.
  -- The input value is the counter difference to 16 transitions ago.
  -- Take then minimum of 512 such values, and then the maximum of
  -- 8 such minima.
  est_filter: entity work.rataser_min_max_filter
    generic map(min_cycle_bits => 9,
                max_cycle_bits => 3)
    port map(clk => clk,

             input => plain_diff_16le,
             take  => take_diff_16le,

             output_min_max => estimate_diff_16le,
             output_new     => new_estimate,

             output         => dummy_out,
             output_max_min => dummy_max_min
             );

  process(clk)
  begin
    if (rising_edge(clk)) then

      last_receive(0) <= receive;
      last_receive(2 downto 1) <= last_receive(1 downto 0);

      -- Normally, we want to add 1
      add_full <= '1';
      add_half <= '0';

      -- Next counter value.
      -- The counter is never reset:  Provided that the period
      -- is nominally correct (within bounds), even if it is at
      -- the opposite phase, the phase-tracking will kick it
      -- faster towards the correct phase than it drifts
      -- into the other side of the opposite phase.
      if (counter_next_wrapped = '1') then
        counter <= counter_next_wrap(counter'range);
      else
        counter <= counter_next(counter'range);
      end if;

      counter_next_above_half_prev <= counter_next_above_half;


      -- Phase correction.
      if (counter_next_wrapped_half = '1') then
        if (had_late_trans = '1' and had_early_trans = '0' and
            had_good_trans = '0') then
          -- Correct us by half a period less.
          add_full <= '0';
          add_half <= '1';
        end if;
        if (had_early_trans = '1' and had_late_trans = '0' and
            had_good_trans = '0') then
          -- Correct us by half a period extra.
          add_half <= '1';
        end if;
      end if;

      if (counter_next_wrapped_half = '1') then
        -- At the half-way point, we reset if we have seen any early
        -- or late.  Seeing a transition now, counts as an early one
        -- for the next bit cycle.
        had_early_trans <= le_transition;
        had_late_trans  <= '0';
        had_good_trans  <= '0';
      elsif (counter_next_wrapped = '1') then
        -- We had a transition at the spot.  Do not kick things around.
        had_good_trans <= le_transition;
      else
        -- See any transition now?
        had_late_trans <=
          had_late_trans or
          (le_transition and not counter_next_above_half);
        had_early_trans <=
          had_early_trans or
          (le_transition and counter_next_above_half);
      end if;

      i_got_bit <= '0';

      -- Bits are sampled at middle of the bit-slot.
      if (counter_next_wrapped_half = '1') then
        i_got_bit <= '1';
        i_bit_val <= sample_high;
      end if;

      -- TODO: we do not need all this latching, could drop one layer?
      bit_val <= i_bit_val;
      got_bit <= i_got_bit;


      -- For initial estimate:

      -- Normally, we do not have a transition.
      take_diff_16le <= '0';
      -- The number of cycles since the 16 transitions ago.
      plain_diff_16le <= plain_counter - plain_counter_prev_le(16);

      if (le_transition = '1') then
        -- Remember the time of previous transition.
        plain_counter_prev_le(1) <= plain_counter;
        -- And shift all the further transition locations.
        plain_counter_prev_le(2 to 16) <=
          plain_counter_prev_le(1 to 15);
        -- We had another transition, take the value.
        take_diff_16le <= '1';
      end if;

      -- The plain counter is always counting.
      plain_counter <= plain_counter + 1;

      if (new_estimate = '1' and
          pattern_synced = '0') then
        -- We have a new estimate for the period.
        -- Only update when next-stage of receiver is not locked.
        target_period(counter'left downto 1) <= estimate_diff_16le;
        -- Since the filter is first doing a min, we expect it to
        -- be somewhat low.  Increase by half a unit.
        target_period(0) <= '1';
      end if;

    end if;
  end process;

end RTL;
