-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

-- Phase/frequency tracking.
--
-- The protocol gives an rising edge every cycle.
-- This is used to simplify the tracking.
--
-- Short-period tracking:
--
-- When the longer-period tracking has (currently) accepted the signal
-- as within phase-range, tracking is done by each cycle evaluating if
-- the edge is before or after the expected point.  This is evaluated
-- half-way between the expected points.  If a correction needs to
-- be made, then half the per-cycle add value is additionally added
-- or subtracted from the counter value, and after four more of one or
-- the other kind has been seen, the frequency estimate (per-cycle add
-- value) is adjusted by one unit.
--
-- The short-period tracking is thus able to handle cases when the
-- per-cycle add value is good to within half a clock cycle.
--
-- Long-period tracking:
--
-- Each time the counter wraps 8 expected pulses, we can evaluate if we
-- have also seen 8 actual pulses.  If 7 or fewer, or 9 or more have
-- been seen, then obviously the frequency estimate is far off, and a
-- larger correction should be applied.  The larger correction is scaled
-- such that for each four consequtive corrections in the same direction,
-- the correction is made twice larger.  (This ensures a speedy lock,
-- but is mostly useful for convenient simulations.)
--
-- This is not necessarily good enough for the short-period tracking.
-- Thus the long-period tracking will apply the same corrections as 
-- the short-period tracking, based on whether the 8th pulse is before
-- or after the expected slot.
--
-- This means as long as the per-cycle add value is further away than
-- 0.5 cycles per pulse period, the spot will wander around, and the
-- main correcting force is the long-term frequency adjustment.
--
-- Eventually the 8th pulse is within -2,..,2 cycles (i.e. 5 locations)
-- of the expected slot.  When this has been the case for several (n)
-- times, then the frequency estimate is good to better than 5/8/(n-1)
-- < 0.5 cycles, and thus handling can be given to the short-period
-- tracker.
--
-- The long-period tracking is always evaluated and thus no special
-- triggers are needed to evaluate overall tracking status and restart
-- in a rough tracking mode.
--
-- The above can work without resetting the counter.  The only caveat
-- is the situation where the frequency is (almost) correct, but the
-- phase is 180 degrees off.  Here the phase-tracking of adding/subtracting
-- an additional half per-cycle add value will however drive the phase
-- correct within 











-- How badly out of sync can we get?
--
-- The real period is (clock cycles per pulse):   p
-- Our counter has fractional bits:               n
-- Target counter period (i.e. counts per pulse): t
--
-- From recv_track_min_transition we have:
--
-- clocks_per_cycles:                          cpc   (p)
-- val_per_clock_scaled:                       vpcs  (= 2^n / p)
--
-- We add vpcs each clock cycle.  As this is rounded down, this
-- can miss (almost) one count each clock cycle, adding up to p counts
-- missing each period.  Out of expected t*2^n counts.  When does this
-- become a major disaster?  Definately at half a clock cycle.
-- Say at a quarter of a cycle:
--
-- p > 1/4*t*2^n   =>   n < log2(4*p/t) = 2 + log2(p) - log2(t)
--
-- Thus we want n > 2 + log2(p) - log2(t),
-- set t = 1 => n > 2 + log2(p).


entity rataser_track_bitstream is
  generic(frac_bits      : integer);
  port (clk      : in  std_logic;

        -- Input serial signal.
        receive  : in  std_logic;

        -- Pulse period (in ns - the target counter increment).
        pulse_period   : in std_logic_vector(9 downto 0);
        
        bit_val        :  out std_logic;
        got_bit        :  out std_logic;

        -- Reset counter to exact 0 on next bit.
        reset_counter_on_bit : in std_logic;
        -- Increment each clock cycle, only full bits.
        incr_per_clock       : out std_logic_vector;
        -- Contribution from the fractional bits.
        incr_per_clock_carry : out std_logic;

        -- The full value (for debugging).
        full_incr_per_clock  : out std_logic_vector;

        -- The filtered min pulse length (clock cyles) (for debugging).
        pulse_length_min     : out std_logic_vector;
        -- The filtered max pulse length (clock cyles) (for debugging).
        pulse_length_max     : out std_logic_vector;

        -- Continuous signal telling if we are (likely) good.
        synced_period :  out std_logic;

        pattern_synced : in std_logic;

        clear_failure     : in std_logic
        );
end rataser_track_bitstream;

architecture RTL of rataser_track_bitstream is

  -- Number of bits to describe the target period.
  constant period_bits   : positive := pulse_period'length;

  -- Number of bits in the fractional counter.  Can have two bits more,
  -- as we have no point in describing periods smaller than 4 clock cycles.
  -- (Empirically, locking does not work below x.yy clock cycles anyhow.)
  constant counter_bits  : positive := period_bits + frac_bits;

  -- Number of bits to describe the value to add each cycle.
  -- Two bits less than the period, since we do not handle signals
  -- faster than 4 cycles anyhow.
  -- (Empirically, locking does not work below x.yy clock cycles anyhow.)
  constant vpcs_bits     : positive := counter_bits - 2;

  -- Previous receive value;
  signal last_receive    : std_logic := '0';
  -- Leading-edge detector;
  signal le_transition   : std_logic := '0';

  -- Transition was seen early
  signal had_early_trans : std_logic := '0';
  -- Transition was seen late
  signal had_late_trans  : std_logic := '0';

  -- Add .5, 1 or 1.5 ?
  signal add_full        : std_logic := '0';
  signal add_half        : std_logic := '0';

  -- Add/sub for this period?
  signal phase_adj_do    : std_logic := '0';
  signal phase_adj_add   : std_logic := '0';
  signal freq_adj_do     : std_logic := '0';
  signal freq_adj_add    : std_logic := '0';
  -- Wandering add/sub for this period?
  signal wander_do       : std_logic := '0';
  signal wander_add      : std_logic := '0';
  -- Previous correction direction
  signal prev_wander_add : std_logic := '0';

  -- Accumulator for wander of correction.
  signal accum_periods_wander      : unsigned(4 downto 0) := (others => '0');
  -- Contibution this cycle.
  signal accum_periods_wander_term : unsigned(4 downto 0) := (others => '0');
  -- Enough to to an actual add or subtract?
  signal accum_periods_wander_add  : std_logic := '0';
  signal accum_periods_wander_sub  : std_logic := '0';

  -- Apply correction?
  signal adjust_do       : std_logic := '0';
  signal adjust_sub      : std_logic := '0';
  -- Corrections was too large last time?
  -- (Init to 1 such that we decrement even in simulation starting large.)
  signal adjust_overflow : std_logic := '1';

  -- Are we phase-tracking?
  signal phase_tracking  : std_logic := '0';

  -- Frequency.
  signal val_per_clock_scaled      : unsigned(vpcs_bits-1 downto 0) :=
    (others => '0');
  -- For division by two
  signal val_per_clock_sc_half     : unsigned(vpcs_bits-2 downto 0) :=
    (others => '0');

  -- Correction value (scaled up when correcting in the same direction,
  -- and not phase tracking).
  signal adjust_incr               : unsigned(frac_bits-1 downto 0) :=
    (others => '0');
  -- Extended to full range of val_per_clock_scaled
  signal adjust_incr_extend        : unsigned(vpcs_bits downto 0) :=
    (others => '0');
  -- Value to add
  signal val_per_clock_scaled_term : unsigned(vpcs_bits downto 0) :=
    (others => '0');
  -- Next value (if addition/subtraction applied)
  signal val_per_clock_scaled_next : unsigned(vpcs_bits downto 0) :=
    (others => '0');
  -- Overflow?
  signal val_per_clock_scaled_overflow : std_logic := '0';

  -- Target period (pulse_period, but clamped at minimum value).
  signal target_period : unsigned(pulse_period'range) := (others => '0');

  -- Reset the counter to exact 0 on next bit.
  signal reset_counter_next_bit : std_logic := '0';

  -- Count fractions of the bit received (and 16 periods).
  signal counter      : unsigned(counter_bits-1 downto 0) := (others => '0');
  -- Next value of counter.
  signal counter_next : unsigned(counter_bits downto 0)   := (others => '0');
  -- Next value of counter, with target period subtracted.
  signal counter_next_wrap      : unsigned(counter_bits downto 0) :=
    (others => '0');
  signal counter_next_wrap_half : unsigned(counter_bits downto 0) :=
    (others => '0');

  -- Helper values for adding and subtraction.
  signal counter_add_full : unsigned(val_per_clock_scaled'range) :=
    (others => '0');
  signal counter_add_half : unsigned(val_per_clock_sc_half'range) :=
    (others => '0');
  -- The actual addition.
  signal counter_add      : unsigned(vpcs_bits downto 0) := (others => '0');
  -- Did the addition carry a bit from the fractional part?
  signal counter_added_frac_carry : std_logic := '0';

  -- Helper values to calculate wrap.
  signal counter_sub_period      : unsigned(counter'range) := (others => '0');
  signal counter_sub_half_period : unsigned(counter'left-1 downto 0) :=
    (others => '0');

  signal counter_next_above_half      : std_logic := '0';
  signal counter_next_above_half_prev : std_logic := '0';

  -- Did counter wrap?
  signal counter_next_wrapped      : std_logic := '0';
  -- Did counter wrap?
  signal counter_next_wrapped_half : std_logic := '0';

  -- For detecting fast wrapping.
  signal counter_wrap_hist2 : std_logic_vector(1 downto 0) := (others => '0');

  -- Count of number of real transitions (max value is 8).
  signal transitions        : unsigned(3 downto 0) := (others => '0');
  -- Number of times in a row 8 transitions were found for 8 wraps.
  signal transitions_8_hist : unsigned(1 downto 0) := (others => '0');

  -- Count of number of counter wraps (wraps at 8 -> 0).
  signal counter_wraps : unsigned(2 downto 0) := (others => '0');

  -- Should we count the cycle towards the pulse being high?
  signal high_counting : std_logic := '0';

  -- Have we already started the counting (once)?
  signal high_count_started : std_logic := '0';

  -- Count of number of cycles a pulse signal is high.
  signal high_count    : unsigned(9 downto 0) := (others => '0');

  -- Count of number of cycles we have been high since half-slot
  -- (if still high) when starting next slot.
  signal high_count_init : unsigned(9 downto 0) := (others => '0');

  -- Did the init count already start the counting?
  signal high_count_init_started : std_logic := '0';

  -- Use value of high_count in filter.
  signal high_take     : std_logic := '0';

  -- The limit for declaring a pulse being a 1 bit.
  signal high_avg      : unsigned(9 downto 0) := (others => '0');

  -- Filter result, useful for debugging.
  signal debug_min_max : unsigned(9 downto 0) := (others => '0');
  signal debug_max_min : unsigned(9 downto 0) := (others => '0');

  -- Output values.
  signal i_bit_val     : std_logic := '0';
  signal i_got_bit     : std_logic := '0';

  signal i_incr_per_clock       : std_logic_vector(incr_per_clock'range) :=
    (others => '0');
  signal i_incr_per_clock_carry : std_logic := '0';

  signal incr2_per_clock        : std_logic_vector(incr_per_clock'range) :=
    (others => '0');
  signal incr2_per_clock_carry  : std_logic := '0';

begin

  le_transition <= receive and not last_receive;

  -- Extend value to full range
  adjust_incr_extend(adjust_incr_extend'left downto
                     adjust_incr'left+1) <= (others =>'0'); -- Silence warning.
  adjust_incr_extend(adjust_incr'range) <= adjust_incr;
  -- Two-complement value if subtracting.  (Then also need to add 1.)
  val_per_clock_scaled_term <=
    (adjust_incr_extend xor (val_per_clock_scaled_term'range => adjust_sub));
  -- The (possible) next value.
  val_per_clock_scaled_next <=
    ("0" & val_per_clock_scaled) +
    val_per_clock_scaled_term + ("0" & adjust_sub);
  -- Does the next value cause an overflow (wrap), such that it should
  -- not be used?
  val_per_clock_scaled_overflow <=
    val_per_clock_scaled_next(val_per_clock_scaled_next'left);

  -- Half value by shift down one bit.
  val_per_clock_sc_half <=
    val_per_clock_scaled(val_per_clock_scaled'left downto 1);

  -- The next counter value is the current value, plus the addition
  -- of the full and/or half value-per-clock.
  counter_add_full <=
    val_per_clock_scaled when (add_full = '1') else (others => '0');
  counter_add_half <=
    val_per_clock_sc_half when (add_half = '1') else (others => '0');
  -- The value to add.
  counter_add <= ('0' & counter_add_full) + counter_add_half + 1;
  -- Do the addition.
  counter_next <= ('0' & counter) + counter_add;
  -- Carry from the fractional part?
  counter_added_frac_carry <=
    counter(frac_bits) xor
    counter_next(frac_bits) xor
    counter_add(frac_bits);

  -- In case we wrap (are larger than the target period), the new
  -- value should have the target subtracted.
  counter_sub_period(counter_sub_period'left - period_bits downto
                     0) <= (others => '0'); -- Silence warning.
  counter_sub_period(counter_sub_period'left downto
                     counter_sub_period'left - period_bits + 1) <=
    target_period;
  -- Do the subtraction.
  counter_next_wrap <= counter_next - counter_sub_period;
  -- Wrapping is determined, if the subtracted value did not become
  -- negative.  This includes becoming zero means wrapped (which is
  -- wanted, i.e perfect wrap).
  counter_next_wrapped <=
    not counter_next_wrap(counter_next_wrap'left);

  -- Similarly, we have passed half-the target value, in case subtraction
  -- of half the target value does not become negative.
  counter_sub_half_period(counter_sub_half_period'left - period_bits downto
                          0) <= (others => '0'); -- Silence warning.
  counter_sub_half_period(counter_sub_half_period'left downto
                          counter_sub_half_period'left - period_bits + 1) <=
    target_period;
  -- Do the subtraction.
  counter_next_wrap_half <= counter_next - counter_sub_half_period;
  -- Are we above?
  counter_next_above_half <=
    not counter_next_wrap_half(counter_next_wrap_half'left);
  -- We may be above for several cycles, so only mark when passing.
  counter_next_wrapped_half <=
    counter_next_above_half and not counter_next_above_half_prev;

  -- Accumulation contribution, either 1 or -1, if wander_do.
  accum_periods_wander_term(0) <= wander_do;    -- 1 if add or sub
  accum_periods_wander_term(accum_periods_wander_term'left downto 1) <=
    (others => (wander_do and not wander_add)); -- all other 1 if sub

  -- Accumulated phase corrections suggest frequency correction?
  -- When negative (i.e. in total < -4 due to start at 4), sub.
  accum_periods_wander_sub <=
    '1' when (accum_periods_wander(4) = '1') else '0';
  -- When > 8 (i.e. in total > 4 due to start at 4), add.
  accum_periods_wander_add <=
    '1' when ((accum_periods_wander(4 downto 3) = "01") and
              (accum_periods_wander(2 downto 0) /= "000")) else '0';

  -- Did we have a call to adjust the per-cycle value?
  wander_do <=
    phase_adj_do or freq_adj_do;
  -- Handle in precendence order (freq adj. more important than phase).
  wander_add <=
    freq_adj_add when (freq_adj_do = '1') else phase_adj_add;

  -- Use high count in filter when we wrap the counter.
  high_take <= counter_next_wrapped;

  high_filter: entity work.rataser_min_max_filter
    generic map(min_cycle_bits => 4,
                max_cycle_bits => 3)
    port map(clk => clk,

             input => high_count,
             take  => high_take,

             output => high_avg,

             output_min_max => debug_min_max,
             output_max_min => debug_max_min
             );

  process(clk)
  begin
    if (rising_edge(clk)) then

      last_receive <= receive;
      
      -- Normally, we want to add 1 * val_per_clock_scaled
      add_full <= '1';
      add_half <= '0';

      -- We are normally not adjusting the frequency.
      adjust_do  <= '0';
      adjust_sub <= '0';

      -- Or suggesting to do (phase tracking).
      phase_adj_do  <= '0';
      phase_adj_add <= '0';
      -- Or suggesting to do (freq tracking).
      freq_adj_do   <= '0';
      freq_adj_add  <= '0';

      -- The target period is the value given from outside.
      target_period <= unsigned(pulse_period);
      -- Unless the value is (illegaly) small, which it may be due
      -- to bad previous message.  Too small value means we might not
      -- be able to lock onto a new signal, and thus get a good value.
      -- Clamp to minimum value of (at least) 8.
      if (unsigned(pulse_period) < 8) then
        target_period(3) <= '1';
      end if;

      -- Next counter value.
      if (counter_next_wrapped = '1') then
        if (reset_counter_next_bit = '1' or
            counter_wrap_hist2 /= "00") then
          -- This allows the increment to be in sync for the timestamp
          -- tracker.
          -- We also abuse this reset if we are wrapping too often.
          -- Helps against counter growing insanely large compared to
          -- target such that it does not decrement itself in one cycle.
          -- (See other use of counter_wrap_hist2 below.)
          counter <= (others => '0');
          reset_counter_next_bit <= '0';
        else
          counter <= counter_next_wrap(counter'range);
        end if;
      else
        counter <= counter_next(counter'range);
      end if;

      if (reset_counter_on_bit = '1') then
        reset_counter_next_bit <= '1';
      end if;
      
      counter_next_above_half_prev <= counter_next_above_half;

      -- Count how many real transitions we have seen (but do not count
      -- above 12 (cheap saturation; only need to get to 9)).
      transitions <= transitions +
                     ("0" & (le_transition and (not transitions(3) or
                                                not transitions(2))));

      -- Count how many times we have wrapped the target count.
      counter_wraps <= counter_wraps + ("0" & counter_next_wrapped);

      if (counter_next_wrapped_half = '1') then

        if (phase_tracking = '1' or
            counter_wraps = 0) then
          -- If we are doing phase tracking, or we reached the
          -- 8th pulse, do the half-extra add/sub phase adjustment.
          if (had_late_trans = '1' and had_early_trans = '0') then
            -- Correct us by half a period less.
            add_full <= '0';
            add_half <= '1';
            -- Slow frequency correction.
            phase_adj_do  <= '1';
            phase_adj_add <= '0'; -- sub (default)
          end if;
          if (had_early_trans = '1' and had_late_trans = '0') then
            -- Correct us by half a period extra.
            add_half <= '1';
            -- Slow frequency correction.
            phase_adj_do  <= '1';
            phase_adj_add <= '1';
          end if;
        end if;

        if (counter_wraps = 0) then
          -- When we have seen 8 counter wraps, and are mid-period
          -- we can determine if we are too slow or too fast.

          -- Note: when we have reached an approximate frequency,
          -- none of these tests will trigger.  Only the phase-tracking
          -- above operates.

          if (transitions < 8) then
            -- We are counting too fast.
            -- But not if previous round was phase tracking (hysteresis).
            freq_adj_do  <= not phase_tracking;
            freq_adj_add <= '0'; -- sub (default)
          end if;
          if (transitions > 8) then
            -- We are counting too slowly.
            -- But not if previous round was phase tracking (hysteresis).
            freq_adj_do  <= not phase_tracking;
            freq_adj_add <= '1';
          end if;

          if (transitions = 8) then
            if (transitions_8_hist = "11") then
              -- We have seen 8 matching transitions 4 times more than
              -- not recently, so enable phase tracking.
              phase_tracking <= '1';
            else
              transitions_8_hist <= transitions_8_hist + 1;
            end if;
          else
            if (transitions_8_hist = "00") then
              -- 8 matching transitions lost twice, give up phase tracking.
              phase_tracking <= '0';
            else
              transitions_8_hist <= transitions_8_hist - 1;
            end if;
          end if;

          -- Reset the transition counter.  If we just had a transition,
          -- do not forget it however.
          transitions <= (0 => le_transition,
                          others => '0');
        end if;
      end if;

      if (phase_tracking = '0' and
          transitions(3 downto 2) = "11") then
        -- We are counting way too slowly, and not phase tracking.
        -- This is not strictly need, but speeds up simulations a lot.
        -- (Without, we are limited by the counter wrapping.  Now we
        -- are driven by the incoming transitions.)
        freq_adj_do  <= '1';
        freq_adj_add <= '1';
        -- Reset the transition counter.  If we reset to 0, then
        -- the next (real) handling above might notice a too small
        -- transition count.  So we reset to 8.
        transitions <= (3 => '1', others => '0');
      end if;

      if (counter_next_wrapped = '1' and
          counter_wrap_hist2 /= "00") then
        -- We have wrapped at least twice in the last three cycles.
        -- We are too fast, must subtract.
        -- (This needed subtraction of the frequency is not always done
        -- by the rules above, when counter_next_wrapped_half fails to
        -- trigger.  Therefore this rule.)
        freq_adj_do  <= '1';
        freq_adj_add <= '0';
      end if;

      -- History of 2 most recent clock cycles.
      counter_wrap_hist2 <= counter_wrap_hist2(0) & counter_next_wrapped;

      if (counter_next_wrapped_half = '1') then
        -- At the half-way point, we reset if we have seen any early
        -- or late.  Seeing a transition now, counts as an early one
        -- for the next bit cycle.
        had_early_trans <= le_transition;
        had_late_trans  <= '0';
      else
        -- See any transition now?
        had_late_trans <=
          had_late_trans or
          (le_transition and not counter_next_above_half);
        had_early_trans <=
          had_early_trans or
          (le_transition and counter_next_above_half);
      end if;

      -- Update the frequency wander seen so far
      accum_periods_wander <=
        accum_periods_wander + accum_periods_wander_term;

      if (wander_do = '1') then
        -- If we are doing phase tracking, or the correction is
        -- in the other direction, set the adjustment to just 1.
        if (wander_add /= prev_wander_add or
            phase_tracking = '1' or
            adjust_overflow = '1') then
          -- Reset the correction.
          adjust_incr <= (others => '0');
          adjust_incr(0) <= '1';
          adjust_overflow <= '0';
        elsif (accum_periods_wander_sub = '1' or
               accum_periods_wander_add = '1') then
          -- If we have done a number of corrections,
          -- then double the size of the correction.
          -- (But do not shift out the '1' bit).
          adjust_incr <=
            (adjust_incr(adjust_incr'left) or
             adjust_incr(adjust_incr'left-1)) &
            adjust_incr(adjust_incr'left-2 downto 0) & '0';
        end if;
        -- Remember previous correction direction.
        prev_wander_add <= wander_add;

        -- Correct frequency?
        if ((freq_adj_do = '1' and freq_adj_add = '0') or
            (freq_adj_do = '0' and accum_periods_wander_sub = '1')) then
          -- Frequency adjustment, or
          -- have seen > 4 more phase subtractions than additions.
          adjust_do  <= '1';
          adjust_sub <= '1';
        end if;
        if ((freq_adj_do = '1' and freq_adj_add = '1') or
            (freq_adj_do = '0' and accum_periods_wander_add = '1')) then
          -- Frequency adjustment, or
          -- have seen > 4 more phase additions than subtractions.
          adjust_do  <= '1';
          adjust_sub <= '0'; -- add (default)
        end if;
        
        if (accum_periods_wander_sub = '1' or
            accum_periods_wander_add = '1') then
          accum_periods_wander <= "00100";
        end if;
      end if;

      -- The adjustment itself happens one cycle later.
      if (adjust_do = '1') then
        if (val_per_clock_scaled_overflow = '1') then
          -- Possibly the correction is too large.
          -- Do not perform, and force adjustment to minimum next time.
          adjust_overflow <= '1';
        else
          val_per_clock_scaled <=
            val_per_clock_scaled_next(val_per_clock_scaled'range);
        end if;
      end if;

      -- Default.
      i_got_bit <= '0';

      -- Bits are sampled at the expected leading edge, by checking
      -- how long (how many high) the previous pulse was (had).
      if (counter_next_wrapped = '1') then
        i_got_bit <= '1';

        if (high_count > high_avg) then
          i_bit_val <= '1';
        else
          i_bit_val <= '0';
        end if;

        -- The counting may have started anytime since the last
        -- mid-slot position (just one cycle earlier is not always
        -- enough).  Else, we will count whatever comes from the first
        -- high location, since it may be later than just the next
        -- cycle.  Counting then continues as long as the signal is
        -- high.
        high_count <= high_count_init + ("" & receive);
        -- Did the start already happen, or just now?
        high_count_started <= high_count_init_started or receive;
        -- If we have started (earlier or now), we are counting.
        high_counting <= high_count_init_started or receive;
      else
        -- Count if we are high, and previous cycle counted, or we
        -- have not started before.
        high_count <=
          high_count + ("0" & (receive and (high_counting or
                                            not high_count_started)));
        -- Whenever we see a pulse, we have started.
        high_count_started <= high_count_started or receive;
        -- Continue to count if are high.
        high_counting <= receive and (high_counting or
                                      not high_count_started);
      end if;

      if (counter_next_wrapped_half = '1' or
          receive = '0') then
        -- Reset pre-nominal counter at mid-slot, or whenever signal
        -- is low.
        high_count_init <= (others => '0');
        high_count_init_started <= '0';
      elsif (receive = '1') then
        -- Continue counting pre-nominal while high.
        high_count_init <= high_count_init + 1;
        high_count_init_started <= '1';
      end if;

      -- In order to not depend on the addition to get the carry bit,
      -- we latch the incr_per_clock{,_carry} signals.  We therefore
      -- also latch the output bit signals, such that they are in sync.

      -- TODO: we do not need all this latching, could drop one layer?

      bit_val <= i_bit_val;
      got_bit <= i_got_bit;

      incr2_per_clock <=
        (incr2_per_clock'range => '0');
      incr2_per_clock(vpcs_bits-frac_bits downto 0) <=
        std_logic_vector(counter_add(vpcs_bits downto frac_bits));
      incr2_per_clock_carry  <= counter_added_frac_carry;

      i_incr_per_clock       <= incr2_per_clock;
      i_incr_per_clock_carry <= incr2_per_clock_carry;

    end if;
  end process;

  incr_per_clock       <= i_incr_per_clock;
  incr_per_clock_carry <= i_incr_per_clock_carry;

  full_incr_per_clock  <= std_logic_vector(val_per_clock_scaled);

  synced_period        <= phase_tracking;

  pulse_length_min     <= std_logic_vector(debug_min_max);
  pulse_length_max     <= std_logic_vector(debug_max_min);

end RTL;


-- Sync between got_bit and incr_per_clock{,_carry} (vs. counter):
--
-- When counter is reset to 0, it is equivalent to wrapping.
--
-- In cycle '0', value of:
--  Input:
--   counter_add              : causes the current counter_next, which causes
--   counter                  : previous
--  Combinatorial:
--   counter_next             : [counter, counter_add]
--   counter_next_wrapped     : [counter_next_wrapped]
--  Latched:
--   i_got_bit               <= [counter_next_wrapped] gives '1' (or '0')
--   incr2_per_clock         <= [counter_add]
--
-- In cycle '1', we get these latched values:
--  Input:
--   incr_per_clock           : just a copy of the above
--  Latched:
--   got_bit                 <= [i_got_bit]
--   i_incr_per_clock        <= [incr2_per_clock]
