-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity rataser_track_edge is
  port (clk:           in std_logic;

        last_edge:     in std_logic_vector(1 downto 0);

        synced_pulse:  in std_logic;
        synced_sym24:  in std_logic;
        synced_time:   in std_logic;

        update_pulse:  in std_logic;

        track_edge:    out std_logic_vector(1 downto 0)
        );
end;

-- Each time we get an update_pulse, we latch the last edge into history.
--
-- We also compare last_edge to the history, and to consider it
-- as an candidate for a good new edge, at least two more items
-- in the history must match (i.e. we have a majority).  And the
-- previous edge must not be completely opposite.

-- To keep lock if we have seen sync status during an update,
-- we will not update as long as that sync status remains.


architecture RTL of rataser_track_edge is

  signal hist1_edge : std_logic_vector(1 downto 0) := "11";
  signal hist2_edge : std_logic_vector(1 downto 0) := "11";
  signal hist3_edge : std_logic_vector(1 downto 0) := "11";
  signal hist4_edge : std_logic_vector(1 downto 0) := "11";

  -- Do not use history until filled.  Mainly avoids lock at start.
  signal hist_valid : std_logic_vector(4 downto 1) := "0000";

  signal hist_ok : std_logic := '0';

  signal is_synced_pulse : std_logic := '0';
  signal is_synced_sym24 : std_logic := '0';
  signal is_synced_time  : std_logic := '0';

  signal match : std_logic_vector(4 downto 1) := (others => '0');

  signal match_two_of_hist : std_logic := '0';

  signal prev_not_opposite : std_logic := '0';

  signal update_pulse2 : std_logic := '0';

  signal drop_is_sync : std_logic := '0';

  signal i_track_edge : std_logic_vector(1 downto 0) := "11";

begin

  match(1) <= '1' when (hist1_edge = last_edge) else '0';
  match(2) <= '1' when (hist2_edge = last_edge) else '0';
  match(3) <= '1' when (hist3_edge = last_edge) else '0';
  match(4) <= '1' when (hist4_edge = last_edge) else '0';

  process(clk)
  begin
    if (rising_edge(clk)) then

      if (update_pulse = '1') then
        hist1_edge <= last_edge;
        hist2_edge <= hist1_edge;
        hist3_edge <= hist2_edge;
        hist4_edge <= hist3_edge;

        hist_valid(1) <= '1';
        hist_valid(4 downto 2) <= hist_valid(3 downto 1);
        
        match_two_of_hist <= (match(1) and match(2)) or
                             (match(1) and match(3)) or
                             (match(1) and match(4)) or
                             (match(2) and match(3)) or
                             (match(2) and match(4)) or
                             (match(3) and match(4));

        -- To not be opposite to the previous, the lowest bit must differ,
        -- or the highest bit must be the same.
        if ((last_edge(0) /= hist1_edge(0)) or
            (last_edge(1)  = hist1_edge(1))) then
          prev_not_opposite <= '1';
        else
          prev_not_opposite <= '0';
        end if;
      end if;

      if (hist_valid = "1111") then
        hist_ok <= '1';
      else
        hist_ok <= '0';
      end if;

      update_pulse2 <= update_pulse;

      if (update_pulse2 = '1') then

        if (match_two_of_hist = '1' and
            prev_not_opposite = '1' and
            hist_ok = '1') then
          -- We are majority of history, and ok relative to previous

          -- If we are not yet in some kind of sync, update the suggested
          -- edge.
          if (is_synced_pulse = '0' and
              is_synced_sym24 = '0' and
              is_synced_time  = '0') then
            i_track_edge <= hist1_edge;
          end if;

          -- If we are (not) updating, and have a lowest kind of sync
          -- reported, then hold that sync status.
          if (i_track_edge = hist1_edge and
              is_synced_pulse = '0') then
            is_synced_pulse <= '1';
          end if;
        end if;

        -- Regardless of what history says - if we are already synced,
        -- we can upgrade our sync level.
        if (is_synced_pulse = '1' and
            synced_sym24 = '1') then
          is_synced_sym24 <= '1';
        end if;

        if (is_synced_sym24 = '1' and
            synced_time = '1') then
          is_synced_time <= '1';
        end if;
      end if;

      drop_is_sync <= '0';
      -- When we loose sync status again, we consider to drop sync status.
      -- But we only drop sync status if no higher sync status has been
      -- reached.  This is such that we do not give up if a lower level
      -- temporarily looses sync.  If we really are bad, then the upper
      -- levels will soon enough loose sync too.
      if ((is_synced_pulse = '1' and
           is_synced_sym24 = '0' and
           is_synced_time  = '0' and
           synced_pulse = '0') or
          (is_synced_sym24 = '1' and
           is_synced_time  = '0' and
           synced_sym24 = '0') or
          (is_synced_time  = '1' and
           synced_time = '0')) then
        -- Drop sync one cycle later, to get less logic in one cycle.
        drop_is_sync <= '1';
      end if;

      if (drop_is_sync = '1') then
        -- But when we drop sync status, we drop it completely.
        is_synced_pulse <= '0';
        is_synced_sym24 <= '0';
        is_synced_time <= '0';
        -- Force recollection of history before using it.
        hist_valid <= "0000";
      end if;

    end if;
  end process;

  track_edge <= i_track_edge;  
end RTL;
