-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.rataser_records.all;
use work.rataser_sym24word8_util;

entity rataser_trig_recv is
  generic(-- Number of bits for pulse period.
          period_bits:   integer;
          -- Pipeline stages to handle diff calculation.
          pipeline_sym_diff_in   : boolean := false;
          pipeline_sym_diff_1    : boolean := false;
          pipeline_sym_diff_2    : boolean := false;
          pipeline_sym_diff_out  : boolean := false;
          pipeline_sym_match_out : boolean := false;
          pipeline_sym_pick      : boolean := false;
          pipeline_sym_out       : boolean := false
          );
  port (clk:           in std_logic;

        -- Input serial signal.
        receive:       in std_logic;

        -- Output data bits.
        data:          out std_logic_vector(10 downto 0);
        -- Strobe signal that we have decoded a new message.
        msg_strobe:    out std_logic;

        -- Auxiliary information.
        info_bit:      out std_logic;
        -- Strobe signal that we have gotten a new info bit.
        info_strobe:   out std_logic;

        -- Report current sync status.
        sync_status:   out std_logic;
        -- Report if any sync status has been lost since clear.
        sync_lost:     out std_logic;
        -- Report number of bad bits (might be underestimate) since clear.
        bad_bit_count: out std_logic_vector;
        -- Clear sync loss errors.
        clear_status:  in std_logic
        );
end;

architecture RTL of rataser_trig_recv is

  signal bit_val        : std_logic := '0';
  signal got_bit        : std_logic := '0';

  signal mark_w16       : std_logic := '0';
  signal out_w16p       : unsigned(15 downto 0) := (others => '0');
  signal out_w16n       : unsigned(15 downto 0) := (others => '0');

  signal sym_synced     : std_logic := '0';
  signal sym_bad_bit    : std_logic;

  signal hamm_data_16p  : std_logic_vector(15 downto 0);
  signal hamm_data_16n  : std_logic_vector(15 downto 0);

  signal hamm_word11    : std_logic_vector(10 downto 0);
  signal hamm_good      : std_logic;
  signal hamm_single_source : std_logic;
  signal hamm_bad_bit   : std_logic;

  signal i_data       : std_logic_vector(10 downto 0) := (others => '0');
  signal i_msg_strobe : std_logic := '0';

  -- Pipeline to get the input into the clock domain, anti-metastable
  -- and one cycle extra.
  signal receive_delayed : std_logic_vector(2 downto 0) := (others => '0');

begin

  -- Get value into clock domain.
  process(clk)
  begin
    if (rising_edge(clk)) then
      receive_delayed <= receive_delayed(1 downto 0) & receive;
    end if;
  end process;

  -- Track the frequency of the sender, and evaluate the bits.
  track_slots: entity work.rataser_track_bitslots
    generic map(period_bits => period_bits,
                frac_bits => 6)
    port map(clk            => clk,
             receive        => receive_delayed(2),
             bit_val        => bit_val,
             got_bit        => got_bit,
             pattern_synced => sym_synced
             );

  decode_sym: entity work.rataser_symburst_decode
    generic map(pipeline_diff_in   => pipeline_sym_diff_in,
                pipeline_diff_1    => pipeline_sym_diff_1,
                pipeline_diff_2    => pipeline_sym_diff_2,
                pipeline_diff_out  => pipeline_sym_diff_out,
                pipeline_match_out => pipeline_sym_match_out,
                pipeline_pick      => pipeline_sym_pick,
                pipeline_out       => pipeline_sym_out)
    port map(clk => clk,

             bit_val => bit_val,
             got_bit => got_bit,

             mark_w16       => mark_w16,
             out_w16p       => out_w16p,
             out_w16n       => out_w16n,

             synced         => sym_synced,
             bad_bit        => sym_bad_bit
             );

  hamm_data_16p <=     std_logic_vector(out_w16p);
  hamm_data_16n <= not std_logic_vector(out_w16n);

  decode_hamm: entity work.rataser_hamming_decode_dual
    port map(message_p     => hamm_data_16p,
             message_n     => hamm_data_16n,

             data          => hamm_word11,
             good          => hamm_good,
             single_source => hamm_single_source,
             bad_bit       => hamm_bad_bit);

  process(clk)
  begin
    if (rising_edge(clk)) then
      i_msg_strobe <= '0';

      if (mark_w16 = '1') then
        if (hamm_good = '1') then
          i_data <= hamm_word11;
          i_msg_strobe <= '1';
        end if;
      end if;
    end if;
  end process;

  data       <= i_data;
  msg_strobe <= i_msg_strobe;

end RTL;
