-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.rataser_records.all;
use work.rataser_symburst_util.all;

-- During idle the transmit contains the 24-slot idle symbol.  The idle
-- symbol with all payload bits as 0 looks like (. denote 01 transitions,
-- and , denote 10 transitions - which would be seen instead in case the
-- signal is inverted):
--
-- 1aAbB0vcCdD01eEfFV0gGhH1
-- 101010v0101010101V001011
--   . .   . . . . .   . .  
--  , , ,   , , , ,     ,

-- To provide the receiver a sequence of also five 10-transitions
-- (giving four minimal periods between transitions), the v bit is taken
-- as 0w11 (i.e. 110w) in four successive idle pattern, where w is
-- the actual aux bit.  The 1 ensures a pattern holding seven
-- transitions of both 01 and 10 kind irrespective of w:
--
-- 101010101010101010001011
--
-- When a transmission is made, either the last half of the 24-slot
-- idle symbol, or 12 slots of a new have payload bits 1.  This is
-- followed by 32 raw data slots (2x16 bits, including inversion).
-- To keep alignment for the 24-slot symbols, the transmission is
-- followed by either 16 or 4 filler slots.
--
-- With the filler consisting of a repeated 10-pattern, it ensures
-- the presence also of a five sequences of 10 transitions:
--
-- fill          idle
-- ====------------------------
-- 1010101010v0101010101V001011
--  , , , , ,  . . . . .





entity rataser_trig_send is
  generic(-- Number of bits for pulse period.
          period_bits:   integer
          );
    port (clk:           in std_logic;

          -- Input data bits.
          data:          in std_logic_vector(10 downto 0);
          -- Flag to initiate a data burst transmission.
          send:          in std_logic;

          -- Auxiliary information.
          info_bit:      in std_logic;

          -- The data burst is waiting for or being sent.
          data_send:     out std_logic := '0';
          -- A new auxiliary information bit should be provided.
          info_new:      out std_logic := '0';

          -- Slot period (in clk cycles, controls sender).
          slot_period_clks:  in std_logic_vector(period_bits-1 downto 0);

          -- For testing of faults only, yank running send counter.
          fault_shift_count: in std_logic_vector(4 downto 0) :=
            (others => '0');

          -- Output message.
          transmit:      out std_logic := '0';
          -- Oscilloscope sync signal for transmit message.
          transmit_sync: out std_logic := '0';
          -- Oscilloscope sync signal for idle message.
          idle_sync:     out std_logic := '0'
          );
end;

architecture RTL of rataser_trig_send is

  -- Latched single signal to transfer.
  signal data_latch : std_logic_vector(10 downto 0) := (others => '0');
  -- Latched auxiliary information.
  signal info_bit_latch : std_logic := '0';
  -- When auxiliary info bit transmission is delayed by data.
  signal data_info_bit_latch : std_logic := '0';

  -- Input payload (message) to Hamming encoder (may also be 3 downto 0).
  signal hamm_word11 : std_logic_vector(10 downto 0) := (others => '0');
  -- Output block from Hamming encoder (may also be 7 downto 0).
  signal hamm_data16 : std_logic_vector(15 downto 0) := (others => '0');

  -- Latched hamming block.
  signal hamm_data16_latch : std_logic_vector(15 downto 0) := (others => '0');

  constant sym16 : std_logic_vector(15 downto 0) := "1000111001001110";

  -- Idle pattern should send match symbol.
  signal idle_sym     : std_logic := '0';
  -- Idle pattern should send aux info bit.
  signal idle_info    : std_logic := '0';
  -- Idle pattern is sending long 10 pattern.
  signal idle_10_long : std_logic := '0';

  -- Transmit bit when sending match pattern.
  signal bit_out_sym  : std_logic;
  -- Transmit bit when sending info bit.
  signal bit_out_info : std_logic;
  -- Transmit bit when sending data-delayed info bit.
  signal bit_out_data_info : std_logic;
  -- Transmit bit when sending 10 pattern.
  signal bit_out_10   : std_logic;
  -- Transmit bit when sending idle pattern.
  signal bit_out_idle : std_logic;
  -- Transmit bit when sending data.
  signal bit_out_data : std_logic;
  -- Bit to transmit, based on current count value.
  signal bit_out      : std_logic;

  -- Value (raw bit) to transmit (internal).
  signal i_transmit      : std_logic := '0';
  -- Scope-watching sync signal for transmit sequence.
  signal i_transmit_sync : std_logic := '0';
  -- Scope-watching sync signal for idle sequence.
  signal i_idle_sync     : std_logic := '0';

  -- Counter of clock cycles.
  signal count_clk      : unsigned(period_bits-1 downto 0) := (others => '0');
  signal count_clk_next : unsigned(period_bits   downto 0) := (others => '0');
  signal count_clk_wrap : std_logic := '0';

  signal count_clk_max  : unsigned(period_bits-1 downto 0) := (others => '0');
  
  -- Send counter, runs through 0-127.  It always jumps past values
  -- 42..63 and 64+42..127.  Effectively, the period is 84.
  -- At 41, the add value is 23.
  signal count_sym_add      : unsigned(4 downto 0) := "00000";
  -- Should next add do the big add.
  signal count_sym_add_jump : std_logic := '0';

  signal count_sym         : unsigned(6 downto 0) := (others => '0');
  signal count_sym_next    : unsigned(7 downto 0) := (others => '0');

  signal count_sym_wrap8   : std_logic := '0';
  signal count_sym_wrap128 : std_logic := '0'; -- Is really wrap 84.

  signal count_data        : unsigned(4 downto 0) := (others => '0');
  signal count_data_next   : unsigned(5 downto 0) := (others => '0');
  signal count_data_wrap2  : std_logic := '0';
  signal count_data_wrap32 : std_logic := '0';

  -- When can a send be started?
  signal start_loc   : std_logic := '0';

  -- Would a started transmission reach to clobber a match symbol
  -- or long 10 pattern?
  signal reach_sym_next : std_logic := '0';
  signal reach_sym_this : std_logic := '0';
  signal reach_10       : std_logic := '0';

  -- Have we clobbered multiple match symbols or long 10 patterns?
  signal recent_clobber_sym_next : std_logic := '0';
  signal recent_clobber_sym_this : std_logic := '0';
  signal recent_clobber_10  : std_logic := '0';

  -- Is it ok to start a messge?
  signal can_start   : std_logic := '0';

  -- Transmission phase (if all 0, then idle pattern).
  -- Waiting in idle to boundary for start pattern (0-11 slots).
  signal send_wait   : std_logic := '0';
  -- Sending start pattern (12 slots).
  signal send_start  : std_logic := '0';
  -- Sending raw data, each bit and inverted (2x16 slots).
  signal send_data   : std_logic := '0';
  -- Always send info bits after data.
  signal send_info   : std_logic := '0';

  -- Which parts of the idle patterns have recently been clobbered by
  -- transmissions?
  signal clobber_sym       : std_logic := '0';
  signal clobber_10        : std_logic := '0';
  signal clobber_prev_sym  : std_logic := '0';
  signal clobber_prev_10   : std_logic := '0';
  signal clobber_prev2_sym : std_logic := '0';
  signal clobber_prev2_10  : std_logic := '0';
  signal clobber_prev3_sym : std_logic := '0';
  signal clobber_prev3_10  : std_logic := '0';

  -- Count bits 0..4  [0..31]  determine which bit of 24-slot pattern

  -- Pipelined
  signal count_sym_p  : unsigned(6 downto 0) := (others => '0');
  signal count_data_p : unsigned(4 downto 0) := (others => '0');
  signal send_start_p : std_logic := '0';
  signal send_data_p  : std_logic := '0';
  signal send_info_p  : std_logic := '0';

begin

  -- Maximum value (start) of pulse counter is pulse period - 1.
  count_clk_max <= unsigned(slot_period_clks) - 1;
  
  -- Counter for the pulse itself (down-count)
  count_clk_next <= ('0' & count_clk) - 1;
  count_clk_wrap <= count_clk_next(count_clk_next'high);

  -- Add 23 when we start from 41, otherwise 1.
  count_sym_add(4 downto 1) <=
    "1011" when (count_clk_wrap = '1' and
                 count_sym_add_jump = '1') else
    "0000";
  count_sym_add(0) <= count_clk_wrap;
  -- Do the addition.  
  count_sym_next   <= ('0' & count_sym) + count_sym_add
                      + unsigned(fault_shift_count);
  count_sym_wrap8   <= count_sym_next(3) xor count_sym(3);
  count_sym_wrap128 <= count_sym_next(7);

  -- Data counter always ticks by 1.
  count_data_next   <= ('0' & count_data) + ("" & count_clk_wrap);
  count_data_wrap2  <= count_data_next(1) xor count_data(1);
  count_data_wrap32 <= count_data_next(5);

  -- We can use the same Hamming encoder, provided that the
  -- upper (unused) data bits that go into the encoder are all 0, so
  -- as to not affect the lower parity bits.

  hamm_word11 <= data_latch;

  hamm: entity work.rataser_hamming_encode
    port map(data   => hamm_word11,
             output => hamm_data16);

  -- Bit of the match-pattern, 10 pattern or info bit.
  bit_out_sym       <= sym_sync16(to_integer(count_sym_p(3 downto 0)));
  bit_out_10        <= not count_sym_p(0);
  bit_out_info      <=      info_bit_latch xor count_sym_p(0);
  bit_out_data_info <= data_info_bit_latch xor count_sym_p(0);
  -- Which bit should idle pattern send?
  -- When <= 16.
  idle_sym  <= '1' when (count_sym_p(6 downto 4) = "000") else '0';
  -- When % 64 >= 32 and % 32 >= 8 gives when >= 40.
  idle_info <= '1' when (count_sym_p(5) = '1' and
                         count_sym_p(4 downto 3) /= "00") else '0';
  -- Long pattern of 10 sent in the second half of the idle pattern.
  idle_10_long <= count_sym_p(6) and not idle_info;
  -- Select the idle pattern bit to transmit.
  bit_out_idle <=
    bit_out_data_info when (send_info_p = '1') else
    bit_out_sym       when (idle_sym = '1') else
    bit_out_info      when (idle_info = '1') else
    bit_out_10;
  -- When sending data.
  bit_out_data <=
    count_sym_p(0) xor
    hamm_data16_latch(to_integer(count_data_p(4 downto 1)));
  -- Choose.
  bit_out <=
    bit_out_data when (send_data_p = '1') else
    bit_out_idle xor send_start_p;

  -- Would a started data transmission clobber the match symbol
  -- in the next idle pattern?
  reach_sym_next <=
    '1' when (count_sym(6) = '1') else '0';
  -- Would it clobber the match symbol in this idle pattern?
  reach_sym_this <=
    '1' when (count_sym(6 downto 3) = "0000") else '0';
  -- Would it clobber the long 10 pattern in this idle pattern?
  reach_10 <=
    '1' when (count_sym_wrap128 = '0') else '0';

  -- We do not want to also clobber the match symbol in the next
  -- idle pattern if the three previous (i.e. this and two
  -- previous) match symbols were clobbered.
  recent_clobber_sym_next <=
    clobber_sym and clobber_prev_sym and clobber_prev2_sym;
  -- Do not clobber the match symbol in this idle pattern if
  -- the three previous were clobbered.
  recent_clobber_sym_this <=
    clobber_prev_sym and clobber_prev2_sym and clobber_prev3_sym;
  recent_clobber_10       <=
    clobber_prev_10  and clobber_prev2_10  and clobber_prev3_10;

  -- A data transmission can be started whenever we are wrapping
  -- transmission of 8 slots.  But not at slot 40 (i.e. if we come
  -- from 39), since this is where the info bits are emitted.
  -- It will be taken soon, at the wrap from 42 (to 64) however.
  start_loc <= '1' when (count_sym_wrap8 = '1' and
                         count_sym(5 downto 3) /= "100") else '0';

  can_start <= start_loc and
               (not (reach_sym_next and recent_clobber_sym_next)) and
               (not (reach_sym_this and recent_clobber_sym_this)) and
               (not (reach_10       and recent_clobber_10));

  process(clk)
  begin
    if (rising_edge(clk)) then

      -- Defaults.
      i_transmit_sync <= '0';
      
      -- Counter for pulse itself.
      count_clk <= count_clk_next(count_clk'range);
      if (count_clk_wrap = '1') then
        count_clk <= count_clk_max;
      end if;

      -- Update the counter.
      count_sym <= count_sym_next(count_sym'range);

      count_sym_add_jump <= '0';
      if (count_sym(5 downto 0) = "101001") then -- 41
        count_sym_add_jump <= '1';
      end if;

      -- Update the data counter.
      count_data <= count_data_next(count_data'range);

      -- If user wants to send data, and we are not already sending.
      if (send = '1' and
          send_wait = '0' and
          send_start = '0' and
          send_data = '0') then
        send_wait <= '1';
        -- Copy the data (give the Hamming encoder the data).
        data_latch <= data;
      end if;

      if (send_wait = '1' and
          can_start = '1') then -- Not if we come from 39.
        send_wait <= '0';
        send_start <= '1';
      end if;

      -- When the start pattern has been sent (8 bit slots), we
      -- can start to transmit the data itself.
      if (send_start = '1' and
          count_sym_wrap8 = '1') then
        send_start <= '0';
        send_data <= '1';
        -- Hold the data to be sent
        hamm_data16_latch <= hamm_data16;
        i_transmit_sync <= '1';
        -- Reset data counter.
        count_data <= (others => '0');
        -- Copy the info bit that would have sent either now, or
        -- during the data.  It will be sent at the end of the data.
        data_info_bit_latch <= info_bit_latch;
      end if;

      -- Data transmission ends after 32 bits have been sent.
      -- After the data, the (possibly) delayed info bit it sent.
      -- This ensures alignment with the idle pattern, since
      -- each data transmission then takes 42 slots, and within
      -- any 42 slots, the idle pattern has one 2-bit info bit
      -- transmission and five 8-slot locations.
      if (send_data = '1' and
          count_data_wrap32 = '1') then
        send_data <= '0';
        send_info <= '1';
      end if;

      -- Stop the info bit transmission after two bits sent.
      if (send_info = '1' and
          count_data_wrap2 = '1') then
        send_info <= '0';
      end if;

      -- Keep track of which parts of the idle pattern have been
      -- overwritten by transmissions.

      clobber_sym <=
        clobber_sym or (idle_sym and
                        (send_start_p or send_data_p));
      clobber_10 <=
        clobber_10 or (idle_10_long and
                       (send_start_p or send_data_p));
      if (count_sym_wrap128 = '1') then
        clobber_prev_sym  <= clobber_sym;
        clobber_prev_10   <= clobber_10;
        clobber_prev2_sym <= clobber_prev_sym;
        clobber_prev2_10  <= clobber_prev_10;
        clobber_prev3_sym <= clobber_prev2_sym;
        clobber_prev3_10  <= clobber_prev2_10;
        clobber_sym <= '0';
        clobber_10  <= '0';
      end if;

      -- Pipeline
      count_sym_p <= count_sym;
      count_data_p <= count_data;
      send_data_p <= send_data;
      send_info_p <= send_info;
      send_start_p <= send_start;

      info_new <= '0';
      -- The info bit is only updated every second time.  The reason
      -- is that they may come close, which could be too often for
      -- the multiplexed receiver.  This way, the receiver only use
      -- every second info bit.
      if (count_sym_wrap128 = '1') then
        info_bit_latch <= info_bit;
        -- Request a new info bit.
        info_new <= '1';
      end if;

      -- Latch the value to transmit
      i_transmit <= bit_out;
      i_idle_sync <= count_sym_wrap128;
    end if;
  end process;

  data_send <= send_wait or send_start or send_data or send_info;

  -- Assign the actual output signals
  transmit      <= i_transmit;
  transmit_sync <= i_transmit_sync;
  idle_sync     <= i_idle_sync;

end RTL;
