-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

package rataser_util is

  function log2   (x : positive) return natural;
  function log2o1 (x : positive) return natural;

  procedure rataser_ppl(constant p : in boolean;
                        signal c : in  std_logic;
                        signal i : in  std_logic;
                        signal o : out std_logic);

  procedure rataser_ppl(constant p : in boolean;
                        signal c : in  std_logic;
                        signal i : in  std_logic_vector;
                        signal o : out std_logic_vector);

  procedure rataser_ppl(constant p : in boolean;
                        signal c : in  std_logic;
                        signal i : in  signed;
                        signal o : out signed);

  procedure rataser_ppl(constant p : in boolean;
                        signal c : in  std_logic;
                        signal i : in  unsigned;
                        signal o : out unsigned);

  function rataser_xorshift(x : std_logic_vector;
                            s1 : natural;
                            s2 : natural;
                            s3 : natural) return
    std_logic_vector;

end package rataser_util;

package body rataser_util is

  -- Call as log2(x)     =>  1->0   2,3->1    4,5,6,7->2
  -- Call as log2(x-1)+1 => (1->0)  2  ->1  3,4      ->2  5,6,7,8 -> 3
  function log2 (x : positive) return natural is
    variable temp, log: natural;
  begin
    if (x = 1) then -- avoid warning about zero number of iterations
      return 0;
    end if;
    temp := x / 2;
    log := 0;
    while (temp /= 0) loop
      temp := temp/2;
      log := log + 1;
    end loop;
    return log;
  end function log2;

  -- Call as log2o1(x) => (1->0)  2  ->1  3,4      ->2  5,6,7,8 -> 3
  function log2o1 (x : positive) return natural is
  begin
    if (x = 1) then
      return 0;
    end if;
    return log2(x-1)+1;
  end function log2o1;

  -- Helper procedures such that inserting an optional pipeline stage,
  -- does not look so intimidating.
  procedure rataser_ppl(constant p : in boolean;
                        signal c : in  std_logic;
                        signal i : in  std_logic;
                        signal o : out std_logic) is
  begin
    if (p) then
      if (rising_edge(c)) then
        o <= i;
      end if;
    else
      o <= i;
    end if;
  end;

  procedure rataser_ppl(constant p : in boolean;
                        signal c : in  std_logic;
                        signal i : in  std_logic_vector;
                        signal o : out std_logic_vector) is
  begin
    if (p) then
      if (rising_edge(c)) then
        o <= i;
      end if;
    else
      o <= i;
    end if;
  end;

  procedure rataser_ppl(constant p : in boolean;
                        signal c : in  std_logic;
                        signal i : in  signed;
                        signal o : out signed) is
  begin
    if (p) then
      if (rising_edge(c)) then
        o <= i;
      end if;
    else
      o <= i;
    end if;
  end;

  procedure rataser_ppl(constant p : in boolean;
                        signal c : in  std_logic;
                        signal i : in  unsigned;
                        signal o : out unsigned) is
  begin
    if (p) then
      if (rising_edge(c)) then
        o <= i;
      end if;
    else
      o <= i;
    end if;
  end;

  function rataser_xorshift(x : std_logic_vector;
                            s1 : natural;
                            s2 : natural;
                            s3 : natural) return
    std_logic_vector is
    variable l : natural := x'length;
    variable y : std_logic_vector(31 downto 0);
  begin
    y := x;

    y := y xor (    y(l-1-s1 downto 0) & (s1-1 downto 0 => '0'));
    y := y xor ((s2-1 downto 0 => '0') & y(l-1 downto s2));
    y := y xor (    y(l-1-s3 downto 0) & (s3-1 downto 0 => '0'));

    return y;
  end;

end rataser_util;
